/*
 * $Header: /Users/pizlo/Repository/tsf/bounds_check_test.c,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * bounds_check_test.c -- tests whether or not the bounds checks are being
 *                        done correctly.
 * by Filip Pizlo, 2003, 2004, 2005
 *
 * FIXME: currently, a failure in this test is pretty much always going to be
 * a seg fault/bus error/trap.  That's not going to be very ergonomic.
 */

#include "tsf.h"
#include "tsf_format.h"
#include "write_cases.h"
#include "gpc.h"

#if defined(TSF_HAVE_MMAP) &&\
    defined(TSF_HAVE_MUNMAP) &&\
    defined(TSF_HAVE_MPROTECT) &&\
    defined(TSF_HAVE_GETPAGESIZE) &&\
    (defined(TSF_HAVE_MAP_ANON) ||\
     defined(TSF_HAVE_MAP_ANONYMOUS))
    
#include "tsf_test.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdarg.h>

#ifdef TSF_HAVE_MAP_ANON
#define TSF_MAP_ANON MAP_ANON
#else
#define TSF_MAP_ANON MAP_ANONYMOUS
#endif

static tsf_bool_t use_generation;

static void doit(tsf_reflect_t *data) {
    /* plan:
     * 1) generate a buffer
     * 2) repeat the following starting with size and going to 0:
     *   1) mmap some stuff and mprotect it.  use MAP_ANON
     *   2) copy the buffer into the mmap'd memory
     *   3) invoke either reflective or generation parser, depending on
     *      what we are configured to do
     *   4) munmap */
    
    tsf_buffer_t *buffer;
    int32_t size;
    size_t page_size=getpagesize();
    tsf_parser_t *parser=NULL;  /* make GCC happy */
    
    TAT(data);
    
    if (use_generation) {
        tsf_type_t *type=tsf_type_dup(tsf_reflect_get_type(data));
        
        TEST_ASSERT_TSFERROR(tsf_native_type_produce_some_mapping(&type,
								  tsf_true));
        
        TEST_ASSERT_TSFERROR(parser=tsf_parser_create(type,
                                                      TSF_EN_DEFAULT,
						      TSF_CD_DEFAULT));
        
        tsf_type_destroy(type);
    }
    
    TEST_ASSERT_TSFERROR(buffer=
        tsf_reflect_make_buffer(data));
    
    for (size=buffer->size;
         size>=0;
         --size) {
        void *addr;
        void *fail_point;
        tsf_buffer_t crazy_buffer;
        
        TEST_ASSERT_ERRNO((addr=mmap(NULL,
                                     size+page_size*2,
                                     PROT_READ|PROT_WRITE,
                                     TSF_MAP_ANON|MAP_PRIVATE,
                                     -1,
                                     0))!=(void*)-1);
        
        fail_point=(void*)((((uintptr_t)addr)+size+page_size-1)
                           &~(page_size-1));
        
        TEST_ASSERT_ERRNO(mprotect(fail_point,
                                   page_size,
                                   0)>=0);
        
        crazy_buffer.region=NULL;
        crazy_buffer.type=buffer->type;
        crazy_buffer.types=buffer->types;
        crazy_buffer.keep=tsf_false;
        crazy_buffer.data=((uint8_t*)fail_point)-size;
        crazy_buffer.size=size;
        
        memcpy(crazy_buffer.data,
               buffer->data,
               size);
        
        if (use_generation) {
            void *data=tsf_parser_parse(parser,&crazy_buffer);
            if ((uint32_t)size==buffer->size) {
                TEST_ASSERT_TSFERROR(data);
                tsf_region_free(data);
            } else {
                TET(data, TSF_E_PARSE_ERROR);
            }
        } else {
            tsf_reflect_t *data=tsf_reflect_from_buffer(&crazy_buffer);
            if ((uint32_t)size==buffer->size) {
                TEST_ASSERT_TSFERROR(data);
                tsf_reflect_destroy(data);
            } else {
                TET(data, TSF_E_PARSE_ERROR);
            }
        }
        
        munmap(addr,size+page_size*2);
    }
    
    if (use_generation) {
        tsf_parser_destroy(parser);
    }
    
    tsf_buffer_destroy(buffer);
    tsf_reflect_destroy(data);
}

int main(int c,char **v) {
    TEST_INIT(c,v);
    
    TEST_PUSH("reflect");
    use_generation=tsf_false;
    do_tests(tsf_false,doit);
    TEST_POP();
    
    TEST_PUSH("generation");
    use_generation=tsf_true;
    do_tests(tsf_false,doit);
    TEST_POP();
    
    TEST_DONE();
    
    return 0;
}

#else

int main(int c,char **v) {
    printf("%s: cannot do test because memory mapping and "
           "protecting doesn't work.\n",v[0]);
    return 0;
}

#endif

