/*
 * $Header: /Users/pizlo/Repository/tsf/error_test.c,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * error_test.c -- tests error handling
 * by Filip Pizlo, 2004, 2005
 *
 * NOTE: this is the one test that I won't want to run with fault injection.
 * Mainly because I don't feel like going to the trouble of telling the fault
 * injection code how to deal with tsf_set_error().
 */

#include "tsf_internal.h"
#include "tsf_test.h"

#include <errno.h>
#include <signal.h>

#ifdef TSF_FAULT_INJECT

int main(int c,char **v) {
    printf("%s: this test makes no sense with fault injection\n",
           v[0]);
    return 0;
}

#else

#ifdef TSF_HAVE_BDB
#include <db.h>
#endif

#ifdef TSF_HAVE_ZLIB
#include <zlib.h>
#endif

#ifdef TSF_HAVE_BZIP2
#include <bzlib.h>
#endif

static void test_set_error(tsf_error_t code,
			   const char *msg) {
    tsf_set_error(code, msg);
    TA(tsf_get_error_code() == code);
    TASE(msg,tsf_get_error_suffix());
}

static void test_set_errno(int my_errno,
			   const char *msg) {
    errno=my_errno;
    tsf_set_errno(msg);
    TA(tsf_get_error_code() == TSF_E_ERRNO);
    TA(tsf_get_system_errno() == my_errno);
    TASE(strerror(my_errno),tsf_get_error_infix());
    TASE(msg,tsf_get_error_suffix());
}

static void test_set_specific_errno(int my_errno,
				    const char *msg) {
    tsf_set_specific_errno(my_errno,msg);
    TA(tsf_get_error_code() == TSF_E_ERRNO);
    TA(tsf_get_system_errno() == my_errno);
    TASE(strerror(my_errno),tsf_get_error_infix());
    TASE(msg,tsf_get_error_suffix());
}

#ifdef TSF_HAVE_BDB

static void test_set_bdb_error_code(int bdb_error,
				    const char *msg) {
    tsf_set_bdb_error(bdb_error,msg);
    TA(tsf_get_error_code() == TSF_E_BDB_ERROR);
    TA(tsf_get_bdb_error_code() == bdb_error);
    TASE(db_strerror(bdb_error),tsf_get_error_infix());
    TASE(msg,tsf_get_error_suffix());
}

static void test_set_errno_via_bdb_error_code(int my_errno,
					      const char *msg) {
    tsf_set_bdb_error(my_errno,msg);
    TA(tsf_get_error_code() == TSF_E_ERRNO);
    TA(tsf_get_system_errno() == my_errno);
    TASE(strerror(my_errno),tsf_get_error_infix());
    TASE(msg,tsf_get_error_suffix());
}

#endif

static void test_set_deadly_signal(int signal,
				   const char *msg) {
    tsf_set_deadly_signal(signal,msg);
    TA(tsf_get_error_code() == TSF_E_DEATH_BY_SIGNAL);
    TA(tsf_get_deadly_signal() == signal);
    TASE(tsf_asprintf("Received signal #%d",signal),
	 tsf_get_error_infix());
    TASE(msg,tsf_get_error_suffix());
}

#ifdef TSF_HAVE_ZLIB

static void test_set_zlib_error(int zlib_error,
                                const char *zlib_msg,
				const char *msg,
                                int resulting_code,
                                int resulting_errno) {
    tsf_set_zlib_error(zlib_error,zlib_msg,msg);
    TA(tsf_get_error_code()==resulting_code);
    if (resulting_code==TSF_E_ZLIB_ERROR) {
        TA(tsf_get_zlib_error_code()==zlib_error);
        if (zlib_msg!=NULL) {
            TA(tsf_get_zlib_error_msg()==zlib_msg);
	    TASE(zlib_msg,tsf_get_error_infix());
        } else {
	    TA(tsf_get_error_infix()!=NULL);
	}
    }
    if (resulting_errno>=0) {
        TA(tsf_get_system_errno()==resulting_errno);
	TASE(strerror(resulting_errno),
	     tsf_get_error_infix());
    }
    TASE(msg,tsf_get_error_suffix());
}

#endif

#ifdef TSF_HAVE_BZIP2

static void test_set_libbzip2_error(int bzip2_error,
				    const char *msg,
                                    int resulting_code,
                                    int resulting_errno) {
    tsf_set_libbzip2_error(bzip2_error,msg);
    TA(tsf_get_error_code()==resulting_code);
    if (resulting_code==TSF_E_LIBBZIP2_ERROR) {
        TA(tsf_get_libbzip2_error_code()==bzip2_error);
	TA(tsf_get_error_infix()!=NULL);
    }
    if (resulting_errno>=0) {
        TA(tsf_get_system_errno()==resulting_errno);
	TASE(strerror(resulting_errno),
	     tsf_get_error_infix());
    }
    TASE(msg,tsf_get_error_suffix());
}

#endif

int main(int c,char **v) {
    tsf_error_t i;

    TEST_INIT(c,v);
    
    for (i=0;
         i<TSF_E_LAST;
         ++i) {
        TEST_PUSH("i=%d",i);
        switch (i) {
            case TSF_E_ERRNO:
            case TSF_E_BDB_ERROR:
            case TSF_E_DEATH_BY_SIGNAL:
                break;
            default:
                DO_TEST(test_set_error(i,NULL));
                DO_TEST(test_set_error(i,"hello, world"));
                break;
        }
        TEST_POP();
    }
    
    /* just try a couple of system errno's */
    DO_TEST(test_set_errno(EFAULT,NULL));
    DO_TEST(test_set_errno(EPERM,NULL));
    DO_TEST(test_set_errno(ENOSYS,NULL));
    DO_TEST(test_set_errno(EINTR,NULL));
    
    DO_TEST(test_set_errno(EFAULT,"hello, world"));
    DO_TEST(test_set_errno(EPERM,"hello, world"));
    DO_TEST(test_set_errno(ENOSYS,"hello, world"));
    DO_TEST(test_set_errno(EINTR,"hello, world"));
    
    /* do it again... */
    DO_TEST(test_set_specific_errno(EFAULT,NULL));
    DO_TEST(test_set_specific_errno(EPERM,NULL));
    DO_TEST(test_set_specific_errno(ENOSYS,NULL));
    DO_TEST(test_set_specific_errno(EINTR,NULL));
    
    DO_TEST(test_set_specific_errno(EFAULT,"hello, world"));
    DO_TEST(test_set_specific_errno(EPERM,"hello, world"));
    DO_TEST(test_set_specific_errno(ENOSYS,"hello, world"));
    DO_TEST(test_set_specific_errno(EINTR,"hello world"));
    
#ifdef TSF_HAVE_BDB
    DO_TEST(test_set_bdb_error_code(DB_NOTFOUND,NULL));
    DO_TEST(test_set_bdb_error_code(DB_LOCK_DEADLOCK,NULL));
    
    DO_TEST(test_set_errno_via_bdb_error_code(EFAULT,NULL));
    DO_TEST(test_set_errno_via_bdb_error_code(EPERM,NULL));
    DO_TEST(test_set_errno_via_bdb_error_code(ENOSYS,NULL));
    DO_TEST(test_set_errno_via_bdb_error_code(EINTR,NULL));

    DO_TEST(test_set_bdb_error_code(DB_NOTFOUND,"hello"));
    DO_TEST(test_set_bdb_error_code(DB_LOCK_DEADLOCK,"hello"));
    
    DO_TEST(test_set_errno_via_bdb_error_code(EFAULT,"bob"));
    DO_TEST(test_set_errno_via_bdb_error_code(EPERM,"fred"));
    DO_TEST(test_set_errno_via_bdb_error_code(ENOSYS,"joe"));
    DO_TEST(test_set_errno_via_bdb_error_code(EINTR,"bill"));
#endif
    
    DO_TEST(test_set_deadly_signal(SIGINT,NULL));
    DO_TEST(test_set_deadly_signal(SIGKILL,NULL));   /* heh */
    DO_TEST(test_set_deadly_signal(SIGQUIT,NULL));
        
    DO_TEST(test_set_deadly_signal(SIGINT,"foo"));
    DO_TEST(test_set_deadly_signal(SIGKILL,"bar"));   /* heh */
    DO_TEST(test_set_deadly_signal(SIGQUIT,"snafu"));
        
#ifdef TSF_HAVE_ZLIB
    errno=EINVAL;
    DO_TEST(test_set_zlib_error(Z_ERRNO,NULL,NULL,TSF_E_ERRNO,EINVAL));
    
    DO_TEST(test_set_zlib_error(Z_MEM_ERROR,NULL,NULL,TSF_E_ERRNO,ENOMEM));
    DO_TEST(test_set_zlib_error(Z_DATA_ERROR,NULL,NULL,TSF_E_ZLIB_ERROR,-1));
    DO_TEST(test_set_zlib_error(Z_DATA_ERROR,"data error",NULL,TSF_E_ZLIB_ERROR,-1));

    errno=EINVAL;
    DO_TEST(test_set_zlib_error(Z_ERRNO,NULL,"z",TSF_E_ERRNO,EINVAL));
    
    DO_TEST(test_set_zlib_error(Z_MEM_ERROR,NULL,"l",TSF_E_ERRNO,ENOMEM));
    DO_TEST(test_set_zlib_error(Z_DATA_ERROR,NULL,"i",TSF_E_ZLIB_ERROR,-1));
    DO_TEST(test_set_zlib_error(Z_DATA_ERROR,"data error","b",TSF_E_ZLIB_ERROR,-1));
#endif
        
#ifdef TSF_HAVE_BZIP2
    errno=EINVAL;
    DO_TEST(test_set_libbzip2_error(BZ_IO_ERROR,NULL,TSF_E_ERRNO,EINVAL));
    
    DO_TEST(test_set_libbzip2_error(BZ_MEM_ERROR,NULL,TSF_E_ERRNO,ENOMEM));
    DO_TEST(test_set_libbzip2_error(BZ_UNEXPECTED_EOF,NULL,TSF_E_ERRONEOUS_EOF,-1));
    DO_TEST(test_set_libbzip2_error(BZ_DATA_ERROR,NULL,TSF_E_LIBBZIP2_ERROR,-1));

    errno=EINVAL;
    DO_TEST(test_set_libbzip2_error(BZ_IO_ERROR,"b",TSF_E_ERRNO,EINVAL));
    
    DO_TEST(test_set_libbzip2_error(BZ_MEM_ERROR,"z",TSF_E_ERRNO,ENOMEM));
    DO_TEST(test_set_libbzip2_error(BZ_UNEXPECTED_EOF,"i",TSF_E_ERRONEOUS_EOF,-1));
    DO_TEST(test_set_libbzip2_error(BZ_DATA_ERROR,"p",TSF_E_LIBBZIP2_ERROR,-1));
#endif

    TEST_DONE();
    
    return 0;
}

#endif

