/*
 * $Header: /Users/pizlo/Repository/tsf/good_system.h,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * good_system.h -- an alternate implementation of system() that suppresses all
 *                  output from the shell.
 * by Filip Pizlo, 2004, 2005
 *
 * I put this in a header because:
 * 1) it is only used from one place in the TSF library, and
 * 2) I want to be able to test this directly in configure.
 *
 * Why go to all this trouble of re-implementing system() just to suppress
 * output?
 * 1) this suppresses the shell's own output, if there were any (for example
 *    due to a verbose bashrc file).
 * 2) this version of system is also interruptible, unlike the libc version.
 * 3) this version is under my control.  This means that if bugs are found,
 *    I can fix them rather than having to fuck around with working around
 *    them.
 */

#ifndef FP_GOOD_SYSTEM_H
#define FP_GOOD_SYSTEM_H

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>

extern char **environ;

/* you should be able to call this just as you would call system() */
static int good_system(const char *shell_command) {
    int status[2];
    int child_pid;
    int stat_bits;
    int res;
    int my_errno;
    char *cur;
    
    if (pipe(status)<0) {
        return -1;
    }
    
    /* make the reading end non-blocking.  this isn't strictly necessary,
     * but it will help us catch errors in the event that we fuck up. */
    res=fcntl(status[0],F_GETFL,0);
    if (res<0) {
        close(status[0]);
        close(status[1]);
        return -1;
    }
    
    if (fcntl(status[0],F_SETFL,res|O_NONBLOCK)<0) {
        close(status[0]);
        close(status[1]);
        return -1;
    }
    
    child_pid=fork();
    if (child_pid<0) {
        close(status[0]);
        close(status[1]);
        return -1;
    } else if (child_pid==0) {
        const char *argv[4];
        int fd,i,n;
        
        for (n=0;environ[n];++n) {}
        
        for (i=0;i<n;++i) {
            if (strstr(environ[i],"DYLD_INSERT_LIBRARIES")==environ[i] ||
                strstr(environ[i],"LD_PRELOAD")==environ[i] ||
                strstr(environ[i],"DYLD_FORCE_FLAT_NAMESPACE")==environ[i]) {
                environ[i--]=environ[--n];
            }
        }
        
        environ[n]=NULL;
        
        fd=open("/dev/null",O_RDONLY);
        if (fd<0) {
            goto fail_in_child;
        }
        
        if (dup2(fd,0)<0) {
            goto fail_in_child;
        }
        
        close(fd);
        
        fd=open("/dev/null",O_WRONLY);
        if (fd<0) {
            goto fail_in_child;
        }
        
        if (dup2(fd,1)<0) {
            goto fail_in_child;
        }
        
        if (dup2(fd,2)<0) {
            goto fail_in_child;
        }
        
        close(fd);
        
        close(status[0]);
        
        /* set close-on-exec */
        if (fcntl(status[1],F_SETFD,1)<0) {
            goto fail_in_child;
        }
        
        argv[0]="/bin/sh";
        argv[1]="-c";
        argv[2]=shell_command;
        argv[3]=NULL;
        
        execv(argv[0],(char * const *)argv);

fail_in_child:
        my_errno=errno;
        write(status[1],&my_errno,sizeof(my_errno));
        
        exit(1);
    }
    
    close(status[1]);
    
    for (;;) {
        if (waitpid(child_pid,
                    &stat_bits,
                    0)<0) {
            if (errno==EINTR) {
                continue;
            }
            close(status[0]);
            return -1;
        }
        break;
    }
    
    cur=(char*)&my_errno;
    while (cur<(char*)(&my_errno+1)) {
        res=read(status[0],cur,(char*)(&my_errno+1)-cur);
        if (res<0) {
            if (errno==EINTR) {
                continue;
            }
            close(status[0]);
            return -1;
        }
        if (res==0) {
            if (cur!=(char*)&my_errno) {
                /* wtf?  this means that there was an EOF
                 * after reading at least one byte but before
                 * reading the whole errno.  that really should
                 * not happen. */
                close(status[0]);
                return -1;
            }
            /* ok!  since no errno was written, we know that
             * everything went just fine. */
            close(status[0]);
            return stat_bits;
        }
        cur+=res;
    }
    
    /* an error occurred */
    errno=my_errno;
    close(status[0]);
    return -1;
}

#endif

