/*
 * gpc_code_gen_util.c -- code generation utilities
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "gpc_internal.h"
#include "tsf_format.h"
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>

/* code generation debugging */

/* NOTE: It is now the case that the count may be incremented multiple times,
 * in order to insure that files are not overwritten.  This allows multiple
 * processes to use the same directory for code gen debugging.  However, the
 * multi-increment appproach may be overkill, since I could have probably just
 * as well have used the PID. */

static volatile uint32_t count=0;

#ifdef HAVE_PTHREAD
#include <pthread.h>

static pthread_once_t once_control = PTHREAD_ONCE_INIT;
static pthread_mutex_t count_lock;

static void count_init() {
    pthread_mutex_init(&count_lock,NULL);
}

static uint32_t next_count() {
    uint32_t result;
    /* NOTE: pthread_once can be expensive on some platforms.  But that
       might not be a problem here. */
    pthread_once(&once_control,count_init);
    pthread_mutex_lock(&count_lock);
    result=count++;
    pthread_mutex_unlock(&count_lock);
    return result;
}
#else
static uint32_t next_count() {
    return count++;
}
#endif

void gpc_dyncc_code_gen_debug(const char *type,
                              const char *src_filename) {
    const char *tsf_code_gen_debug=getenv("TSF_CODE_GEN_DEBUG");
    const char *tsf_code_gen_debug_path;
    char *dest_filename;
    int fd_src,fd_trg;
    uint32_t cur_cnt;
    
    if (tsf_code_gen_debug==NULL || strcasecmp(tsf_code_gen_debug,"YES")) {
        return;
    }
    
    tsf_code_gen_debug_path=getenv("TSF_CODE_GEN_DEBUG_PATH");
    
    if (tsf_code_gen_debug_path==NULL) {
        return;
    }
    
    fd_src=open(src_filename,O_RDONLY);
    if (fd_src<0) {
        return;
    }
    
    /* loop until we hit a filename that isn't in use */
    for (cur_cnt=count;;cur_cnt=next_count()) {
        dest_filename=tsf_asprintf("%s/" fui32 "_%s",
                                   tsf_code_gen_debug_path,
                                   cur_cnt,
                                   type);
        if (dest_filename==NULL) {
            close(fd_src);
            return;
        }
        
        fd_trg=open(dest_filename,O_CREAT|O_EXCL|O_WRONLY,0600);
        free(dest_filename);
        
        if (fd_trg>=0) {
            break;
        }
        
        if (errno!=EEXIST) {
            close(fd_src);
            return;
        }
    }
    
    for (;;) {
        char buf[1024];
        int res=read(fd_src,buf,sizeof(buf));
        if (res<=0) {
            break;
        }
        if (write(fd_trg,buf,res)!=res) {
            break;
        }
    }
    
    close(fd_src);
    close(fd_trg);
}

void gpc_code_gen_debug(const char *type,
                        gpc_proto_t *proto) {
    const char *tsf_code_gen_debug=getenv("TSF_CODE_GEN_DEBUG");
    const char *tsf_code_gen_debug_path;
    char *dest_filename;
    FILE *out;
    uint32_t cur_cnt;
    
    /* printf("dumping %s\n",type); */

    if (tsf_code_gen_debug==NULL || strcasecmp(tsf_code_gen_debug,"YES")) {
        return;
    }
    
    tsf_code_gen_debug_path=getenv("TSF_CODE_GEN_DEBUG_PATH");
    
    if (tsf_code_gen_debug_path==NULL) {
        return;
    }
    
    for (cur_cnt=count;;cur_cnt=next_count()) {
        int fd;
        
        dest_filename=tsf_asprintf("%s/" fui32 "_%s",
                                   tsf_code_gen_debug_path,
                                   cur_cnt,
                                   type);
        if (dest_filename==NULL) {
            return;
        }
        
        fd=open(dest_filename,O_CREAT|O_EXCL|O_WRONLY,0600);
        free(dest_filename);

        if (fd>=0) {
            out=fdopen(fd,"w");
            
            if (out==NULL) {
		close(fd);
                return;
            }
            
            break;
        }
        
        if (errno!=EEXIST) {
            /* printf("error not eexist\n"); */
            return;
        }
    }
    
    gpc_proto_print(proto,out);
    
    fclose(out);
}

/* stupid stuff */

void gpc_unlink_free(char *name) {
    unlink(name);
    free(name);
}

void gpc_rmdir_free(char *name) {
    rmdir(name);
    free(name);
}

/* this isn't a real rm -rf; it only goes down one level. */
void gpc_rmdashrf_free(char *name) {
    DIR *dir;
    dir=opendir(name);
    if (dir!=NULL) {
	for (;;) {
	    struct dirent entry,*result=NULL;
	    int res=readdir_r(dir,&entry,&result);
	    if (res!=0 || result==NULL) {
		break;
	    }
	    if (strcmp(entry.d_name,".") &&
		strcmp(entry.d_name,"..")) {
		char *flnm=tsf_asprintf("%s/%s",name,entry.d_name);
		if (flnm!=NULL) {
		    unlink(flnm);
		    free(flnm);
		}
	    }
	}
	closedir(dir);
	rmdir(name);
    }
    free(name);
}

#ifdef TSF_HAVE_MKDTEMP
char *gpc_mkdtemp(char *tmpdir,char *prefix) {
    char *temp_file_name=tsf_asprintf("%s/%sXXXXXXXX",tmpdir,prefix);
    char *result;
    tsf_bool_t in_safe_code;
    
    if (temp_file_name==NULL) {
        return NULL;
    }
    
    in_safe_code=tsf_set_in_safe_code(tsf_false);
    result=mkdtemp(temp_file_name);
    tsf_set_in_safe_code(in_safe_code);
    
    return result;
}
#else
#ifdef TSF_HAVE_TEMPNAM
char *gpc_mkdtemp(char *tmpdir,char *prefix) {
    char *result;
    tsf_bool_t in_safe_code;
    
    in_safe_code=tsf_set_in_safe_code(tsf_false);
    result=tempnam(tmpdir,prefix);
    tsf_set_in_safe_code(in_safe_code);
    
    if (result==NULL) {
        return NULL;
    }
    
    if (mkdir(result,0700)<0) {
        free(result);
        return NULL;
    }

    return result;
}
#else

#error "Dunno how to create temporary directories on your system"

#endif
#endif


