/*
 * gpc_dyncc.c -- dynamic C compilation
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "gpc_internal.h"
#include "tsf_format.h"
#include "tsf_build_defines.h" /* not a real header file; gives us DYNCC_BUILD_CMD_SUFFIX */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>

#ifndef SKIP_DYNCC
    #ifdef CAN_BUILD_MODULE
    
        #ifdef HAVE_NS_LINK_MODULE_IN_DYLD_H
            #include <mach-o/dyld.h>
            #define USE_NS_MODULE
        #else
            #ifdef HAVE_DLOPEN_IN_DLFCN_H
                #define USE_DLFCN
                #include <dlfcn.h>
            #else
                #define SKIP_DYNCC
            #endif
        #endif
    
    #else
    
        #define SKIP_DYNCC
    
    #endif
#endif

#ifdef HAVE_PTHREAD
#include <pthread.h>

static pthread_once_t once_control = PTHREAD_ONCE_INIT;
static pthread_mutex_t dyncc_lock;
static volatile pthread_t holder;
static volatile unsigned lock_cnt;

static void dccl_init() {
    pthread_mutex_init(&dyncc_lock,NULL);
    lock_cnt=0;
}

void gpc_dyncc_lock() {
    pthread_t me;
    /* NOTE: pthread_once can be expensive on some platforms.  But that
       might not be a problem here (it's a fracking global lock we're
       initializing). */
    pthread_once(&once_control,dccl_init);
    me=pthread_self();
    if (lock_cnt==0 || !pthread_equal(me,holder)) {
        pthread_mutex_lock(&dyncc_lock);
        holder=me;
    }
    lock_cnt++;
}

void gpc_dyncc_unlock() {
    lock_cnt--;
    if (lock_cnt==0) {
        pthread_mutex_unlock(&dyncc_lock);
    }
}

#else

void gpc_dyncc_lock() {}
void gpc_dyncc_unlock() {}

#endif

#ifndef SKIP_DYNCC

#ifdef GOOD_SYSTEM_WORKS
#include "good_system.h"
#endif

tsf_bool_t gpc_dyncc_supported() {
    return tsf_true;
}

/* this checker is for fprintf's */
#define CF(exp) do {				\
	if ((exp)<0) {				\
	    goto fprintf_failure;		\
	}					\
    } while (0)

tsf_bool_t do_search(gpc_intable_t *intable,
		     uint32_t hash_code,
		     const char *cache_dir,
		     char **key_name_to_use,
		     char **filename) {
    unsigned i;
    char *dir_name;
    int res;
    
    if (key_name_to_use!=NULL) {
	*key_name_to_use=NULL;
    }
    *filename=NULL;

    dir_name=tsf_asprintf("%s/data",cache_dir);
    if (dir_name==NULL) {
	tsf_set_errno("Could not tsf_asprintf()");
	return tsf_false;
    }

    res=mkdir(dir_name,0755);
    if (res<0 && errno!=EEXIST) {
	tsf_set_errno("Could not make directory at %s",
		      dir_name);
	free(dir_name);
	return tsf_false;
    }

    free(dir_name);

    for (i=0;;++i) {
	char *key_name;
	int fd;
	tsf_comparing_writer_t comp_wtr;
	uint8_t c;
	tsf_bool_t bres;

	key_name=tsf_asprintf("%s/data/" fui32 "_%u_key",
			      cache_dir,
			      hash_code,
			      i);
	if (key_name==NULL) {
	    tsf_set_errno("Could not tsf_asprintf()");
	    return tsf_false;
	}
	
	fd=open(key_name,O_RDONLY);
	if (fd<0) {
	    if (errno==ENOENT) {
		/* not found, but I know where to put the bucket
		   entry */
		if (key_name_to_use!=NULL) {
		    *filename=tsf_asprintf("%s/data/" fui32 "_%u_data",
					   cache_dir,
					   hash_code,
					   i);
		    if (*filename==NULL) {
			tsf_set_errno("Could not tsf_asprintf()");
			return tsf_false;
		    }
		    *key_name_to_use=key_name;
		}
		return tsf_true;
	    }
	    tsf_set_errno("Could not open key file at %s",
			  key_name);
	    free(key_name);
	    return tsf_false;
	}

	free(key_name);
	
	comp_wtr.reader=tsf_fd_reader;
	comp_wtr.arg=&fd;
	if (!gpc_intable_write(intable,
			       tsf_comparing_writer,
			       &comp_wtr)) {
	    close(fd);
	    if (tsf_get_error_code() == TSF_E_COMPARE_FAIL ||
		tsf_get_error_code() == TSF_E_ERRONEOUS_EOF ||
		tsf_get_error_code() == TSF_E_UNEXPECTED_EOF) {
		/* comparison failed. */
		continue;
	    }
	    return tsf_false;
	}
	
	bres=tsf_fd_reader(&fd,&c,1);
	close(fd);

	if (bres) {
	    /* comparison failed, since the file is larger than
	       our intable */
	    continue;
	} else if (tsf_get_error_code()==TSF_E_UNEXPECTED_EOF) {
	    /* yar!!! */
	    *filename=tsf_asprintf("%s/data/" fui32 "_%u_data",
				   cache_dir,
				   hash_code,
				   i);
	    if (*filename==NULL) {
		tsf_set_errno("Could not tsf_asprintf()");
		return tsf_false;
	    }
	    return tsf_true;
	} else {
	    return tsf_false;
	}
    }
}

tsf_bool_t do_code_gen(gpc_intable_t *intable,
		       const char *dir_name,
		       const char *mod_filename,
		       tsf_bool_t unlink_src) {
    char *temp_src_file_name;
    FILE *fout;
    const char *include_prefix;
    uint32_t i;
    gpc_cell_t *cur;
    const char *build_cmd,*build_cmd_suffix;
    char *real_build_cmd;
    int res;

    include_prefix=getenv("TSF_CODE_GEN_INCLUDE_PREFIX");
    if (include_prefix==NULL) {
	include_prefix=TSF_CODE_GEN_INCLUDE_PREFIX;
    }
    
    temp_src_file_name=tsf_asprintf("%s/dyncc.c",dir_name);
    if (temp_src_file_name==NULL) {
	tsf_set_errno("Could not tsf_asprintf() in gpc_dyncc_from_intable()");
	return tsf_false;
    }
    
    fout=fopen(temp_src_file_name,"w");
    if (fout==NULL) {
	tsf_set_errno("Could not fopen() the temporary file: %s",
		      temp_src_file_name);
	free(temp_src_file_name);
	return tsf_false;
    }
    
    /* note: there are some symbols in libtsf that are referred to by
       the code that we generate here.  this is fine most of the time.
       but if libtsf is static, these symbols won't be found.  the
       symbols being referred to are:
       
       tsf_set_errno
       tsf_set_error
       tsf_buffer_read_from_buf
       tsf_buffer_skip_in_buf
       tsf_buffer_write_to_buf
       tsf_buffer_clone_in
       tsf_buffer_create
       tsf_buffer_create_in
       tsf_type_out_map_create
       tsf_type_out_map_destroy
       tsf_type_in_map_from_type_out_map
       tsf_type_in_map_from_type_out_map_in 
       
       we can get around this static linking issue by defining all of
       these functions locally, and having the definition call the real
       function by using a pointer constant (which is available to us).
    */
    
    CF(fprintf(fout,"#include <%sgpc_internal.h>\n",include_prefix));
    CF(fprintf(fout,"#include <%stsf_format.h>\n",include_prefix));
    CF(fprintf(fout,"#define INT_FAILURE() goto failure\n"));

#if defined(HAVE_ADDRESS_OF_LABEL)
    CF(fprintf(fout,"#define LOOKUP_GOTO(targ) goto *labels[targ]\n"));
#else
    CF(fprintf(fout,"#define LOOKUP_GOTO(targ) do {\\\n"));
    CF(fprintf(fout,"    switch (values[targ]) {\\\n"));
    for (cur=gpc_intable_get_stream(intable);
	 cur<gpc_intable_get_stream_end(intable);
	 ++cur) {
	if (!gpc_intable_is_target(intable,cur) ||
	    !gpc_intable_is_live(intable,cur)) {
	    continue;
	}
	CF(fprintf(fout,"        case TSF_UINTPTR_C(" fuptr "):\\\n",cur));
	CF(fprintf(fout,"            goto cell" fuptr ";\\\n",cur));
    }
    CF(fprintf(fout,"        default:\\\n"));
    CF(fprintf(fout,"            goto bad_label;\\\n"));
    CF(fprintf(fout,"    }\\\n"));
    CF(fprintf(fout,"} while (0)\n"));
#endif

    CF(fprintf(fout,"uintptr_t program(void *region,\n"));
    CF(fprintf(fout,"                  gpc_cell_t *args) {\n"));
    
    CF(fprintf(fout,"    static const gpc_cell_t values[]={\n"));
    
    for (cur=gpc_intable_get_stream(intable);
	 cur<gpc_intable_get_stream_end(intable);
	 ++cur) {
	CF(fprintf(fout,"        TSF_UINTPTR_C(" fuptr ")",*cur));
	if (cur+1<gpc_intable_get_stream_end(intable)) {
	    CF(fprintf(fout,","));
	}
	CF(fprintf(fout,"\n"));
    }
    
    CF(fprintf(fout,"    };\n"));

#ifdef HAVE_ADDRESS_OF_LABEL
    CF(fprintf(fout,"    static const void *labels[]={\n"));
    
    for (cur=gpc_intable_get_stream(intable);
	 cur<gpc_intable_get_stream_end(intable);
	 ++cur) {
	CF(fprintf(fout,"        &&"));
	if (gpc_intable_is_target(intable,*cur) &&
	    gpc_intable_is_live(intable,*cur)) {
	    CF(fprintf(fout,"cell" fuptr,*cur));
	} else {
	    CF(fprintf(fout,"bad_label"));
	}
	if (cur+1<gpc_intable_get_stream_end(intable)) {
	    CF(fprintf(fout,","));
	}
	CF(fprintf(fout,"\n"));
    }
    
    CF(fprintf(fout,"    };\n"));
#endif
    
    CF(fprintf(fout,"    uint8_t *buf=NULL,*cur=NULL,*end=NULL;\n"));
    CF(fprintf(fout,"    tsf_type_in_map_t *types;\n"));
    CF(fprintf(fout,"    tsf_type_out_map_t *out_map=NULL;\n"));
    CF(fprintf(fout,"    uint32_t size=0;\n"));
    CF(fprintf(fout,"    tsf_bool_t keep=tsf_false,keep_region=tsf_false;\n"));
    CF(fprintf(fout,"    int8_t i8_tmp;\n"));
    CF(fprintf(fout,"    uint8_t ui8_tmp;\n"));
    CF(fprintf(fout,"    int16_t i16_tmp;\n"));
    CF(fprintf(fout,"    uint16_t ui16_tmp;\n"));
    CF(fprintf(fout,"    int32_t i32_tmp;\n"));
    CF(fprintf(fout,"    uint32_t ui32_tmp;\n"));
    CF(fprintf(fout,"    int64_t i64_tmp;\n"));
    CF(fprintf(fout,"    uint64_t ui64_tmp;\n"));
    CF(fprintf(fout,"    float f_tmp;\n"));
    CF(fprintf(fout,"    double d_tmp;\n"));
    
    for (i=0;i<intable->max_height;++i) {
	CF(fprintf(fout,"    gpc_cell_t stack_" fui32,i));
	if (i<gpc_intable_get_num_args(intable)) {
	    CF(fprintf(fout,"=args[" fui32 "]",i));
	}
	CF(fprintf(fout,";\n"));
    }
    
    for (cur=gpc_intable_get_stream(intable);
	 cur<gpc_intable_get_stream_end(intable);
	 cur+=gpc_instruction_size(cur)) {
	intptr_t height=gpc_intable_get_height(intable,cur);
	if (height<0) {
	    continue;
	}
        
	if (gpc_instruction_stack_effects(*cur)>0) {
	    height+=gpc_instruction_stack_effects(*cur);
	}
        
	if (gpc_intable_is_target(intable,cur)) {
	    CF(fprintf(fout,"cell" fuptr ":\n",cur));
	}
        
	CF(fprintf(fout,"    /* %s */\n",gpc_instruction_to_string(*cur)));
	CF(fprintf(fout,"    {\n"));
#include "gpc_compiler.gen"
	CF(fprintf(fout,"    }\n"));
    }
    
    CF(fprintf(fout,"bad_label:\n"));
    CF(fprintf(fout,"    tsf_set_error(TSF_E_SEMANTIC,\"Jump using value that "
	       "is not a label\");\n"));
    CF(fprintf(fout,"    goto failure;\n"));
    CF(fprintf(fout,"bounds_error:\n"));
    CF(fprintf(fout,"    tsf_set_error(TSF_E_PARSE_ERROR,\"Bounds check "
	       "failure\");\n"));
    CF(fprintf(fout,"failure:\n"));
    CF(fprintf(fout,"    if (keep) {\n"));
    CF(fprintf(fout,"        free(buf);\n"));
    CF(fprintf(fout,"    }\n"));
    CF(fprintf(fout,"    if (keep_region) {\n"));
    CF(fprintf(fout,"        tsf_region_free(region);\n"));
    CF(fprintf(fout,"    }\n"));
    CF(fprintf(fout,"    if (out_map!=NULL) {\n"));
    CF(fprintf(fout,"        tsf_type_out_map_destroy(out_map);\n"));
    CF(fprintf(fout,"    }\n"));
    CF(fprintf(fout,"    return 0;\n"));
    CF(fprintf(fout,"}\n"));

    goto fprintf_success;
 fprintf_failure:
    tsf_set_errno("Failed in fprintf()");
    free(temp_src_file_name);
    fclose(fout);
    return tsf_false;

 fprintf_success:

    fclose(fout);
    
    gpc_dyncc_code_gen_debug("dyncc.c",temp_src_file_name);
    
    build_cmd=getenv("TSF_DYNCC_BUILD_CMD");
    if (build_cmd==NULL) {
	build_cmd=TSF_DYNCC_BUILD_CMD;
    }
    
    build_cmd_suffix=getenv("TSF_DYNCC_BUILD_CMD_SUFFIX");
    if (build_cmd_suffix==NULL) {
	build_cmd_suffix=DYNCC_BUILD_CMD_SUFFIX;
    }
    
    real_build_cmd=tsf_asprintf("%s -o %s %s %s",
				build_cmd,
				mod_filename,
				temp_src_file_name,
				build_cmd_suffix);
    
    if (real_build_cmd==NULL) {
	tsf_set_errno("Could not do tsf_asprintf() to create the "
		      "build command");
	gpc_unlink_free(temp_src_file_name);
	return tsf_false;
    }
    
    /* printf("real_build_cmd = %s\n",real_build_cmd); */
    
#ifdef GOOD_SYSTEM_WORKS
    res=good_system(real_build_cmd);
#else
    res=system(real_build_cmd);
#endif

    if (!WIFEXITED(res) || WEXITSTATUS(res)!=0) {
	/* some sort of error */
        
	if (res<0) {
	    tsf_set_errno("Could not do system() to build the file using "
			  "the command: %s",real_build_cmd);
#ifndef GOOD_SYSTEM_WORKS
	} else if (res==127) {  /* only for vanilla system();
				 * good_system() will report failure to
				 * start the shell as -1 */
	    tsf_set_error(TSF_E_COMPILE_ERROR,
			  "The shell failed to start for the command: %s",
			  real_build_cmd);
#endif
	} else if (WIFSIGNALED(res)) {
	    tsf_set_deadly_signal(WTERMSIG(res),
				  "While running \'%s\'",
				  real_build_cmd);
	} else if (WIFEXITED(res)) {
	    tsf_set_error(TSF_E_COMPILE_ERROR,
			  "Running \'%s\' returned with %d",
			  real_build_cmd,WEXITSTATUS(res));
	} else {
	    tsf_set_error(TSF_E_INTERNAL,
			  "Compiler did not exit but was not signaled, yet "
			  "system() returned.  The command was: \'%s\'",
			  real_build_cmd);
	}

	free(real_build_cmd);
	gpc_unlink_free(temp_src_file_name);
	return tsf_false;
    }
    
    if (unlink_src) {
	gpc_unlink_free(temp_src_file_name);
    } else {
	free(temp_src_file_name);
    }
    free(real_build_cmd);
    return tsf_true;
}

gpc_dyncc_t *do_loading(const char *mod_filename) {
    gpc_dyncc_t *ret;
#ifdef USE_NS_MODULE
    NSObjectFileImage fileImage;
    NSObjectFileImageReturnCode returnCode;
    NSSymbol *symbol;
    NSLinkEditErrors ler;
    int lerno;
    const char *lerfile;
    const char *dyld_error;
    tsf_bool_t in_safe_code;
#endif

    ret=malloc(sizeof(gpc_dyncc_t));
    if (ret==NULL) {
        tsf_set_errno("Could not malloc gpc_dyncc_t");
        return NULL;
    }
    
    gpc_dyncc_lock();

#ifdef USE_DLFCN
    ret->library=dlopen(mod_filename,RTLD_NOW|RTLD_LOCAL);
    
    if (ret->library==NULL) {
        tsf_set_error(TSF_E_DLOADING_ERROR,
                      "Could not dynamically load library: %s",
                      dlerror());
        free(ret);
        gpc_dyncc_unlock();
        return NULL;
    }
    
    /* try both the symbol and the symbol with a leading underscore.
     * the leading underscope seems to be a problem with some versions of
     * dlcompat. */
    ret->symbol=dlsym(ret->library,"program");
    if (ret->symbol==NULL) {
        char *first_error=strdup(dlerror());
        
        if (first_error==NULL) {
            tsf_set_errno("Could not strdup() the return from dlerror()");
            dlclose(ret->library);
            free(ret);
            gpc_dyncc_unlock();
            return NULL;
        }
        
        ret->symbol=dlsym(ret->library,"_program");
        
        if (ret->symbol==NULL) {
            tsf_set_error(TSF_E_DLOADING_ERROR,
                          "Could not load symbol 'program' or '_program' from library "
                          "because: %s; and then %s",
                          first_error,
                          dlerror());
            dlclose(ret->library);
            free(ret);
            free(first_error);
            gpc_dyncc_unlock();
            return NULL;
        }

        free(first_error);
    }
#endif
#ifdef USE_NS_MODULE
    returnCode=NSCreateObjectFileImageFromFile(mod_filename,
                                               &fileImage);
    
    if (returnCode!=NSObjectFileImageSuccess) {
        switch (returnCode) {
            case NSObjectFileImageInappropriateFile:
                tsf_set_error(TSF_E_DLOADING_ERROR,
                              "File is not a bundle");
                break;
            case NSObjectFileImageFailure:
                tsf_set_error(TSF_E_DLOADING_ERROR,
                              "Object file setup failure");
                break;
            case NSObjectFileImageArch:
                tsf_set_error(TSF_E_DLOADING_ERROR,
                              "No object for this architecture");
                break;
            case NSObjectFileImageFormat:
                tsf_set_error(TSF_E_DLOADING_ERROR,
                              "Bad object file format");
                break;
            case NSObjectFileImageAccess:
                tsf_set_error(TSF_E_DLOADING_ERROR,
                              "Cannot read file");
                break;
            default:
                tsf_set_error(TSF_E_DLOADING_ERROR,
                              "Could not load file because of some "
                              "mystery error");
                break;
        }
        free(ret);
        gpc_dyncc_unlock();
        return NULL;
    }
    
    in_safe_code=tsf_set_in_safe_code(tsf_false);
    ret->library=NSLinkModule(fileImage,
                              mod_filename,
                              NSLINKMODULE_OPTION_RETURN_ON_ERROR |
                              NSLINKMODULE_OPTION_PRIVATE |
                              NSLINKMODULE_OPTION_BINDNOW);
    tsf_set_in_safe_code(in_safe_code);
    
    NSDestroyObjectFileImage(fileImage);
    
    if (ret->library==NULL) {
        NSLinkEditError(&ler,&lerno,&lerfile,&dyld_error);
        if (dyld_error && *dyld_error) {
            tsf_set_error(TSF_E_DLOADING_ERROR,
                          "NSLinkModule() failed: %s",
                          dyld_error);
        } else {
            tsf_set_error(TSF_E_DLOADING_ERROR,
                          "NSLinkModule() failed");
        }
        free(ret);
        gpc_dyncc_unlock();
        return NULL;
    }
    
    symbol=NSLookupSymbolInModule(ret->library,"_program");
    
    if (symbol==NULL) {
        NSLinkEditError(&ler,&lerno,&lerfile,&dyld_error);
        if (dyld_error && *dyld_error) {
            tsf_set_error(TSF_E_DLOADING_ERROR,
                          "NSLookupSymbolInModule() for '_program' "
                          "failed: %s",
                          dyld_error);
        } else {
            tsf_set_error(TSF_E_DLOADING_ERROR,
                          "NSLookupSymbolInModule() for '_program' "
                          "failed");
        }
        NSUnLinkModule(ret->library,0);
        free(ret);
        gpc_dyncc_unlock();
        return NULL;
    }
    
    ret->symbol=NSAddressOfSymbol(symbol);
#endif
    
    gpc_dyncc_unlock();
    
    return ret;
}

gpc_dyncc_t *gpc_dyncc_from_intable(gpc_intable_t *intable,
				    const char *cache_dir) {
    char *mod_filename;
    char *dir_name;
    gpc_dyncc_t *ret;
    
    if (cache_dir==TSF_CD_DEFAULT) {
	tsf_set_error(TSF_E_INVALID_ARG,
		      "TSF_CD_DEFAULT is not a valid argument to "
		      "gpc_dyncc_from_intable()");
	return NULL;
    }

    if (cache_dir!=TSF_CD_NO_CACHE) {
	/* how does this work?
	   
	   first, I acquire a shared lock and search to see if the cache
	   contains a module for the intable that I'm dealing with.  note
	   that the key that I use for lookup is a mostly-canonical
	   representation of the intable.  (the only way in which it is
	   not canonical is if it contains pointer constants into the
	   application, which currently should only happen for constructors
	   and destructors.)  if this lookup succeeds, I just load the
	   module.

	   otherwise, I release the shared lock and create a temporary
	   directory as a subdirectory of the cache directory and proceed
	   to do code generation there.  after the code gen and compilation
	   (invocation of the C compiler) is done, I acquire the exclusive
	   lock.  then, I once again do the search to see if the key can
	   be found (it is possible that some other thread or process is
	   doing the same as me and has also gotten this far but before me).
	   if it is found, I delete the temporary directory that I had
	   created and just use whatever I found for the dynamic loading.
	   
	   otherwise, I move the temporary directory into its permanent
	   location.  I also create a key file.  then I do dynamic loading
	   while holding the exclusive lock - this is important, as it
	   allows me to blow away the cache entry if it turns out to not
	   be loadable.  once this either fails or succeeds, I release the
	   exclusive lock.
	   
	   note that the shared/exclusive locking (read/write locking in
	   proper speak) is done using a lock file that is specific to the
	   cache.  The file is locked using flock(), since fcntl()'s read/
	   write locks are a bit more goofy.
	   
	   what does this style of locking mean?  if two applications are
	   using different caches, there will be no contention between
	   them.  if two application use the same cache, there will be
	   some contention.  but if the two applications are not generating
	   types often, then the contention will be minimal (since only
	   shared locks are used for cache hits).

	   one interesting note.  notice that no lock is held while the
	   actual code generation and compilation is going on.  this might
	   at first seem like an optimization.  it isn't - it's actually
	   necessary for correctness, because of the unclear semantics of
	   doing a fork() while holding a lock with flock().  basically,
	   a lock held by flock() is held by both parent and child after
	   the fork(), but if either one of them releases the lock, then
	   neither of them will hold it.  notice that when I invoke the C
	   compiler, I am doing a fork-and-exec.  I don't have a good way
	   of knowing what the exec() will do to any locks that I am
	   holding.  if it is the case that it releases it, then I really
	   cannot protect the code generation phase with a flock(), since
	   the lock might get released as soon as the C compiler is
	   invoked.  this is why I am using a strategy that allows me to
	   release the lock while doing code generation.
	
	   of course, this strategy can be seen as an optimization, since
	   it reduces the amount of time during which the exclusive lock
	   is held.  but on the other hand, it might also be detrimental
	   to performance in the case when two threads or processes
	   simultaneously generate code for the same intable. */
	
	uint32_t hash=0;
	int fd;
	char *lock_filename;
	char *key_name;

	if (!gpc_intable_write(intable,
			       tsf_hashing_writer,
			       &hash)) {
	    return NULL;
	}
	
	if (mkdir(cache_dir,0755)<0 &&
	    errno!=EEXIST) {
	    tsf_set_errno("Could not create directory %s",
			  cache_dir);
	    return NULL;
	}

	lock_filename=tsf_asprintf("%s/lock",
				   cache_dir);
	if (lock_filename==NULL) {
	    tsf_set_errno("Could not asprintf()");
	    return NULL;
	}

	fd=open(lock_filename,O_CREAT|O_RDWR,0644);
	if (fd<0) {
	    tsf_set_errno("Could not open lock file at %s",
			  lock_filename);
	    free(lock_filename);
	    return NULL;
	}
	
	if (flock(fd,LOCK_SH)<0) {
	    tsf_set_errno("Could not acquire shared lock on %s",
			  lock_filename);
	    free(lock_filename);
	    close(fd);
	    return NULL;
	}
	
	if (!do_search(intable,hash,cache_dir,NULL,&dir_name)) {
	    flock(fd,LOCK_UN);
	    close(fd);
	    free(lock_filename);
	    return NULL;
	}

	if (dir_name==NULL) {
	    char *real_temp_dir_name;
	    char *temp_dir_name;

	    /* didn't find it, proceed to step 5 */
	    flock(fd,LOCK_UN);

	    /* now, create a temporary directory, do the code generation,
	       and then acquire the exclusive lock and attempt to move
	       this temp directory to its 'rightful' place.

	       this reduces lock contention and also prevents me from
	       having to think too hard about what happens to a lock when
	       you fork-and-exec while holding it. */
	    
	    real_temp_dir_name=tsf_asprintf("%s/temp",cache_dir);
	    if (real_temp_dir_name==NULL) {
		tsf_set_errno("Could not tsf_asprintf()");
		close(fd);
		free(lock_filename);
		return NULL;
	    }
	    
	    if (mkdir(real_temp_dir_name,0755)<0 &&
		errno!=EEXIST) {
		tsf_set_errno("Could not make directory at %s",
			      real_temp_dir_name);
		free(real_temp_dir_name);
		close(fd);
		free(lock_filename);
		return NULL;
	    }

	    temp_dir_name=gpc_mkdtemp(real_temp_dir_name,"data");
	    if (temp_dir_name==NULL) {
		tsf_set_errno("Could not gpc_mkdtemp()");
		free(real_temp_dir_name);
		close(fd);
		free(lock_filename);
	    }

	    free(real_temp_dir_name);

	    mod_filename=tsf_asprintf("%s/dyncc.out",
				      temp_dir_name);
	    if (mod_filename==NULL) {
		tsf_set_errno("Could not tsf_asprintf()");
		gpc_rmdir_free(temp_dir_name);
		close(fd);
		free(lock_filename);
		return NULL;
	    }
	    
	    if (!do_code_gen(intable,
			     temp_dir_name,
			     mod_filename,
			     tsf_false)) {
		gpc_unlink_free(mod_filename);
		gpc_rmdir_free(temp_dir_name);
		close(fd);
		free(lock_filename);
		return NULL;
	    }

	    free(mod_filename);

	    if (flock(fd,LOCK_EX)<0) {
		tsf_set_errno("Could not acquire exclusive lock on %s",
			      lock_filename);
		close(fd);
		free(lock_filename);
		gpc_rmdashrf_free(temp_dir_name);
		return NULL;
	    }
	    
	    /* search again to make sure something didn't change between
	       lock release and reacquisition */
	    if (!do_search(intable,hash,cache_dir,&key_name,&dir_name)) {
		flock(fd,LOCK_UN);
		close(fd);
		gpc_rmdashrf_free(temp_dir_name);
		free(lock_filename);
		return NULL;
	    }
	    
	    mod_filename=tsf_asprintf("%s/dyncc.out",
				      dir_name);
	    if (mod_filename==NULL) {
		tsf_set_errno("Could not tsf_asprintf()");
		gpc_rmdashrf_free(temp_dir_name);
		flock(fd,LOCK_UN);
		close(fd);
		free(lock_filename);
		return NULL;
	    }
	    
	    if (key_name==NULL) {
		/* found it, so we should remove our crap. */
		gpc_rmdashrf_free(temp_dir_name);
	    } else {
		/* ok.  now make the key file and move the directory
		   in place. */
		int key_fd;
		key_fd=open(key_name,O_WRONLY|O_TRUNC|O_CREAT,0644);
		if (key_fd<0) {
		    tsf_set_errno("Could not create key file at %s",
				  key_name);
		    flock(fd,LOCK_UN);
		    close(fd);
		    gpc_rmdashrf_free(temp_dir_name);
		    free(key_name);
		    free(dir_name);
		    free(lock_filename);
		    free(mod_filename);
		    return NULL;
		}
		
		if (!gpc_intable_write(intable,
				       tsf_fd_writer,
				       &key_fd)) {
		    gpc_rmdashrf_free(temp_dir_name);
		    free(dir_name);
		    free(lock_filename);
		    free(mod_filename);
		    close(key_fd);
		    unlink(key_name);
		    free(key_name);
		    flock(fd,LOCK_UN);
		    close(fd);
		    return NULL;
		}
		
		close(key_fd);

		/* ok, now move the directory in place. */
		if (rename(temp_dir_name,dir_name)<0) {
		    tsf_set_errno("Could not move data directory to %s",
				  dir_name);
		    gpc_unlink_free(key_name);
		    gpc_rmdashrf_free(temp_dir_name);
		    gpc_rmdashrf_free(dir_name);
		    flock(fd,LOCK_UN);
		    close(fd);
		    free(lock_filename);
		    free(mod_filename);
		    return NULL;
		}

		free(temp_dir_name);

		/* load under the lock, so that we can make sure that
		   the module we linked is loadable before letting the
		   rest of the world touch it. */
		ret=do_loading(mod_filename);
		if (ret==NULL) {
		    /* if loading fails, remove the whole thing (key
		       and directory) from the cache, so that if this
		       failure is caused by a temporal error in code
		       generation, the next attempt will be given a
		       chance to succeed. */
		    gpc_unlink_free(key_name);
		    gpc_rmdashrf_free(dir_name);
		    flock(fd,LOCK_UN);
		    close(fd);
		    free(lock_filename);
		    free(mod_filename);
		    return NULL;
		}
		
		free(key_name);
		free(mod_filename);
		free(lock_filename);
		free(dir_name);
		flock(fd,LOCK_UN);
		close(fd);
		return ret;
	    }
	    
	    free(dir_name);
	} else {
	    /* this is when we find the module on the first search
	       under a shared lock. */

	    mod_filename=tsf_asprintf("%s/dyncc.out",
				      dir_name);
	    if (mod_filename==NULL) {
		tsf_set_errno("Could not tsf_asprintf()");
		flock(fd,LOCK_UN);
		close(fd);
		free(lock_filename);
		free(dir_name);
		return NULL;
	    }
	    
	    free(dir_name);
	}

	ret=do_loading(mod_filename);
	
	free(lock_filename);
	free(mod_filename);
	flock(fd,LOCK_UN);
	close(fd);
    } else {
	dir_name=gpc_mkdtemp(TSF_TMP_DIRECTORY,"gpc_dyncc");
	
	if (dir_name==NULL) {
	    tsf_set_errno("Could not tsf_mkdtemp() in gpc_dyncc_from_intable()");
	    return NULL;
	}

	mod_filename=tsf_asprintf("%s/dyncc.out",
				  dir_name);
	if (mod_filename==NULL) {
	    tsf_set_errno("Could not tsf_asprintf()");
	    gpc_rmdir_free(dir_name);
	    return NULL;
	}

	if (!do_code_gen(intable,
			 dir_name,
			 mod_filename,
			 tsf_true)) {
	    gpc_unlink_free(mod_filename);
	    gpc_rmdir_free(dir_name);
	    return NULL;
	}
	
	ret=do_loading(mod_filename);
	
	gpc_unlink_free(mod_filename);
	gpc_rmdir_free(dir_name);
    }

    return ret;
}

void gpc_dyncc_destroy(gpc_dyncc_t *dyncc) {
#ifdef USE_DLFCN
    dlclose(dyncc->library);
#endif
#ifdef USE_NS_MODULE
    NSUnLinkModule(dyncc->library,0);
#endif
    free(dyncc);
}

uintptr_t gpc_dyncc_run(gpc_dyncc_t *prog,
                        void *region,
                        gpc_cell_t *args) {
     return ((uintptr_t (*)(void *,gpc_cell_t*))prog->symbol)(region,args);
}

#else

tsf_bool_t gpc_dyncc_supported() {
    return tsf_false;
}

gpc_dyncc_t *gpc_dyncc_from_intable(gpc_intable_t *intable,
				    const char *cache_dir) {
    tsf_set_error(TSF_E_NOT_SUPPORTED,
                  "Dynamic C compilation support was not compiled into "
                  "the TSF package.  This is because no suitable "
                  "dynamic library loading facility was found.");
    return NULL;
}

void gpc_dyncc_destroy(gpc_dyncc_t *dyncc) {
    tsf_abort("Call to gpc_dyncc_destroy() in a program that should "
              "never have been able to create a gpc_dyncc_t, because "
              "dynamic C compilation is not supported.");
}

uintptr_t gpc_dyncc_run(gpc_dyncc_t *prog,
                        void *region,
                        gpc_cell_t *args) {
    tsf_abort("Call to gpc_threaded_run() in a program that should "
              "never have been able to create a gpc_dyncc_t, because "
              "dynamic C compilation is not supported.");
}

#endif



