/*
 * $Header: /Users/pizlo/Repository/tsf/gpc_program.c,v 1.1.1.1 2007/05/02 23:31:05 pizlo Exp $
 *
 * gpc_program.c -- encapsulates the whole GPC execution engine maze.
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "gpc_internal.h"

char const * const TSF_CD_NO_CACHE = (char const * const) 0;
char const * const TSF_CD_DEFAULT  = (char const * const) 1;

gpc_program_t *gpc_program_from_proto(gpc_proto_t *proto,
                                      tsf_engine_t conf,
				      const char *cache_dir) {
    gpc_intable_t *intable;
    gpc_program_t *ret;
    
    intable=gpc_intable_from_proto(proto);
    
    if (intable==NULL) {
        return NULL;
    }
    
    ret=gpc_program_from_intable(intable,conf,cache_dir);

    gpc_intable_destroy(intable);
    
    return ret;
}

gpc_program_t *gpc_program_from_intable(gpc_intable_t *intable,
                                        tsf_engine_t conf,
					const char *cache_dir) {
    gpc_program_t *ret;
    const char *env_def;
    
    switch (conf) {
        case TSF_EN_DEFAULT:
            env_def=getenv("TSF_ENGINE_CONFIG");
            if (env_def==NULL || !strcmp(env_def,"DEFAULT")) {
                conf=TSF_EN_GPINT;
            } else if (!strcmp(env_def,"GPINT")) {
                conf=TSF_EN_GPINT;
            } else if (!strcmp(env_def,"DYNCC")) {
                conf=TSF_EN_DYNCC;
            } else {
                tsf_set_error(TSF_E_NOT_SUPPORTED,
                              "\"%s\" is not a supported engine.  To "
                              "resolve this problem, reset the "
                              "TSF_ENGINE_CONFIG environment variable "
                              "to \"GPINT\", \"DYNCC\", or "
                              "\"DEFAULT\".",
                              env_def);
                return NULL;
            }
            break;
        case TSF_EN_NOT_SUPPORTED:
            return NULL;
        default:
            break;
    }
    
    ret=malloc(sizeof(gpc_program_t));
    if (ret==NULL) {
        tsf_set_errno("Could not malloc gpc_program_t");
        return NULL;
    }
    
    ret->conf=conf;
    
    switch (conf) {
        case TSF_EN_GPINT:
            if (gpc_threaded_supported()) {
                ret->payload=gpc_threaded_from_intable(intable);
                if (ret->payload==NULL) {
                    free(ret);
                    return NULL;
                }
                ret->runner=(gpc_program_runner_t)gpc_threaded_run;
                ret->destroyer=(gpc_program_destroyer_t)gpc_threaded_destroy;
            } else {
                ret->payload=gpc_intable_clone(intable);
                if (ret->payload==NULL) {
                    free(ret);
                    return NULL;
                }
                ret->runner=(gpc_program_runner_t)gpc_intable_run;
                ret->destroyer=(gpc_program_destroyer_t)gpc_intable_destroy;
            }
            break;
        case TSF_EN_DYNCC:
	    /* why handle TSF_CD_DEFAULT here?
	       because in the future, other engines may use the cache
	       directory facility, and we want the implementation to
	       be uniform.  hence we put it here. */
	    if (cache_dir==TSF_CD_DEFAULT) {
		const char *env;
		env=getenv("TSF_USE_CACHE_DIR");
		if (env!=NULL && !strcasecmp(env,"yes")) {
		    env=getenv("TSF_CACHE_DIR");
		    if (env!=NULL) {
			cache_dir=env;
		    } else {
			cache_dir=TSF_CD_NO_CACHE;
		    }
		} else {
		    cache_dir=TSF_CD_NO_CACHE;
		}
	    }
            ret->payload=gpc_dyncc_from_intable(intable,cache_dir);
            if (ret->payload==NULL) {
                free(ret);
                return NULL;
            }
            ret->runner=(gpc_program_runner_t)gpc_dyncc_run;
            ret->destroyer=(gpc_program_destroyer_t)gpc_dyncc_destroy;
            break;
        default:
            tsf_set_error(TSF_E_NOT_SUPPORTED,
                          "%d is not a supported engine",
                          conf);
            free(ret);
            return NULL;
    }
    
    return ret;
}

void gpc_program_destroy(gpc_program_t *prog) {
    (prog->destroyer)(prog->payload);
    free(prog);
}

uintptr_t gpc_program_run(gpc_program_t *prog,
                          void *region,
                          gpc_cell_t *args) {
    return (prog->runner)(prog->payload,region,args);
}




