/*
 * tsf.h -- typed stream format header
 * by Filip Pizlo, 2003, 2004, 2005, 2006
 *
 * Stuff to know:
 *
 * Memory management conventions: unless otherwise stated, all objects
 * allocated by functions that you call become your responsibility after
 * a successful return.  also, unless otherwise stated, passing an
 * object as a first argument to a function does not clear your ownership
 * of that object.  and unless otherwise stated, objects passed as anything
 * but the first argument become the responsibility of that function,
 * regardless of whether or not that function returns successfully.
 *
 * Error management conventions: most functions return pointers or booleans.
 * in the case of pointers, if the pointer returned is NULL, then you must
 * check for errors.  in the case of booleans, you must check if the
 * return value is tsf_false.
 *
 * Error propagation: those functions that claim to be capable of error
 * propagation will return immediately as if with an error without actually
 * setting the error if one of their object arguments is NULL.  This allows
 * you to do:
 * if (!tsf_do_operation(object1,obj2, ..., tsf_do_op_that_fails(),...)) {
 *      // process error.
 * }
 * ... and safely expect that this will correctly catch errors from
 * tsf_do_op_that_fails() in addition to catching errors from
 * tsf_do_operation().
 *
 * Code generation/native code support/struct mapping: this functionality is
 * captured under the 'tsf_native' namespace, but it does not actually have
 * any struct or class or object of its own.  Instead, this functionality's
 * state is spread across the rest of the system.  Fields that have the sole
 * purpose of supporting code generation or struct mapping are marked as
 * such with a comment.
 */

#ifndef FP_TSF_H
#define FP_TSF_H

#include "tsf_config.h"
#include "tsf_st.h"
#include "tsf_region.h"
#include "tsf_types.h"
#include "tsf_zip_abstract.h"

#ifdef TSF_HAVE_PTHREAD
#include <pthread.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#ifdef __cplusplus
extern "C" {
#endif

#if 0
/* hack to prevent emacs from being confused */
}
#endif

/* macro you'll find useful */
#define tsf_offsetof(type,element) ((size_t)(&(((type*)NULL)->element)))

/* any errors in this enumeration must also be placed in tsf_error.c's
 * init_error_array() function. */
enum tsf_error {
    TSF_E_ERRNO,            /* system error; examine errno value returned
                             * from tsf_get_system_errno() */
    TSF_E_BDB_ERROR,        /* Berkeley DB error; examine the error value
                             * returned from tsf_get_bdb_error_code() */
    TSF_E_ZLIB_ERROR,       /* ZLib error; examine the error code returned
                             * from tsf_get_zlib_error_code() and also the
                             * error message returned from
                             * tsf_get_zlib_error_msg() */
    TSF_E_LIBBZIP2_ERROR,   /* libbzip2 error; examine the error code returned
                             * from tsf_get_libbzip2_error_code() and also the
                             * error message returned from
                             * tsf_get_libbzip2_error_msg().  note that the
                             * latter function returns textual messages
                             * selected by TSF and not by libbzip2.  these
                             * messages were written by me based on the
                             * libbzip2 manual. */
    TSF_E_INVALID_ARG,      /* invalid argument */
    TSF_E_BAD_STATE,        /* invalid state for the given operation */
    TSF_E_STR_OVERFLOW,     /* string overflow.  strings used as element names
                             * in structs and choices cannot be longer than
                             * TSF_STR_SIZE (including null byte).  comment
                             * strings also have some limit. */
    TSF_E_PARSE_ERROR,      /* parse error */
    TSF_E_UNEXPECTED_TYPE,  /* did not parse the type that I expected to
                             * parse.  this may happen if you're asking the
                             * serial in man to re-read some data that it's
                             * already read.  you can do this by using its
                             * mark and seek mechanism. */
    TSF_E_ELEMENT_NULL,     /* element in container is null */
    TSF_E_NAME_COLLISION,   /* name collision - the name is already used! */
    TSF_E_EOF,              /* end of file or end of stream */
    TSF_E_UNEXPECTED_EOF,   /* unexpected end of file or end of stream.  when
                             * returned from a reader, this error means that
                             * the EOF happened before any data was read. */
    TSF_E_ERRONEOUS_EOF,    /* erroneous end of file or end of stream.  when
                             * returned from a reader, this error means that
                             * the EOF happened after some data was already
                             * read.  to everyone else, this error should
                             * have the same exact meaning as
                             * TSF_E_UNEXPECTED_EOF */
    TSF_E_NO_EOF,           /* expected an EOF but didn't get one. */
    TSF_E_BAD_TYPE,         /* parameter has the wrong tsf_type - thrown
                             * when the type for a typed object (such as
                             * a reflect object) is not what is expected. */
    TSF_E_BAD_TYPE_KIND,    /* parameter has the wrong tsf_type_kind -
                             * thrown when you pass the wrong type object
                             * into a function. */
    TSF_E_ELE_NOT_FOUND,    /* element not found */
    TSF_E_NO_STRUCT_MAP,    /* the given type lacks a struct mapping */
    TSF_E_BAD_STRUCT_MAP,   /* the structure mapping offset or size is
                             * invalid because it does not agree with the
                             * offsets and/or sizes of the other elements */
    TSF_E_TYPE_MISMATCH,    /* type mismatch */
    TSF_E_BAD_CHOICE_VALUE, /* bad choice value.  note that if a bad choice
                             * value appears while parsing, you will get a
                             * TSF_E_PARSE_ERROR instead.  also, there will
                             * be times when you don't get this error even
                             * though you should have.  so be happy when
                             * you see it and don't cry to me about it if you
                             * don't. */
    TSF_E_COMPARE_FAIL,     /* a comparison failed. */
    TSF_E_COMPILE_ERROR,    /* compilation error in dyncc */
    TSF_E_DEATH_BY_SIGNAL,  /* child process died because it received a
                             * signal (again, this'll happen because of
                             * dyncc). */
    TSF_E_DLOADING_ERROR,   /* dynamic loading error in dyncc */
    TSF_E_BAD_JUMP,         /* jump to bad label in GPC prototype */
    TSF_E_SEMANTIC,         /* violation of GPC semantics */
    TSF_E_CAS_FAIL,         /* a compare-and-swap-like operation failed. */
    TSF_E_NOT_ALLOWED,      /* you get this error if you had specified that
                             * a certain action was not allowed, but for
                             * whatever reason TSF would have had to perform
                             * that action. */
    TSF_E_NOT_IMPLEMENTED,  /* not implemented.  this differs from not
                             * supported in that something that is 'not
                             * implemented' is actually advertised as being
                             * supported (and implemented), but I just
                             * haven't moved my ass to implement it.  you
                             * should only see 'not implemented' errors in
                             * development versions. */
    TSF_E_NOT_SUPPORTED,    /* not supported.  this differs from not
                             * implemented in that 'not supported' refers
                             * to functionality that is present but cannot
                             * be used due to some limitations of your
                             * system.  a good example: TSF has code to glue
                             * itself to Berkeley DB, but if Berkeley DB
                             * cannot be found on your system, TSF's
                             * Berkeley DB functions will return 'not
                             * supported'. */
    TSF_E_EXTERNAL,         /* some external error (like a condition in
			     * in a callback provided by someone else).
			     * the C++ binding uses this for
			     * std::exception objects caught in callbacks. */
    TSF_E_INTERNAL,         /* internal error - should never happen. */
    TSF_E_LAST              /* placeholder */
};

typedef enum tsf_error tsf_error_t;

/* standard string size within tsf */
#define TSF_STR_SIZE            128

/* max size for comment strings */
#define TSF_MAX_COMMENT_SIZE    4096

/* Instanceof rules:
 * -> VOID type.  everything is instanceof VOID, but VOID is instaceof
 *    VOID only.
 * -> INT types.  An INT type is instanceof any INT type that is
 *    composed of at least as many bits as it.  For example, INT8
 *    is instanceof INT8, INT16, INT32, and INT64.  INT types are
 *    not instanceof any UINT types.  INT types are instanceof
 *    those FP types that have a mantissa that is made up of at
 *    least as many bits as it.
 * -> UINT types.  A UINT type is instanceof any UINT type that is
 *    composed of at least as many bits as it.  Also, a UINT type
 *    is instanceof any INT type that has at least one more bits
 *    than it (so UINT16 will be instanceof INT32 and INT64, but
 *    not INT16).  UINT types are instanceof those FP types that
 *    have a mantissa that is made up of at least one more bits as
 *    it.
 * -> FP types.  An FP type is instanceof any FP type that has at
 *    least as large of a mantissa, and at least as large of an
 *    exponent.  So FLOAT is instanceof FLOAT and DOUBLE, while
 *    DOUBLE is only instanceof DOUBLE.
 * -> ARRAY.  if A and B are array types, then A is instanceof B
 *    if element_type(A) is instanceof element_type(B).
 * -> STRUCT.  one struct is instanceof another if it contains all
 *    of the mandatory elements of the other struct, and if each
 *    such element is instanceof its respective element in the other
 *    struct.  note that if a struct contains an optional field for
 *    another struct's mandatory field, this will violate the
 *    instanceof relation.
 * -> BIT.  a BIT is only instanceof BIT.  the justification is that
 *    despite the name, which would seem to imply that a BIT is a
 *    type of integer, a BIT is intended to be used in such a
 *    fundamentally different way that it would not make sense for
 *    a BIT to be instanceof any integer.
 * -> CHOICE.  if A and B are choice types, then A is instanceof B
 *    if for those element names that A and B have in common, A's
 *    element is instanceof B's element.
 * -> STRING.  a STRING is only instanceof STRING.
 * -> ANY.  a ANY is only instanceof ANY.  this is of
 *    course quite ironic, since ANY types are our main source of
 *    polymorhpism. :-)
 *
 * Note that if you change this enumeration, you must change some of the
 * #define's further down in this file.
 */
enum tsf_type_kind {
    TSF_TK_VOID     = 0,
    
    /* ints and floats are primitives.  bits are not. */
    
    TSF_TK_INT8     = 1,
    TSF_TK_UINT8    = 2,
    TSF_TK_INT16    = 3,
    TSF_TK_UINT16   = 4,
    TSF_TK_INT32    = 5,
    TSF_TK_UINT32   = 6,
    TSF_TK_INT64    = 7,
    TSF_TK_UINT64   = 8,
    /* reserve 9-19 for bizarre integer types. */

    TSF_TK_FLOAT    = 20,
    TSF_TK_DOUBLE   = 21,
    /* reserve 22-39 for bizarre float types.  recommend that if
     * we ever have a customizable float, it should be 30. */

    /* FIXME: should be able to convert between bit and numbers */
    TSF_TK_BIT      = 40,
    
    TSF_TK_STRING   = 45,

    TSF_TK_ARRAY    = 50,
    TSF_TK_STRUCT   = 51,
    TSF_TK_CHOICE   = 52,
    
    TSF_TK_ANY      = 60
};

typedef uint8_t tsf_type_kind_t;

struct tsf_type;
typedef struct tsf_type tsf_type_t;

struct tsf_named_type;
typedef struct tsf_named_type tsf_named_type_t;

/* ordered table of types keyed by int and string */
struct tsf_type_table {
    tsf_named_type_t **elements;
    uint32_t num_elements;
    
    tsf_st_table *ele_table;
};

typedef struct tsf_type_table tsf_type_table_t;

/* shared table of types keyed only by int */
struct tsf_type_in_map {
    void *region;           /* NULL if not allocated in a region */
    tsf_type_t **elements;
    uint32_t num_elements;
    uint32_t capacity;
};

typedef struct tsf_type_in_map tsf_type_in_map_t;

typedef tsf_st_table tsf_type_out_map_t;

struct tsf_ra_type_info;
struct tsf_ra_tc_node;
struct tsf_ra_type_man;

typedef struct tsf_ra_type_info tsf_ra_type_info_t;
typedef struct tsf_ra_tc_node tsf_ra_tc_node_t;
typedef struct tsf_ra_type_man tsf_ra_type_man_t;

struct tsf_ra_type_info {
    uint32_t type_code;
    uint64_t ref_count;
};

struct tsf_ra_tc_node {
    uint32_t type_code;
    tsf_ra_tc_node_t *next;
};

struct tsf_ra_type_man {
    tsf_ra_tc_node_t *type_codes;
    tsf_st_table *out_map;
    tsf_st_table *in_map;
};

struct tsf_type {
    tsf_type_kind_t kind_code;
    
    /* true if this is an ANY or if is a container that contains an ANY. */
    tsf_bool_t is_dynamic;
    
    union {
        /* additional data for array */
        struct {
            tsf_type_t *element_type;
        } a;
        
        /* additional data for struct */
        struct {
            /* struct mapping support - need to know the size of
             * the struct in memory.  this is NOT the sum of the
             * sizes of the elements, because that doesn't take
             * into account padding or the possibility of additional
             * fields in that struct that the user isn't telling us
             * about.  HOWEVER! we can be sure that this size must
             * be greater than or equal to the sum of the sizes of
             * the elements. */
            tsf_bool_t has_size;
            size_t size;

	    void (*constructor)(void *);
	    void (*destructor)(void *);

            tsf_type_table_t *tt;
        } s;
        
        /* additional data for choice */
        struct {
            tsf_type_table_t *tt;
            
            /* struct mapping support */
            tsf_bool_t has_mapping;
	    
	    tsf_bool_t in_place;    /* use union? */

            size_t value_offset;    /* offset to the value field */
            size_t data_offset;     /* offset to the union if in_place
				       is true, or to the pointer if
				       it is false. */
            size_t size;            /* size of the choice structure */
        } h;
    } u;
    
    /* precomputed hash code.  since each tsf_type object is immutable
     * (that's why all functions that 'modify' a tsf_type actually take
     * a pointer to a pointer to a tsf_type - so that instead of modifying
     * the actual tsf_type, it allocates a new one and changes your
     * pointer), the hash code can safely be precomputed when the
     * tsf_type is created. */
    int hash_code;
    
    /* lazily computed type size in bytes.  if it hasn't been computed yet,
     * its value will be 0. */
    volatile uint32_t byte_size;
    
    /* holder for an optional comment supplied by the user.  only
     * commentable types can hold comments (use tsf_type_kind_is_commentable()
     * to find out if it is commentable).  the comment is a UTF-8 string
     * supplied by the user via a function call (see tsf_type_set_comment())
     * or by using javadoc comments in definitions passed to tsf_define. */
    char *comment;
    
    /* only used for non-singletons */
#if TSF_HAVE_PTHREAD
    pthread_mutex_t ref_count_lock;
#endif
    volatile uint32_t ref_count;
};

struct tsf_named_type {
    char *name;
    tsf_type_t *type;
    uint32_t index;
    
    /* for struct entries - struct entries may be marked optional.  note
     * that if we have structure A, B, and C, so that A does not have
     * field F, B has an optional field F, and C has field F, then, assuming
     * the structs are otherwise equal, the instanceof 'matrix' goes as
     * follows:
     *
     *   A B C
     * A Y Y N
     * B Y Y N
     * C Y Y Y
     * 
     * In other words, a structure with an optional F is not instanceof a
     * structure with a mandatory F. */
    tsf_bool_t optional;
    
    /* struct mapping support - need to know the offset of this
     * element within its parent struct.
     *
     * Note that this does not apply to choice types, where
     * all elements have the same offset. */
    tsf_bool_t has_offset;
    size_t offset;
    
    /* holder for an optional comment supplied by the user either through
     * a function call or through tsf_define. */
    char *comment;
};

struct tsf_buffer {
    void *region; /* the region in which we were allocated */
    tsf_type_t *type;
    tsf_type_in_map_t *types;
    tsf_bool_t keep;
    void *data;
    uint32_t size;
    
#if TSF_HAVE_PTHREAD
    pthread_mutex_t ref_count_lock;
#endif
    volatile uint32_t ref_count;
};

typedef struct tsf_buffer tsf_buffer_t;

/* type used for arrays referenced from structs within our native
 * structure mapping.  you do not have to ever touch this type.  but
 * if you do struct mapping (which surely you will!), then your
 * arrays should resemble this here tsf_native_array. */
struct tsf_native_array {
    void *data;
    uint32_t len;
};

typedef struct tsf_native_array tsf_native_array_t;

/* types used internally for primitive arrays */
#define TSF_DEFINE_NATIVE_ARRAY(type,suffix) \
    struct tsf_native_##type##_array {\
        type##suffix *data;\
        uint32_t len;\
    };\
    \
    typedef struct tsf_native_##type##_array tsf_native_##type##_array_t

TSF_DEFINE_NATIVE_ARRAY(int8,_t);
TSF_DEFINE_NATIVE_ARRAY(int16,_t);
TSF_DEFINE_NATIVE_ARRAY(int32,_t);
TSF_DEFINE_NATIVE_ARRAY(int64,_t);
TSF_DEFINE_NATIVE_ARRAY(uint8,_t);
TSF_DEFINE_NATIVE_ARRAY(uint16,_t);
TSF_DEFINE_NATIVE_ARRAY(uint32,_t);
TSF_DEFINE_NATIVE_ARRAY(uint64,_t);
TSF_DEFINE_NATIVE_ARRAY(float,);
TSF_DEFINE_NATIVE_ARRAY(double,);

/* special type used for arrays of unknown type */
struct tsf_native_void_array {
    uint32_t len;
};

typedef struct tsf_native_void_array tsf_native_void_array_t;

/* special type used for arrays of bits referenced from structs
 * within our native structure mapping. */
struct tsf_native_bitvector {
    uint8_t *bits;  /* array of length (num_bits+7)>>3 */
    uint32_t num_bits;
};

typedef struct tsf_native_bitvector tsf_native_bitvector_t;

/* type of the choice itself within choice types. */
typedef uint32_t tsf_choice_t;

struct tsf_reflect;
typedef struct tsf_reflect tsf_reflect_t;

struct tsf_reflect {
    tsf_type_t *type;
    union {
        /* container, used for both struct and array.  note that it
         * is used differently for struct and for array. */
        struct {
            tsf_reflect_t **array;
            uint32_t size;
        } c;
        
        /* choice data. */
        struct {
            uint32_t choice;    /* 'unknown' if equal to UINT32_MAX */
            tsf_reflect_t *data;
        } h;
        
        /* string data. */
        char *str;
        
        /* primitives */
        int8_t   p_int8_t;
        uint8_t  p_uint8_t;
        int16_t  p_int16_t;
        uint16_t p_uint16_t;
        int32_t  p_int32_t;
        uint32_t p_uint32_t;
        int64_t  p_int64_t;
        uint64_t p_uint64_t;
        float    p_float;
        double   p_double;
        
        /* bit */
        tsf_bool_t bit;
        
        /* for the any type */
        tsf_buffer_t *any;
    } u;
};

struct tsf_genrtr {
    tsf_type_t *type;
    gpc_program_t *generator;
};

struct tsf_parser {
    tsf_type_t *expected_type;
    tsf_engine_t conf;
    char *cache_dir;
    tsf_st_table *parsers;
    
#if TSF_HAVE_PTHREAD
    pthread_mutex_t lock;
#endif
};

struct tsf_copier {
    gpc_program_t *copier;
};

typedef struct tsf_genrtr tsf_genrtr_t;
typedef struct tsf_parser tsf_parser_t;
typedef struct tsf_copier tsf_copier_t;

/* function callback used for reading.  return tsf_true on success and
 * tsf_false on error.  on error, use tsf_set_error() and tsf_set_errno()
 * to indicate what error occurred (this part is not optional - you
 * cannot return tsf_false without first calling tsf_set_error() or
 * tsf_set_errno()).  an all-or-nothing read is expected, so if you're
 * implementing this as a read on a socket or pipe, you may have to
 * loop a couple of times. */
typedef tsf_bool_t (*tsf_reader_t)(void *arg,
                                   void *buf,
                                   uint32_t len);

/* function callback used for write.  return tsf_true on success and
 * tsf_false on error.  on error, use tsf_set_error() and tsf_set_errno()
 * to indicate what error occurred (this part is not optional - you
 * cannot return tsf_false without first calling tsf_set_error() or
 * tsf_set_errno()).  an all-or-nothing write is expected, so if you're
 * implementing this as a write on a socket or pipe, you may have to
 * loop a couple of times. */
typedef tsf_bool_t (*tsf_writer_t)(void *arg,
                                   void *buf,
                                   uint32_t len);

/* function callback used for reading.  returns non-negative count of
 * the number of bytes read on success and -1 on error.  on error, use
 * tsf_set_error() and friends.  the caller expects read-as-much-as-
 * you-can semantics, so this function should never block if some data
 * is already available. */
typedef int32_t (*tsf_partial_reader_t)(void *arg,
                                        void *buf,
                                        uint32_t len);

struct tsf_limits {
    /* maximum number of types accepted in a stream file in between reset
     * types calls. */
    uint32_t max_types;
    
    /* byte size limits */
    
    /* maximum size in bytes of a serialized type */
    uint32_t max_type_size;
    
    /* maximum size that all types read during a session can have put
     * together. */
    uint32_t max_total_type_size;
    
    /* maximum size in bytes of a buffer */
    uint32_t max_buf_size;
    
    /* container depth limit */
    uint32_t depth_limit;
    
    /* whether or not to allow comments */
    tsf_bool_t allow_comments;
    
    /* allowing individual types.  access this as a regular bitvector. */
    uint8_t allow_type_kind[32];
    
    /* type-specific limits */
    
    /* maximum array length */
    uint32_t max_array_length;  /* unimplemented */
    
    /* maximum struct size, in number of elements */
    uint32_t max_struct_size;
    
    /* maximum choice size, in number of elements */
    uint32_t max_choice_size;
    
    /* maximum string size */
    uint32_t max_string_size;   /* unimplemented */
};

typedef struct tsf_limits tsf_limits_t;

struct tsf_serial_in_man;
struct tsf_serial_out_man;

typedef struct tsf_serial_in_man tsf_serial_in_man_t;
typedef struct tsf_serial_out_man tsf_serial_out_man_t;

/* callback that gets called by the input manager when a new type comes
 * in. */
typedef void (*tsf_type_cback_t)(tsf_serial_in_man_t *in_man,
                                 tsf_type_t *type,
                                 void *arg);

typedef struct {
    tsf_type_cback_t cback;
    void *arg;
} tsf_serial_in_man_cback_t;

typedef uint32_t tsf_serial_in_man_state_t;

struct tsf_serial_in_man {
    tsf_type_in_map_t *types;
    tsf_serial_in_man_state_t state;
    
    uint32_t types_byte_size;
    
    tsf_serial_in_man_cback_t *cbacks;
    uint32_t num_cbacks;
    
    tsf_reader_t reader;
    void *arg;
    
    tsf_limits_t *limits;
    
#if TSF_HAVE_PTHREAD
    pthread_mutex_t lock;
#endif
};

struct tsf_serial_out_man {
    tsf_type_out_map_t *types;
    
    tsf_writer_t writer;
    void *arg;
    
#if TSF_HAVE_PTHREAD
    pthread_mutex_t lock;
#endif
};

struct tsf_buf_rdr {
    int fd;
    tsf_bool_t keep_fd;
    uint8_t *buf;
    uint32_t size;
    uint32_t puti;
    uint32_t geti;
};

struct tsf_buf_wtr {
    int fd;
    tsf_bool_t keep_fd;
    uint8_t *buf;
    uint32_t size;
    uint32_t pos;
};

typedef struct tsf_buf_rdr tsf_buf_rdr_t;
typedef struct tsf_buf_wtr tsf_buf_wtr_t;

/* the zlib rdr/wtr can also do bzip2. */

struct tsf_zip_wtr_attr {
    tsf_zip_mode_t mode;
    
    uint32_t buf_size;
    
    union {
        struct {
            tsf_bool_t advanced_init;
            
            int level;
            
            int method;
            int windowBits;
            int memLevel;
            int strategy;
        } z;
        struct {
            int blockSize100k;
            int verbosity;
            int workFactor;
        } b;
    } u;
};

struct tsf_zip_rdr_attr {
    struct {
        tsf_bool_t allow;
        
        uint32_t buf_size;
        tsf_bool_t advanced_init;
        int windowBits;
    } z;
    struct {
        tsf_bool_t allow;
        
        uint32_t buf_size;
        int verbosity;
        int small;
    } b;
};

typedef struct tsf_zip_wtr_attr tsf_zip_wtr_attr_t;
typedef struct tsf_zip_rdr_attr tsf_zip_rdr_attr_t;

struct tsf_zip_rdr {
    tsf_zip_mode_t mode;
    
    tsf_partial_reader_t reader;
    void *reader_arg;
    
    void *stream;   /* can be either of type z_stream or bz_stream */
    
    tsf_zip_abstract_t abstract;
    
    tsf_bool_t decomp_eof;  /* true if the decompression algorithm has
                             * reached EOF. */
    
    uint8_t *buf;
    uint32_t buf_size;
};

struct tsf_zip_wtr {
    tsf_zip_mode_t mode;
    
    tsf_writer_t writer;
    void *writer_arg;
    
    void *stream;   /* can be either of type z_stream of bz_stream */
    
    tsf_zip_abstract_t abstract;
    
    uint8_t *buf;
    uint32_t buf_size;
};

typedef struct tsf_zip_rdr tsf_zip_rdr_t;
typedef struct tsf_zip_wtr tsf_zip_wtr_t;

struct tsf_adpt_rdr {
    tsf_zip_mode_t mode;
    
    tsf_partial_reader_t reader;
    void *reader_arg;
    
    void *stream;   /* can be either of type z_stream or bz_stream */
    
    tsf_zip_abstract_t abstract;
    
    tsf_bool_t expect_switch;
    
    uint8_t *buf;
    uint32_t puti,geti; /* same as in buffered reader */
    uint32_t buf_size;

    tsf_zip_rdr_attr_t attr;
    uint32_t nozip_buf_size;
};

typedef struct tsf_adpt_rdr tsf_adpt_rdr_t;

/* you're responsible for allocating these guys yourself.  but that does
 * not mean that you can edit them!  treat this as an opaque!  also note
 * that the state is effectively a pointer, so this mark is undefined if
 * serialized... so you cannot use these things to build an external
 * index.  sorry. */
struct tsf_stream_file_mark {
    tsf_off_t offset;
    tsf_serial_in_man_state_t state;
};

typedef struct tsf_stream_file_mark tsf_stream_file_mark_t;

struct tsf_stream_file_input {
    int fd;
    tsf_bool_t keep_fd;
    
    tsf_serial_in_man_t *in_man;
    
    tsf_bool_t use_adpt;
    union {
        tsf_buf_rdr_t *buf;
        tsf_adpt_rdr_t *adpt;
        void *ptr;
    } reader;
};

struct tsf_stream_file_output {
    int fd;
    tsf_bool_t keep_fd;
    
    char *rename_to;
    char *rename_from;
    
    tsf_serial_out_man_t *out_man;
    
    tsf_bool_t use_zip;
    union {
        tsf_buf_wtr_t *buf;
        tsf_zip_wtr_t *zip;
        void *ptr;
    } writer;
    
    tsf_bool_t all_good;
};

typedef struct tsf_stream_file_input tsf_stream_file_input_t;
typedef struct tsf_stream_file_output tsf_stream_file_output_t;

struct tsf_bdb_file {
    void *db;
    
    tsf_bool_t update;

    tsf_ra_type_man_t *types;
};

struct tsf_bdb_cursor {
    void *dbc;
};

typedef struct tsf_bdb_file tsf_bdb_file_t;
typedef struct tsf_bdb_cursor tsf_bdb_cursor_t;

/* get an error code.  this is the error code from the last function that
 * failed in the current thread.  functions in the TSF API always tell you
 * whether or not they failed, usually by returning NULL (if the function
 * returns pointers), tsf_false (if the function returns tsf_bool_t), or -1
 * (if the function returns some sort of integer).  calls to this function
 * made when no error has occurred result in undefined behavior (read:
 * segfault). */
tsf_error_t     tsf_get_error_code(void);

/* given an error code, tells you what the default message is.  the string
 * returned here is truly a constant, so you can safely refer to it
 * indefinitely.  note that if you want a more descriptive message for an
 * error that just occurred, use tsf_get_error().  tsf_get_error() will
 * almost always return a string that contains more information than using
 * tsf_get_message_for_error_code(tsf_get_error_code()).  note that this
 * function does not check if code is a valid TSF error code; if it isn't,
 * bad things will happen. */
const char*     tsf_get_message_for_error_code(tsf_error_t code);

/* get the system errno associated with the current error.  the return
 * value is undefined if the last function that failed did not fail due to
 * a system error. */
int             tsf_get_system_errno(void);

/* get the BDB error code assiciated with the current error.  the return
 * value is undefined if the last function that failed did not fail due to
 * a BDB error. */
int             tsf_get_bdb_error_code(void);

/* get the ZLib error code associated with the current error.  the return
 * value is undefined if the last function that failed did not fail due to
 * a ZLib error. */
int             tsf_get_zlib_error_code(void);

/* get the ZLib error message associated with the current error.  the return
 * value is undefined if the last function that failed did not fail due to
 * a ZLib error.  (NOTE: if a ZLib error actually did happen, this will
 * not be NULL.)*/
const char*     tsf_get_zlib_error_msg(void);

/* get the libbzip2 error code associated with the current error.  the return
 * value is undefined if the last function that failed did not fail due to
 * a libbzip2 error. */
int             tsf_get_libbzip2_error_code(void);

/* get the libbzip2 error message associated with the current error.  the
 * return value is undefined if the last function that failed did not fail
 * due to a libbzip2 error.*/
const char*     tsf_get_libbzip2_error_msg(void);

/* get the signal that killed the child process (only for
 * TSF_E_DEATH_BY_SIGNAL).  the return value is undefined if the last
 * function that failed did not fail due to TSF_E_DEATH_BY_SIGNAL. */
int             tsf_get_deadly_signal(void);

/* get a description of the error.  the string that is returned
 * is not owned by you; it is only valid until the next call to
 * any tsf function other than tsf_get_error_code(), tsf_get_error(),
 * and the other error-getting functions above.  (If the library is
 * compiled with thread-safety then the previous statement refers to
 * the next call by the same thread, rather than the next call
 * globally.)  calls to this function made when no error has
 * occurred (within this thread) result in undefined behavior
 * (quite possibly and very likely a segfault inside tsf_get_error()). */
const char*     tsf_get_error(void);

/* get just the message for the current error code.  this will always
 * be the prefix of what tsf_get_error() returns. */
#define         tsf_get_error_prefix() \
    (tsf_get_message_for_error_code(tsf_get_error_code()))

/* get just the message that is acquired from the source of the error.
 * for errno, this is strerror(errno); for the BDB error, this is
 * db_strerror(...), and so on.  this will be NULL for errors in which
 * this does not apply.  if the error is set using the ordinary functions
 * in this header file, this string will fall somewhere in the string
 * returned by tsf_get_error() - but to be safe you should not rely on
 * this fact. */
const char*     tsf_get_error_infix(void);

/* get just the message that was supplied by the user.  this will be
 * NULL if the user did not supply anything.  if non-null, this will
 * always be the suffix of what tsf_get_error() returns. */
const char*     tsf_get_error_suffix(void);

/* specify that an error occurred.  format may be NULL, in which case
 * the only error string is the one that comes with the given error
 * code. */
void            tsf_set_error(tsf_error_t code,
                              const char *format,
                              ...);

/* specify that a system error ocurred.  if format is NULL, then the
 * error message contains the string "System error" and the strerror
 * for the given errno.  if format is non-NULL, then your text
 * gets appended at the end of that. */
void            tsf_set_specific_errno(int cur_errno,
                                       const char *format,...);

/* specify that a system error ocurred.  exactly like
 * tsf_set_specific_errno() except that the global errno is used
 * instead of the given errno. */
void            tsf_set_errno(const char *format,...);

/* specify that a BDB error ocurred.  if the error code is positive,
 * this function will behave exactly like tsf_set_errno().  if format
 * is NULL, then the error message contains the string "Berkeley DB
 * error" and the db_strerror for the given error code.  if format is
 * non-NULL, then your text gets appended at the end of that.
 *
 * Note that if BDB support is not compiled in, this function will
 * mean an immediate abort. */
void            tsf_set_bdb_error(int bdb_error,
                                  const char *format,...);

/* specify that some child process died with the given signal. */
void            tsf_set_deadly_signal(int deadly_signal,
                                      const char *format,...);

/* specify that a ZLib error occurred.  if format is NULL, then the
 * error message contains the string "ZLib error" and the error message
 * you provide.  If the format is not NULL, the text specified by the
 * format is appended to that.
 *
 * Note that if ZLib support is not compiled in, this function will
 * mean an immediate abort. */
void            tsf_set_zlib_error(int error_code,
                                   const char *error_msg,
                                   const char *format,
                                   ...);

/* specify that a libbzip2 error occurred.  if format is NULL, then the
 * error message contains the string "libbzip2 error" and an error message
 * that is appropriate to the error code.  If the format is not NULL, the
 * text specified by the format is appended to that.
 *
 * Note that if libbzip2 support is not compiled in, this function will
 * mean an immediate abort. */
void            tsf_set_libbzip2_error(int error_code,
                                       const char *format,
                                       ...);

/* allows you to specify that you're in code that should not be considered
 * for fault injection.  This is mainly for internal use, but can also be
 * of use if you find yourself using TSF's fault injection features. */
tsf_bool_t      tsf_set_in_safe_code(tsf_bool_t value);

/* query if you're in code that is 'safe';  only safe code will be
 * considered for fault injection. */
tsf_bool_t      tsf_in_safe_code(void);

/* returns the library's version string. */
const char*     tsf_version(void);

/* returns the library's major version number. */
int             tsf_version_major(void);

/* returns the library's minor version number. */
int             tsf_version_minor(void);

/* returns the library's patch version number. */
int             tsf_version_patch(void);

/* creates default limits */
tsf_limits_t*   tsf_limits_create(void);

/* destroys limits, but only if they're not NULL */
void            tsf_limits_destroy(tsf_limits_t *limits);

/* clones limits.  returns NULL if the type it's given is NULL.  this
 * counts as error propagation. */
tsf_limits_t*   tsf_limits_clone(tsf_limits_t *limits);

/* set the allow bit of the particular kind type */
static TSF_inline
void            tsf_limits_set_allow_type_kind(tsf_limits_t *limits,
                                               tsf_type_kind_t kind_code) {
    limits->allow_type_kind[kind_code>>3]|=(1<<(kind_code&7));
}

/* reset the allow bit of the particular kind type */
static TSF_inline
void            tsf_limits_reset_allow_type_kind(tsf_limits_t *limits,
                                                 tsf_type_kind_t kind_code) {
    limits->allow_type_kind[kind_code>>3]&=~(1<<(kind_code&7));
}

/* determine if the allow bit of the particular kind type is set.  always
 * returns tsf_true if limits is NULL. */
static TSF_inline
tsf_bool_t      tsf_limits_is_type_kind_allowed(tsf_limits_t *limits,
                                                tsf_type_kind_t kind_code) {
    if (limits==NULL) {
        return tsf_true;
    }
    return (limits->allow_type_kind[kind_code>>3]&(1<<(kind_code&7)))
           ?tsf_true:tsf_false;
}

/* given a kind code, gives you the textual tsf name for it. */
static TSF_inline
const char*     tsf_type_kind_tsf_name(tsf_type_kind_t kind_code) {
    switch (kind_code) {
        case TSF_TK_INT8:   return "TSF_TK_INT8";
        case TSF_TK_UINT8:  return "TSF_TK_UINT8";
        case TSF_TK_INT16:  return "TSF_TK_INT16";
        case TSF_TK_UINT16: return "TSF_TK_UINT16";
        case TSF_TK_INT32:  return "TSF_TK_INT32";
        case TSF_TK_UINT32: return "TSF_TK_UINT32";
        case TSF_TK_INT64:  return "TSF_TK_INT64";
        case TSF_TK_UINT64: return "TSF_TK_UINT64";
        case TSF_TK_FLOAT:  return "TSF_TK_FLOAT";
        case TSF_TK_DOUBLE: return "TSF_TK_DOUBLE";
        case TSF_TK_ARRAY:  return "TSF_TK_ARRAY";
        case TSF_TK_STRUCT: return "TSF_TK_STRUCT";
        case TSF_TK_BIT:    return "TSF_TK_BIT";
        case TSF_TK_CHOICE: return "TSF_TK_CHOICE";
        case TSF_TK_VOID:   return "TSF_TK_VOID";
        case TSF_TK_STRING: return "TSF_TK_STRING";
        case TSF_TK_ANY:    return "TSF_TK_ANY";
        default:            return "unknown";
    }
}

/* tells you if a kind code represents a primitive type */
static TSF_inline
tsf_bool_t      tsf_type_kind_is_primitive(tsf_type_kind_t kind_code) {
    return kind_code>=TSF_TK_INT8 && kind_code<=TSF_TK_DOUBLE;
}

/* given a primitive kind code, it returns to you the size in bytes. */
static TSF_inline
uint32_t        tsf_primitive_type_kind_size_of(tsf_type_kind_t kind_code) {
    switch (kind_code) {
        case TSF_TK_INT8:   return 1;
        case TSF_TK_UINT8:  return 1;
        case TSF_TK_INT16:  return 2;
        case TSF_TK_UINT16: return 2;
        case TSF_TK_INT32:  return 4;
        case TSF_TK_UINT32: return 4;
        case TSF_TK_INT64:  return 8;
        case TSF_TK_UINT64: return 8;
        case TSF_TK_FLOAT:  return 4;
        case TSF_TK_DOUBLE: return 8;
        default:            return 0;
    }
}

/* returns a range lower bound ID */
static TSF_inline
int32_t         tsf_primitive_type_kind_lower_bound(tsf_type_kind_t kind_code) {
    switch (kind_code) {
        case TSF_TK_INT8:   return -1;
        case TSF_TK_UINT8:  return 0;
        case TSF_TK_INT16:  return -3;
        case TSF_TK_UINT16: return 0;
        case TSF_TK_INT32:  return -5;
        case TSF_TK_UINT32: return 0;
        case TSF_TK_INT64:  return -7;
        case TSF_TK_UINT64: return 0;
        case TSF_TK_FLOAT:  return -9;
        case TSF_TK_DOUBLE: return -10;
        default:            return 0;
    }
}

/* returns a range upper bound ID */
static TSF_inline
int32_t         tsf_primitive_type_kind_upper_bound(tsf_type_kind_t kind_code) {
    switch (kind_code) {
        case TSF_TK_INT8:   return 1;
        case TSF_TK_UINT8:  return 2;
        case TSF_TK_INT16:  return 3;
        case TSF_TK_UINT16: return 4;
        case TSF_TK_INT32:  return 5;
        case TSF_TK_UINT32: return 6;
        case TSF_TK_INT64:  return 7;
        case TSF_TK_UINT64: return 8;
        case TSF_TK_FLOAT:  return 9;
        case TSF_TK_DOUBLE: return 10;
        default:            return 0;
    }
}

/* returns the precision of a primitive type */
static TSF_inline
int32_t         tsf_primitive_type_kind_precision(tsf_type_kind_t kind_code) {
    switch (kind_code) {
        case TSF_TK_INT8:   return 8;
        case TSF_TK_UINT8:  return 8;
        case TSF_TK_INT16:  return 16;
        case TSF_TK_UINT16: return 16;
        case TSF_TK_INT32:  return 32;
        case TSF_TK_UINT32: return 32;
        case TSF_TK_INT64:  return 64;
        case TSF_TK_UINT64: return 64;
        case TSF_TK_FLOAT:  return 23;
        case TSF_TK_DOUBLE: return 52;
        default:            return 0;
    }
}

/* given a primitive kind code, tells you if it is floating point. */
static TSF_inline
tsf_bool_t      tsf_primitive_type_kind_is_float(tsf_type_kind_t kind_code) {
    return kind_code==TSF_TK_FLOAT || kind_code==TSF_TK_DOUBLE;
}

/* same as above. */
#define         tsf_type_kind_is_float(kind_code) \
                    (tsf_primitive_type_kind_is_float(kind_code))

/* given a primitive kind code, tells you if it is floating point. */
static TSF_inline
tsf_bool_t      tsf_primitive_type_kind_is_int(tsf_type_kind_t kind_code) {
    return tsf_type_kind_is_primitive(kind_code) &&
           !tsf_type_kind_is_float(kind_code);
}

/* same as above. */
#define         tsf_type_kind_is_int(kind_code) \
                    (tsf_primitive_type_kind_is_int(kind_code))

/* given a primitive kind code, tells you if it is signed. */
static TSF_inline
tsf_bool_t      tsf_primitive_type_kind_is_signed(tsf_type_kind_t kind_code) {
    switch (kind_code) {
        case TSF_TK_INT8:
        case TSF_TK_INT16:
        case TSF_TK_INT32:
        case TSF_TK_INT64:  return tsf_true;
        default:            return tsf_primitive_type_kind_is_float(kind_code);
    }
}

/* same as above. */
#define         tsf_type_kind_is_signed(kind_code) \
                    (tsf_primitive_type_kind_is_signed(kind_code))

/* given a primitive kind code, gives you the textual name of the
 * C type. */
static TSF_inline
const char*     tsf_primitive_type_kind_c_name(tsf_type_kind_t kind_code) {
    switch (kind_code) {
        case TSF_TK_INT8:   return "int8_t";
        case TSF_TK_UINT8:  return "uint8_t";
        case TSF_TK_INT16:  return "int16_t";
        case TSF_TK_UINT16: return "uint16_t";
        case TSF_TK_INT32:  return "int32_t";
        case TSF_TK_UINT32: return "uint32_t";
        case TSF_TK_INT64:  return "int64_t";
        case TSF_TK_UINT64: return "uint64_t";
        case TSF_TK_FLOAT:  return "float";
        case TSF_TK_DOUBLE: return "double";
        default:            return "unknown";
    }
}

/* given a primitive kind code, gives you the textual name of the type
 * using lower case. */
static TSF_inline
const char*     tsf_primitive_type_kind_lc_name(tsf_type_kind_t kind_code) {
    switch (kind_code) {
        case TSF_TK_INT8:   return "int8";
        case TSF_TK_UINT8:  return "uint8";
        case TSF_TK_INT16:  return "int16";
        case TSF_TK_UINT16: return "uint16";
        case TSF_TK_INT32:  return "int32";
        case TSF_TK_UINT32: return "uint32";
        case TSF_TK_INT64:  return "int64";
        case TSF_TK_UINT64: return "uint64";
        case TSF_TK_FLOAT:  return "float";
        case TSF_TK_DOUBLE: return "double";
        default:            return "unknown";
    }
}

/* given a kind code, gives you the textual lower-case name */
static TSF_inline
const char*     tsf_type_kind_lc_name(tsf_type_kind_t kind_code) {
    switch (kind_code) {
        case TSF_TK_ARRAY:  return "array";
        case TSF_TK_STRUCT: return "struct";
        case TSF_TK_BIT:    return "bit";
        case TSF_TK_CHOICE: return "choice";
        case TSF_TK_VOID:   return "void";
        case TSF_TK_STRING: return "string";
        case TSF_TK_ANY:    return "any";
        default:            return tsf_primitive_type_kind_lc_name(kind_code);
    }
}

/* given a primitive kind code, gives you the ntoh/hton name that
 * we use for it. */
static TSF_inline
const char*     tsf_primitive_type_kind_nh_name(tsf_type_kind_t kind_code) {
    switch (kind_code) {
        case TSF_TK_INT8:   return "c";
        case TSF_TK_UINT8:  return "c";
        case TSF_TK_INT16:  return "s";
        case TSF_TK_UINT16: return "s";
        case TSF_TK_INT32:  return "l";
        case TSF_TK_UINT32: return "l";
        case TSF_TK_INT64:  return "ll";
        case TSF_TK_UINT64: return "ll";
        case TSF_TK_FLOAT:  return "f";
        case TSF_TK_DOUBLE: return "d";
        default:            return "unknown";
    }
}

/* does a primitive instance of.  one primitive is instance of another
 * if the domain of the first is contained in the domain of the second.
 * if you wish to assign one primitive to another, then to avoid
 * a loss of significant or an overflow, you must ensure that the
 * source is instance of the destination. */
tsf_bool_t      tsf_primitive_type_kind_instanceof(tsf_type_kind_t one,
                                                   tsf_type_kind_t two);

/* returns a primitive type kind that matches the requires specification. */
tsf_type_kind_t tsf_primitive_type_kind_for_spec(int32_t lower_bound,
						 int32_t upper_bound,
						 int32_t precision);

/* returns the smallest common subtype of the two types, or void if
 * a common subtype does not exist. */
tsf_type_kind_t tsf_primitive_type_kind_mix(tsf_type_kind_t one,
					    tsf_type_kind_t two);

/* tells you if types with the given kind type are containers. */
static TSF_inline
tsf_bool_t      tsf_type_kind_is_container(tsf_type_kind_t kind_code) {
    switch (kind_code) {
        case TSF_TK_ARRAY:
        case TSF_TK_STRUCT:
        case TSF_TK_CHOICE: return tsf_true;
        default:            return tsf_false;
    }
}

/* tells you if types with the given kind type can store comments. */
static TSF_inline
tsf_bool_t      tsf_type_kind_is_commentable(tsf_type_kind_t kind_code) {
    return tsf_type_kind_is_container(kind_code);
}

/* create a type table (you should almost never have to do this
 * directly.) */
tsf_type_table_t* tsf_type_table_create(void);

/* copy the contents of one type table into another */
tsf_bool_t      tsf_type_table_copy(tsf_type_table_t *dest,
                                    tsf_type_table_t *src);

/* destroy a type table */
void            tsf_type_table_destroy(tsf_type_table_t *tt);

/* write the type table */
tsf_bool_t      tsf_type_table_write(tsf_type_table_t *tt,
                                     tsf_writer_t writer,
                                     void *arg);

/* append a type to the type table */
tsf_bool_t      tsf_type_table_append(tsf_type_table_t *tt,
                                      const char *name,
                                      tsf_type_t *type);

/* find a named type node in the table using the name. */
tsf_named_type_t* tsf_type_table_find_by_name(tsf_type_table_t *tt,
                                              const char *name);

/* find a named type node in the table using the index. */
#define         tsf_type_table_find_by_index(tt,index) \
                    ((tt)->elements[index])

/* get the number of elements in the type table. */
#define         tsf_type_table_get_num_elements(tt) \
                    ((tt)->num_elements)

/* get the hash for a type table.  you are guaranteed that this is just
 * the sum of the hash codes of the constituent tsf_named_type_t objects.
 * hence, you can always determine what the hash code is after an
 * apppend operation. */
int             tsf_type_table_get_hash(tsf_type_table_t *tt);

/* determine is the type table contains any dynamic types. */
tsf_bool_t      tsf_type_table_contains_dynamic(tsf_type_table_t *tt);

/* compare two type tables. */
tsf_bool_t      tsf_type_table_compare(tsf_type_table_t *a,
                                       tsf_type_table_t *b);

/* create a table of types. */
tsf_type_in_map_t* tsf_type_in_map_create(void);

/* create a table of types inside of the provided region. */
tsf_type_in_map_t* tsf_type_in_map_create_in(void *region);

/* return the global empty table of types.  this one should never
 * be modified. */
tsf_type_in_map_t* tsf_type_in_map_get_empty_singleton(void);

/* clone a table of types. */
tsf_type_in_map_t* tsf_type_in_map_clone(tsf_type_in_map_t *types);

/* clone a table of types, creating the new one in a region. */
tsf_type_in_map_t* tsf_type_in_map_clone_in(tsf_type_in_map_t *types,
                                            void *region);

/* destroy a table of types.  has no effect if passed the
 * empty singleton.  cannot be used on region-allocated tables. */
void            tsf_type_in_map_destroy(tsf_type_in_map_t *types);

/* prepare some number of slots for types in the table.  a call to this
 * function serves as a hint and so does not have any guaranteed effect.
 * as such, success in this function should not be taken to mean that
 * the subsequent append operations are guaranteed to succeed and
 * failure in this operation should not be taken to mean that any
 * number of subsequent appends should fail.
 *
 * (In other words, the return value can be safely ignored.  It is there
 * to aid in debugging only.) */
tsf_bool_t      tsf_type_in_map_prepare(tsf_type_in_map_t *types,
                                        uint32_t num);

/* append a type to the table.  the index will be the number of
 * types prior to the append. */
tsf_bool_t      tsf_type_in_map_append(tsf_type_in_map_t *types,
                                       tsf_type_t *type);

/* returns the number of types in the table. */
#define         tsf_type_in_map_get_num_types(types) \
                    ((types)->num_elements)

/* returns the type at the given index in the table. */
#define         tsf_type_in_map_get_type(types,i) \
                    ((types)->elements[i])

/* reduces the size of the type table to the specified new size.
 * unpredictably horrible things will happen if the new size is
 * greater than the current size. */
#define         tsf_type_in_map_truncate(types,new_size) \
                    ((void)((types)->num_elements = (new_size)))

/* creates a new empty type map. */
tsf_type_out_map_t* tsf_type_out_map_create(void);

/* destroy a type map. */
void            tsf_type_out_map_destroy(tsf_type_out_map_t *type_map);

/* look up a type code asserting that it must be found.  an error is
 * returned if it is not found. */
tsf_bool_t      tsf_type_out_map_find_type_code(tsf_type_out_map_t *type_map,
                                                tsf_type_t *type,
                                                uint32_t *type_code);

/* look up a type code, potentially creating a new one.  fails only if
 * a type code had to be created AND that creation failed. */
tsf_bool_t      tsf_type_out_map_get_type_code(tsf_type_out_map_t *type_map,
                                               tsf_type_t *type,
                                               uint32_t *type_code,
                                               tsf_bool_t *created);

/* destroy a type out map and produce a type in map thingy from it. */
tsf_type_in_map_t* tsf_type_in_map_from_type_out_map(
                                                tsf_type_out_map_t *type_map);

/* destroy a type out map and produce a type in map thingy from it.
 * creates the type in thingy in the provided region. */
tsf_type_in_map_t* tsf_type_in_map_from_type_out_map_in(
                                                tsf_type_out_map_t *type_map,
                                                void *region);

/* create an empty random-access type manager. */
tsf_ra_type_man_t* tsf_ra_type_man_create(void);

/* parse a random-access type manager. */
tsf_ra_type_man_t* tsf_ra_type_man_read(tsf_reader_t reader,
					void *arg);

/* destroy a random-access type manager. */
void            tsf_ra_type_man_destroy(tsf_ra_type_man_t *man);

/* write the stuff in the type manager out to somewhere. */
tsf_bool_t      tsf_ra_type_man_write(tsf_ra_type_man_t *man,
				      tsf_writer_t writer,
				      void *arg);

/* inset a type into the type manager.  do this only if you know
   that the type is not already in the type manager. */
tsf_bool_t      tsf_ra_type_man_insert(tsf_ra_type_man_t *man,
				       tsf_type_t *type,
				       uint32_t type_code,
				       uint64_t ref_count);

/* remove a type from the type manager.  do this only if you know
   that the type is in the type manager. */
void            tsf_ra_type_man_remove(tsf_ra_type_man_t *man,
				       tsf_type_t *type);

/* find a type in the type manager using the type code.  if it's
   not found, NULL will be retorned and an error code will be set. */
tsf_type_t*     tsf_ra_type_man_find(tsf_ra_type_man_t *man,
				     uint32_t type_code);

/* determine if the type manager has the given type. */
tsf_bool_t      tsf_ra_type_man_has_type(tsf_ra_type_man_t *man,
					 tsf_type_t *type);

/* get the type code for a type.  only call this if you know that
   the type manager has the type. */
uint32_t        tsf_ra_type_man_get_type_code(tsf_ra_type_man_t *man,
					      tsf_type_t *type);

/* increase the reference count for the given type.  if the type
   is not in the type manager, it is added to the type manager,
   assigned a type code, and given a reference count of 1. */
tsf_bool_t      tsf_ra_type_man_get_reference(tsf_ra_type_man_t *man,
					      tsf_type_t *type);

/* decrease the reference count for a given type.  if the reference
   count drops to 0, the type is removed. */
void            tsf_ra_type_man_remove_reference(tsf_ra_type_man_t *man,
						 tsf_type_t *type);

/* determine if the type manager contains all of the types used by
   the given buffer. */
tsf_bool_t      tsf_ra_type_man_has_buf_types(tsf_ra_type_man_t *man,
					      tsf_buffer_t *buf);

/* increase the reference count for the types in the given buffer.
   creates any types that don't exist. */
tsf_bool_t      tsf_ra_type_man_get_buf_references(tsf_ra_type_man_t *man,
						   tsf_buffer_t *buf);

/* decrease the reference count for the types in the given buffer. */
void            tsf_ra_type_man_remove_buf_references(tsf_ra_type_man_t *man,
						      tsf_buffer_t *buf);

/* calculate the number of type codes on the free-list. */
uint32_t        tsf_ra_type_man_get_type_code_free_list_size(
						     tsf_ra_type_man_t *man);

/* create a type object.  this function will not accept TSF_TK_ARRAY.
 * if you want to create a type with kind ccode TSF_TK_ARRAY, use
 * tsf_type_create_array(). */
tsf_type_t*     tsf_type_create(tsf_type_kind_t kind_code);

/* create an array type object. */
tsf_type_t*     tsf_type_create_array(tsf_type_t *element_type);

/* create a struct type all at once.  note that you can use
 * tsf_type_create() in combination with tsf_struct_type_append() to do
 * what this method does.  also, you can safely use
 * tsf_struct_type_append() on the type returned from this method.
 *
 * the way you would actually use this is to pass struct element names
 * paired with types.  for example:
 *
 * tsf_type_t *my_type =
 *     tsf_type_create_struct("x",
 *                            tsf_type_create(TSF_TK_FLOAT),
 *                            "y",
 *                            tsf_type_create(TSF_TK_FLOAT),
 *                            NULL);
 *
 * note that you must pass NULL after the last pair.  otherwise you'll
 * get random crashes and stuff.
 *
 * the types passed to this function become owned by the function
 * regardless of success or error. */
tsf_type_t*     tsf_type_create_struct(const char *first_name,
                                       ...);

/* the same as the above but tailored to allow you to use it more easily
 * from your own subrutines that use variable args. */
tsf_type_t*     tsf_type_create_structv(const char *first_name,
                                        va_list lst);

/* create a choice type all at once.  note that you can use
 * tsf_type_create() in combination with tsf_choice_type_append() to do
 * what this method does.  also, you can safely use
 * tsf_choice_type_append() on the type returned from this method.
 *
 * the way you would actually use this is to pass struct element names
 * paired with types.  for example:
 *
 * tsf_type_t *my_type =
 *     tsf_type_create_choice("x",
 *                            tsf_type_create(TSF_TK_FLOAT),
 *                            "y",
 *                            tsf_type_create(TSF_TK_FLOAT),
 *                            NULL);
 *
 * note that you must pass NULL after the last pair.  otherwise you'll
 * get random crashes and stuff.
 *
 * the types passed to this function become owned by the function
 * regardless of success or error. */
tsf_type_t*     tsf_type_create_choice(const char *first_name,
                                       ...);

/* the same as the above but tailored to allow you to use it more easily
 * from your own subrutines that use variable args. */
tsf_type_t*     tsf_type_create_choicev(const char *first_name,
                                        va_list lst);

/* read a type. */
tsf_type_t*     tsf_type_read(tsf_reader_t reader,
                              void *arg,
                              tsf_limits_t *limits);

/* write a type. */
tsf_bool_t      tsf_type_write(tsf_type_t *type,
                               tsf_writer_t writer,
                               void *arg);

/* duplicate a type.  this gives you a copy of the type that must be
 * separately destroyed.  so, if you wish to retain the type after a
 * call to a function that might destroy it, call this first. */
tsf_type_t*     tsf_type_dup(tsf_type_t *type);

/* destroy a type object. */
void            tsf_type_destroy(tsf_type_t *type);

/* get the size in bytes of the type.  note that this operation cannot
 * fail.  note further that the size is lazily computed, so the first time
 * you call this, it will be expensive, but it will become cheap for all
 * subsequent calls. */
uint32_t        tsf_type_get_byte_size(tsf_type_t *type);

/* verify a type against some limits; this is a fairly costly process,
 * probably at least as expensive as writing or reading a type.  tsf_true
 * is returned if the type is OK.  if tsf_false is returned,
 * TSF_E_NOT_ALLOWED is set.  if limits is NULL, all types should
 * verify */
tsf_bool_t      tsf_type_verify(tsf_type_t *type,
                                tsf_limits_t *limits);

/* get a hash code for a type. */
static TSF_inline
int             tsf_type_get_hash(tsf_type_t *type) {
    return type->hash_code;
}

/* determine if the type is dynamic (that is, if it has any anies in
   it) */
#define         tsf_type_is_dynamic(type) \
                    ((type)->is_dynamic)

/* compare two types.  neither type becomes owned by the function; both
 * are still your responsibility after the call.  returns tsf_true
 * if the types are equal, tsf_false otherwise. */
tsf_bool_t      tsf_type_compare(tsf_type_t *a,tsf_type_t *b);

/* determines if the first type is an instance of the second type.
 * the instanceof relationship is defined in a big comment up above
 * the definition of the tsf_type_kind enumeration. */
tsf_bool_t      tsf_type_instanceof(tsf_type_t *one,
                                    tsf_type_t *two);

/* get the kind code for a type object. */
#define         tsf_type_get_kind_code(type) \
                    ((type)->kind_code)

/* if the size of the type is known statically, it returns it.  otherwise,
 * it returns UINT32_MAX. */
uint32_t        tsf_type_get_static_size(tsf_type_t *type);

/* tells you the comment on the type.  may be NULL if there is no comment
 * or if the type is not commentable. */
#define         tsf_type_get_comment(type) \
                    ((type)->comment)

/* tells you if the type has a comment */
#define         tsf_type_has_comment(type) \
                    (tsf_type_get_comment(type)!=NULL)

/* sets the comment on the type.  will return TSF_E_BAD_TYPE_KIND if the
 * type is not commentable. */
tsf_bool_t      tsf_type_set_comment(tsf_type_t **type,
                                     const char *comment);

/* get the element type of an array type object.  the returned
 * object is not owned by you; it is owned by the array type. */
#define         tsf_array_type_get_element_type(type) \
                    ((type)->u.a.element_type)

/* append a new field to a struct type.  the string argument
 * does not become owned by the function; it is still your
 * responsibility to free it.  note that the string cannot
 * exceed TSF_STR_SIZE; if it does, an error will be returned.
 * this will propagate errors on the *ele_type. */
tsf_bool_t      tsf_struct_type_append(tsf_type_t **type,
                                       const char *name,
                                       tsf_type_t *ele_type);

/* set the 'optional' attribute of a structure field. */
tsf_bool_t      tsf_struct_type_set_optional(tsf_type_t **type,
					     const char *name,
					     tsf_bool_t optional);

/* set the comment for a structure field. */
tsf_bool_t      tsf_struct_type_set_comment(tsf_type_t **type,
                                            const char *name,
                                            const char *comment);

/* find an element in a struct type.  this search should be
 * fast, as it uses a hashtable. */
tsf_type_t*     tsf_struct_type_find(tsf_type_t *type,
                                     const char *name);

/* find the pair element in a struct type; the pair
 * element contains more info.  this search should be fast,
 * as it uses a hashtable. */
#define         tsf_struct_type_find_node(type,name) \
                    (tsf_type_table_find_by_name((type)->u.s.tt,(name)))

/* returns the number of elements in the struct. */
#define         tsf_struct_type_get_num_elements(type) \
                    (tsf_type_table_get_num_elements((type)->u.s.tt))

/* get the ith element in the struct */
#define         tsf_struct_type_get_element(type,i) \
                    (tsf_type_table_find_by_index((type)->u.s.tt,(i)))

/* get the name of the ith element in the struct */
#define         tsf_struct_type_get_element_name(type,i) \
                    (tsf_named_type_get_name(\
                         tsf_struct_type_get_element(type,i)))

/* get the type of the ith element in the struct */
#define         tsf_struct_type_get_element_type(type,i) \
                    (tsf_named_type_get_type(\
                         tsf_struct_type_get_element(type,i)))

/* append a new field to a choice type.  the string argument
 * does not become owned by the function; it is still your
 * responsibility to free it.  note that the string cannot
 * exceed TSF_STR_SIZE; if it does, an error will be returned.
 * this will propagate errors on the *ele_type. */
tsf_bool_t      tsf_choice_type_append(tsf_type_t **type,
                                       const char *name,
                                       tsf_type_t *ele_type);

/* set the comment for a choice field. */
tsf_bool_t      tsf_choice_type_set_comment(tsf_type_t **type,
                                            const char *name,
                                            const char *comment);

/* find an element in a choice type.  this search should be
 * fast, as it uses a hashtable. */
tsf_type_t*     tsf_choice_type_find(tsf_type_t *type,
                                     const char *name);

/* find the pair element in a choice type; the pair
 * element contains more info.  this search should be fast,
 * as it uses a hashtable. */
#define         tsf_choice_type_find_node(type,name) \
                    (tsf_type_table_find_by_name((type)->u.h.tt,(name)))

/* returns the number of elements in the choice. */
#define         tsf_choice_type_get_num_elements(type) \
                    (tsf_type_table_get_num_elements((type)->u.h.tt))

/* get the ith element in the choice */
#define         tsf_choice_type_get_element(type,i) \
                    (tsf_type_table_find_by_index((type)->u.h.tt,(i)))

/* get the name of the ith element in the choice */
#define         tsf_choice_type_get_element_name(type,i) \
                    (tsf_named_type_get_name(\
                         tsf_choice_type_get_element(type,i)))
		     
/* get the type of the ith element in the choice */
#define         tsf_choice_type_get_element_type(type,i) \
                    (tsf_named_type_get_type(\
                         tsf_choice_type_get_element(type,i)))

/* returns true if the choice has any non-void types. */
tsf_bool_t      tsf_choice_type_has_non_void(tsf_type_t *type);

/* returns true if:
 * -> the latter choice type has at least as many elements as the
 *    first, and
 * -> for each of the elements in the first choice type, the element
 *    in the second choice type that has the same index also has the
 *    same name.
 *
 * note that the element types do not have to match for this
 * predicate to hold.  however, if you are interested in adding an
 * instanceof constraint for the corresponding elements of the two
 * choice types, you can do so by performing an instanceof test on
 * a and b. */
tsf_bool_t      tsf_choice_type_uses_same_indices(tsf_type_t *a,
                                                  tsf_type_t *b);

/* creates a named type object with the given name, type, and index.
 * failure to create this thing results in the type being destroyed.
 * the has_offset field is set to false, the offset field is set to 0,
 * and the comment field is set to NULL. */
tsf_named_type_t* tsf_named_type_create(const char *name,
                                        tsf_type_t *type,
                                        uint32_t index);

/* destroy a named type object.  destroys the type.  frees the
 * comment if there is one. */
void            tsf_named_type_destroy(tsf_named_type_t *node);

/* returns the name of the type in the node.  what gets returned
 * should be treated as a constant string that you do not own. */
#define         tsf_named_type_get_name(node) \
                    ((node)->name)

/* returns the type that the node holds.  what gets returned is
 * of type tsf_type_t* and should be treated as constant that you
 * do not own. */
#define         tsf_named_type_get_type(node) \
                    ((node)->type)

/* returns the index of a named type node within its structure
 * type. */
#define         tsf_named_type_get_index(node) \
                    ((node)->index)
		     
/* set the optional attribute */
#define         tsf_named_type_get_optional(node) \
                    ((node)->optional)

/* set the optional attribute */
#define         tsf_named_type_set_optional(node,val) \
                    ((node)->optional=(val))

/* returns the comment of a names type.  may be NULL, indicating that
 * no comment was ever set */
#define         tsf_named_type_get_comment(node) \
                    ((node)->comment)

/* tells you if the named type has a comment */
#define         tsf_named_type_has_comment(node) \
                    (tsf_named_type_get_comment(node)!=NULL)

/* set the comment on a named type.  deletes the old one if
 * necessary. */
tsf_bool_t      tsf_named_type_set_comment(tsf_named_type_t *node,
                                           const char *comment);

/* get the hash code for the named type.  you are guaranteed that types
 * of type TSF_TK_STRUCT and TSF_TK_CHOICE will use the sum of these hash
 * codes as part of computing their own hash code.  you're guaranteed
 * that tsf_type_table_table_t will use the sum of these hash codes as
 * part of computing their own hash code. */
int             tsf_named_type_get_hash(tsf_named_type_t *node);

/* create a buffer.  this is a bit of a trusted function.  the type
 * is dupped by the buffer.  the types table becomes owned.  and, to
 * make things more interesting, the boolean argument determines if
 * the void* is owned by the buffer.  if it is owned by the buffer,
 * then if this function fails, the buffer will be freed. */
tsf_buffer_t*   tsf_buffer_create(tsf_type_t *type,
                                  tsf_type_in_map_t *types,
                                  void *data,
                                  uint32_t size,
                                  tsf_bool_t keep);

/* create a buffer in a region.  exactly like tsf_buffer_create()
 * except that the buffer is created in a region.  it is assumed
 * that the void* data is allocated in the same region and that
 * tsf_type_in_map_t* is, too.  note that the tsf_bool_t argument
 * is omitted since tsf_buffer_destroy() is not applicable to
 * region-allocated buffers. */
tsf_buffer_t*   tsf_buffer_create_in(tsf_type_t *type,
                                     tsf_type_in_map_t *types,
                                     void *data,
                                     uint32_t size,
                                     void *region);

/* get the empty buffer singleton. */
tsf_buffer_t*   tsf_buffer_get_empty_singleton(void);

/* clone the buffer.  the resulting buffer has the keep bit set to
 * true.  note that this is a deep clone, so that the actual data
 * itself is also copied.  this means that this operation, and any
 * operations on the new buffer have no effect on the original buffer. */
tsf_buffer_t*   tsf_buffer_clone(tsf_buffer_t *buf);

/* clone the buffer into the given region.  exactly like
 * tsf_buffer_clone() except for the region part. */
tsf_buffer_t*   tsf_buffer_clone_in(tsf_buffer_t *buf,
                                    void *region);

/* duplicate a buffer.  this will give you access to the given buffer
 * until you call tsf_buffer_destroy on it.  if the buffer was heap-
 * allocated and has the keep bit set to true, this operation will be
 * very fast - it will increment a count and maybe do some simple
 * locking/unlocking.  if the buffer was region allocated or if the
 * keep bit was set to false, then this has the same cost as
 * tsf_buffer_clone().  as such, you must do error checking after a
 * a call to tsf_buffer_dup(). */
tsf_buffer_t*   tsf_buffer_dup(tsf_buffer_t *buf);

/* destroy a buffer.  cannot be used on region allocated buffers. */
void            tsf_buffer_destroy(tsf_buffer_t *buf);

/* tells you the size of the buffer */
#define         tsf_buffer_get_size(buf) \
                    ((buf)->size)

/* read a buffer and associate it with the given type.  this is
 * a bit of a trusted routine... */
tsf_buffer_t*   tsf_buffer_read(tsf_type_in_map_t *enclosing_types,
                                tsf_reader_t reader,
                                void *arg,
                                tsf_limits_t *limits);

/* read a buffer from another buffer and associate it with the given
 * type.  this is also a bit of a trusted routine...
 * if the region pointer is non-NULL, then the resulting buffer will
 * be region-allocated.  otherwise it will be malloc-allocated and
 * you will be expected to call tsf_buffer_destroy() at some point. */
tsf_buffer_t*   tsf_buffer_read_from_buf(tsf_type_in_map_t *enclosing_types,
                                         uint8_t **cur,
                                         uint8_t *end,
                                         void *region);

/* skip a buffer (in a buffer, not on disk).  takes care of skipping the
   translation table if dealing with a dynamic type. */
tsf_bool_t      tsf_buffer_skip_in_buf(tsf_type_in_map_t *enclosing_types,
                                       uint8_t **cur,
                                       uint8_t *end);

/* write a buffer.  the type in the buffer is ignored. */
tsf_bool_t      tsf_buffer_write(tsf_buffer_t *buf,
                                 tsf_type_out_map_t *type_map,
                                 tsf_writer_t writer,
                                 void *arg);

/* calculate the size that would be required to write this buffer into
 * another buffer. */
static TSF_inline
uint32_t        tsf_buffer_calc_size(tsf_buffer_t *buf) {
    return (tsf_type_is_dynamic(buf->type)?
            4+4*tsf_type_in_map_get_num_types(buf->types):0)+
           buf->size+4+4;
}

/* write a buffer into another buffer */
tsf_bool_t      tsf_buffer_write_to_buf(tsf_buffer_t *buf,
                                        tsf_type_out_map_t *type_map,
                                        uint8_t **target);

/* given a buffer, gives you the type. */
#define         tsf_buffer_get_type(buf) \
                    ((buf)->type)

/* given a buffer, tells you if its type is an instance of another type */
#define         tsf_buffer_instanceof(buf,type) \
                    (tsf_type_instanceof(tsf_buffer_get_type(buf),(type)))

/* given a buffer, gives you the shared types. */
#define         tsf_buffer_get_shared_types(buf) \
                    ((buf)->types)

/* create an serial input manager.  a serial input manager expects zero or
 * more type headers to precede data. */
tsf_serial_in_man_t* tsf_serial_in_man_create(tsf_reader_t reader,
                                              void *arg,
                                              tsf_limits_t *limits);

/* destroy an input manager. */
void            tsf_serial_in_man_destroy(tsf_serial_in_man_t *in_man);

/* force the serial input manager to read a TSF_SP_C_RESET_TYPES command.
 * if this command is not what is found in the input stream, an error is
 * reported.  provided that a TSF_SP_C_RESET_TYPES command is in fact
 * found, the type mapping will be reset.
 *
 * will set error to TSF_E_UNEXPECTED_EOF if there weren't enough bytes for
 * a TSF_SP_C_RESET_TYPES command.
 *
 * will set error to TSF_E_PARSE_ERROR if the bytes read were not the
 * currect ones (effectively, this is complaining of a bad magic number).*/
tsf_bool_t      tsf_serial_in_man_read_reset_types(tsf_serial_in_man_t *in_man);

/* take the given file descriptor and read the reset types command from
 * it.  returns true if there was a reset types command and false on
 * error.  returns the same errors as tsf_serial_in_man_read_reset_types()
 * unless something really weird happens. */
tsf_bool_t      tsf_serial_in_man_read_reset_types_from_fd(int fd);

/* take the given file, open it, and invoke
 * tsf_serial_in_man_read_reset_types_from_fd on it. */
tsf_bool_t      tsf_serial_in_man_read_reset_types_from_file(
                                                        const char *filename);

/* register a callback that will get called every time the input manager
 * encounters a new type. */
tsf_bool_t      tsf_serial_in_man_register_type_cback(
                                                tsf_serial_in_man_t *in_man,
                                                tsf_type_cback_t cback,
                                                void *arg);

/* unregister all type callbacks that match the given function pointer
 * and void* argument. */
uint32_t        tsf_serial_in_man_unregister_type_cback(
                                                tsf_serial_in_man_t *in_man,
                                                tsf_type_cback_t cback,
                                                void *arg);

/* read data using the serial input manager. */
tsf_buffer_t*   tsf_serial_in_man_read(tsf_serial_in_man_t *in_man);

/* get the state of the input manager right now.  if you are using random
 * access files and you wish to seek back to the position you're in now,
 * you also need to save the state of the input manager and restore it
 * right after the seek. */
#define         tsf_serial_in_man_get_state(in_man) \
                    ((in_man)->state)

/* set the state of the input manager.  do this whenever you seek the file
 * that the serial input manager is being used to read. */
void            tsf_serial_in_man_set_state(tsf_serial_in_man_t *in_man,
                                            tsf_serial_in_man_state_t state);

/* create an output manager. */
tsf_serial_out_man_t* tsf_serial_out_man_create(tsf_writer_t writer,
                                                void *arg);

/* destroy an output manager. */
void            tsf_serial_out_man_destroy(tsf_serial_out_man_t *out_man);

tsf_bool_t      tsf_serial_out_man_reset_types(tsf_serial_out_man_t *out_man);

/* write data using an output manager.  this will do error
 * propagation on *buf. */
tsf_bool_t      tsf_serial_out_man_write(tsf_serial_out_man_t *out_man,
                                         tsf_buffer_t *buf);

/* create a tsf data structure using the given type.  the type
 * gets dupped for the returned structure.  this will
 * do error propagation on *type.
 *
 * Note that this function should never be used to create
 * primitives, bits, voids, or strings. */
tsf_reflect_t*  tsf_reflect_create(tsf_type_t *type);

/* create a tsf data structure to hold a value of the given type. */
tsf_reflect_t*  tsf_reflect_create_int8_t(int8_t value);
tsf_reflect_t*  tsf_reflect_create_uint8_t(uint8_t value);
tsf_reflect_t*  tsf_reflect_create_int16_t(int16_t value);
tsf_reflect_t*  tsf_reflect_create_uint16_t(uint16_t value);
tsf_reflect_t*  tsf_reflect_create_int32_t(int32_t value);
tsf_reflect_t*  tsf_reflect_create_uint32_t(uint32_t value);
tsf_reflect_t*  tsf_reflect_create_int64_t(int64_t value);
tsf_reflect_t*  tsf_reflect_create_uint64_t(uint64_t value);
tsf_reflect_t*  tsf_reflect_create_float(float value);
tsf_reflect_t*  tsf_reflect_create_double(double value);

/* macro that allows one to parameterize the above. */
#define         tsf_reflect_create_primitive(type,value) \
                    (tsf_reflect_create_##type(value))

/* create a tsf data struct to hold a bit with the given value. */
tsf_reflect_t*  tsf_reflect_create_bit(tsf_bool_t value);

/* create a 'void' data structure; this is just a singleton for
 * the void type.  you'll probably only have to deal with this if
 * you use the choice type for enumerations. */
tsf_reflect_t*  tsf_reflect_create_void(void);

/* create a string.  the actual char* is duplicated. */
tsf_reflect_t*  tsf_reflect_create_string(const char *str);

/* create an 'any'.  this is a dynamic holder for damn-near
 * anything.  the buffer becomes owned by the reflect object,
 * so that even if this creation fails, the buffer is no longer
 * available to you. */
tsf_reflect_t*  tsf_reflect_create_any(tsf_buffer_t *buf);

/* clone a data structure */
tsf_reflect_t*  tsf_reflect_clone(tsf_reflect_t *data);

/* destroy a data structure */
void            tsf_reflect_destroy(tsf_reflect_t *data);

/* produce a buffer from data */
tsf_buffer_t*   tsf_reflect_make_buffer(tsf_reflect_t *data);

/* produce data from a buffer */
tsf_reflect_t*  tsf_reflect_from_buffer(tsf_buffer_t *buf);

#define         tsf_reflect_get_type(data) \
                    ((data)->type)

#define         tsf_reflect_get_kind_code(data) \
                    (tsf_type_get_kind_code(tsf_reflect_get_type(data)))

/* set a structure element.  this will do error propagation on
 * *ele_data. */
tsf_bool_t      tsf_struct_reflect_set_element(tsf_reflect_t *data,
                                               const char *name,
                                               tsf_reflect_t *ele_data);

/* set a structure element by index.  this will do error propagation
   on *ele. */
tsf_bool_t      tsf_struct_reflect_set_element_by_index(tsf_reflect_t *data,
                                                        uint32_t index,
                                                        tsf_reflect_t *ele);

/* get a structure element */
tsf_reflect_t*  tsf_struct_reflect_get_element(tsf_reflect_t *data,
                                               const char *name);

/* get a structure element by index */
tsf_reflect_t*  tsf_struct_reflect_get_element_by_index(tsf_reflect_t *data,
                                                        uint32_t index);

/* get the number of elements in the structure */
#define         tsf_struct_reflect_get_num_elements(data) \
                    (tsf_struct_type_get_num_elements( \
                        tsf_reflect_get_type(data)))

/* set the choice.  as a convenience shortcut, if choice is NULL, this is
   equivalent to calling tsf_choice_reflect_set_unknown().  this means
   that ele_data is ignored. */
tsf_bool_t      tsf_choice_reflect_set(tsf_reflect_t *data,
                                       const char *choice,
                                       tsf_reflect_t *ele_data);

/* set the choice by index.  as a convenience shortcut, if choice is
   UINT32_MAX, this is equivalent to calling
   tsf_choice_reflect_set_unknown().  this means that ele_data is ignored. */
tsf_bool_t      tsf_choice_reflect_set_by_index(tsf_reflect_t *data,
                                                uint32_t index,
                                                tsf_reflect_t *ele_data);

/* set the choice to unknown. */
tsf_bool_t      tsf_choice_reflect_set_unknown(tsf_reflect_t *data);

/* tells you the choice index */
#define         tsf_choice_reflect_get_index(data) \
                    ((data)->u.h.choice)

/* tells you if the choice is know (not unknown) */
#define         tsf_choice_reflect_is_known(data) \
                    ((data)->u.h.choice!=TSF_UINT32_MAX)

/* get the choice.  if unknown, will return NULL. */
static TSF_inline
const char*     tsf_choice_reflect_get_choice(tsf_reflect_t *data) {
    return tsf_choice_reflect_is_known(data)?
               tsf_choice_type_get_element(data->type,
                                           data->u.h.choice)->name:
               NULL;
}

/* get the data associated with the choice.  if unknown, will
 * return NULL. */
#define         tsf_choice_reflect_get_data(d) \
                    ((d)->u.h.data)

/* append an element to the array.  this will do error propagation
 * on *ele_data. */
tsf_bool_t      tsf_array_reflect_append(tsf_reflect_t *data,
                                         tsf_reflect_t *ele_data);

/* get an array element */
#define         tsf_array_reflect_get_element(data,index) \
                    ((data)->u.c.array[index])

/* get the number of elements in the array */
#define         tsf_array_reflect_get_num_elements(data) \
                    ((data)->u.c.size)

/* get the primitives... */
#define         tsf_primitive_reflect_get(data,type) \
                    ((data)->u.p_##type)

/* get a bit */
#define         tsf_bit_reflect_get(data) \
                    ((data)->u.bit)

/* get the string. */
#define         tsf_string_reflect_get(data) \
                    ((const char*)((data)->u.str))

/* get the buffer from an any. */
#define         tsf_any_reflect_get_buffer(data) \
                    ((data)->u.any)

/* returns tsf_true if the given type has a complete structure mapping. 
 * this does not tell you if the given type has a mapping within its
 * parent type - it only tells you if the types it contains have
 * mappings.  hence, this always returns tsf_true for types that are not
 * containers.  if this function returns tsf_false, then calling any of
 * the tsf_native functions on this type will result in unpredictable
 * behavior (unless otherwise noted). */
tsf_bool_t      tsf_native_type_has_struct_mapping(tsf_type_t *type);

/* get the native size of a particular type within its parent
 * struct.  this can only be called on types that have a struct
 * mapping. */
size_t          tsf_native_type_get_size(tsf_type_t *type);

/* returns true if a memcpy() from source to destination with the source's
 * size will give the 'desired' result. */
tsf_bool_t      tsf_native_type_can_blit(tsf_type_t *dest_type,
                                         tsf_type_t *src_type);

/* returns the offset of a particular named type node. */
#define         tsf_native_named_type_get_offset(n) \
                    ((n)->offset)

/* gives a mapping (offset) for an element within a struct.  note
 * that this function is not particularly fast.  however, you shouldn't
 * care about the speed of this function since you should only ever
 * call it once for each structure element for the lifetime of your
 * program. */
tsf_bool_t      tsf_native_struct_type_map(tsf_type_t **type,
                                           const char *name,
                                           size_t offset);

/* gives a size to a struct.  note that like tsf_native_struct_type_map(),
 * this function is not particularly fast. */
tsf_bool_t      tsf_native_struct_type_set_size(tsf_type_t **type,
                                                size_t size);

/* set the constructor and destructor that get called when a struct
   gets allocated and deallocated. */
tsf_bool_t      tsf_native_struct_type_set_callbacks(tsf_type_t **type,
						     void (*cons)(void*),
						     void (*dest)(void*));

/* get the constructor to call when the struct gets allocated */
#define         tsf_native_struct_type_get_constructor(type) \
                    ((type)->u.s.constructor)

/* get the destructor to call when the struct gets deallocated */
#define         tsf_native_struct_type_get_destructor(type) \
                    ((type)->u.s.destructor)

/* set all of the offsets and things required by choice */
tsf_bool_t      tsf_native_choice_type_map(tsf_type_t **type,
					   tsf_bool_t in_place,
                                           size_t value_offset,
                                           size_t data_offset,
                                           size_t size);

/* is the choice mapping in place? */
#define         tsf_native_choice_type_is_in_place(type) \
                    ((type)->u.h.in_place)

/* retrieve the data offset */
#define         tsf_native_choice_type_get_data_offset(type) \
                    ((type)->u.h.data_offset)

/* retrieve the value offset */
#define         tsf_native_choice_type_get_value_offset(type) \
                    ((type)->u.h.value_offset)

/* automatically produces some sort of mapping.  you probably don't want
 * to use this method, since you get no guarantee that this will map to
 * any of your structures.  this method's main intent is to aid in the
 * writing of regression tests.
 *
 * if do_align is tsf_true, each element will be aligned according to
 * the detected native alignment (the alignment necessary for the largest
 * data type, which is usually determined by either pointers or doubles).
 *
 * if do_align is tsf_false, no alignment guarantee is made.  you have
 * to be quite careful when using this option, as it might cause the
 * GPC programs generated for this type to crash. */
tsf_bool_t      tsf_native_type_produce_some_mapping(tsf_type_t **type,
						     tsf_bool_t do_align);

/* given a type that has complete struct mapping, produce a generator
 * that can be used to produce tsf_buffer_t objects from pointers
 * to the struct. */
tsf_genrtr_t*   tsf_generator_create(tsf_type_t *type,
                                     tsf_engine_t conf,
				     const char *cache_dir);

/* generate a buffer using some data and a generator. */
tsf_buffer_t*   tsf_generator_generate(tsf_genrtr_t *gen,
                                       void *data);

/* generate a buffer in a region using some data and a generator. */
tsf_buffer_t*   tsf_generator_generate_in(tsf_genrtr_t *gen,
                                          void *data,
                                          void *region);

/* destroy the generator. */
void            tsf_generator_destroy(tsf_genrtr_t *gen);

/* given a type that has a complete struct mapping, produce a parser
 * that can be used to read tsf_buffer_t objects and produce data
 * that has the structure you specified in the struct mapping. */
tsf_parser_t*   tsf_parser_create(tsf_type_t *type,
                                  tsf_engine_t conf,
				  const char *cache_dir);

/* parse a buffer and generate data according to the struct mapping
 * in the type you gave to tsf_parser_create().
 *
 * stuff returned from this function can be freed using
 * tsf_region_free() (see tsf_region.h). */
void*           tsf_parser_parse(tsf_parser_t *parser,
                                 tsf_buffer_t *buffer);

/* parse a buffer and generate data according to the struct mapping
 * in the type you gave to tsf_parser_create().  the data is
 * allocated in the given region instead of having a new region
 * created. */
void*           tsf_parser_parse_in(tsf_parser_t *parser,
                                    tsf_buffer_t *buffer,
                                    void *region);

/* destroy a parser. */
void            tsf_parser_destroy(tsf_parser_t *parser);

/* create a copier given a destination type and a source type.  if the
 * source is not instanceof the destination, you will get an error. */
tsf_copier_t*   tsf_copier_create(tsf_type_t *dest,
                                  tsf_type_t *src,
                                  tsf_engine_t conf,
				  const char *cache_dir);

/* destroy the copier. */
void            tsf_copier_destroy(tsf_copier_t *copier);

/* clone the source data, which has the source type, and produce an
 * object of the destination type.  the object returned can be
 * freed with tsf_region_free(). */
void*           tsf_copier_clone(tsf_copier_t *copier,
                                 void *src);

/* clone the source data, which has the source type, and produce an
 * object of the destination type.  the object returned will be
 * allocated in the given region. */
void*           tsf_copier_clone_in(tsf_copier_t *copier,
				    void *src,
				    void *region);

/* copy from the source to the destination, following the specified types.
 * also, use the provided region for allocation.  note that the region
 * cannot be NULL. */
tsf_bool_t      tsf_copier_copy(tsf_copier_t *copier,
                                void *dest,
                                void *src,
                                void *region);

/* file descriptor writer; the void *arg should be a pointer to
 * an integer holding the file descriptor. */
tsf_bool_t      tsf_fd_writer(void *arg,
                              void *data,
                              uint32_t len);

/* file descriptor reader; the void *arg should be a pointer to
 * an integer holding the file descriptor.  if an EOF happens before
 * any data is ready, this function will set the TSF_E_UNEXPECTED_EOF
 * error.  if an EOF happens after any data is ready,
 * TSF_E_ERRONEOUS_EOF will be set instead. if the EOF was expected,
 * you should change the TSF_E_UNEXPECTED_EOF error to TSF_E_EOF if
 * you propagate that error (see tsf_in_man.c for a good example of
 * what I'm talking about). */
tsf_bool_t      tsf_fd_reader(void *arg,
                              void *data,
                              uint32_t len);

/* file descriptor partial reader; the void *arg should be a pointer
 * to an integer holding the file descriptor.  this differs from
 * tsf_fd_reader() in that complete writes are never guaranteed.
 * if an EOF happens, this function will set the TSF_E_UNEXPECTED_EOF
 * error.  if the EOF was expected, you should change the
 * TSF_E_UNEXPECTED_EOF error to TSF_E_EOF if you propagate that error
 * (see tsf_in_man.c for a good example of what I'm talking about). */
int32_t         tsf_fd_partial_reader(void *arg,
                                      void *data,
                                      uint32_t len);

tsf_bool_t      tsf_full_read_of_partial(tsf_partial_reader_t reader,
					 void *arg,
					 void *data,
					 uint32_t len);

/* create a buffered reader using the given file descriptor and
 * buffer size.  if keep_fd is tsf_true, the file descriptor will be
 * closed if: an error occurs while trying to create the buffer,
 * and when the buffer is destroyed. */
tsf_buf_rdr_t*  tsf_buf_reader_create(int fd,
                                      uint32_t buf_size,
                                      tsf_bool_t keep_fd);

/* destroy a read buffer */
void            tsf_buf_reader_destroy(tsf_buf_rdr_t *buf);

/* read using a read buffer */
tsf_bool_t      tsf_buf_reader_read(void *buf,
                                    void *data,
                                    uint32_t len);

/* tells you how many bytes of data are remaining in the buffer. */
static TSF_inline
uint32_t        tsf_buf_reader_get_remaining(tsf_buf_rdr_t *buf) {
    return buf->puti - buf->geti;
}

/* resets the buffer. */
void            tsf_buf_reader_reset(tsf_buf_rdr_t *buf);

/* create a buffered writer using the given file descriptor, buffer size.
 * if keep_fd is tsf_true, the file descriptor will be closed if: an error
 * occurs while trying to create the buffer, and when the buffer is destroyed. */
tsf_buf_wtr_t*  tsf_buf_writer_create(int fd,
                                      uint32_t buf_size,
                                      tsf_bool_t keep_fd);

/* destroy the buffer.  if keep_fd is tsf_true, the file descriptor
 * is closed.  note that this operation does not result in the
 * buffer being flushed, so calling this without calling the
 * flush function (see below) first may result in data loss. */
void            tsf_buf_writer_destroy(tsf_buf_wtr_t *buf);

/* this is the function that you pass to all of the writing operations
 * if you actually want to put this writer to good use. */
tsf_bool_t      tsf_buf_writer_write(void *buf,
                                     void *data,
                                     uint32_t len);

/* flush the buffer. */
tsf_bool_t      tsf_buf_writer_flush(tsf_buf_wtr_t *buf);

/* get the 'default' zip mode.  this is actually NONE, which will
 * result in illegal argument errors in tsf_zip_reader_create() and
 * tsf_zip_writer_create().  however, tsf_stream_file_input and
 * tsf_stream_file_output will gracefully accept it and take it
 * to mean that normal buffered IO should be performed. */
void            tsf_zip_wtr_attr_get_default(tsf_zip_wtr_attr_t *attr);

/* get the default ZLib mode.  this will cause ZLib to be used with
 * default settings (a compression level of Z_DEFAULT_COMPRESSION and
 * advanced_init set to false).  this function will return tsf_true
 * if ZLib is supported, and tsf_false otherwise.  if tsf_false is
 * returned, the error will be set to TSF_E_NOT_SUPPORTED.  to make
 * life easier in some cases, you can pass NULL to this function and
 * the function will still perform the check if Zlib is supported. */
tsf_bool_t      tsf_zip_wtr_attr_get_zlib_default(tsf_zip_wtr_attr_t *attr);

/* get the default libbzip2 mode.  this will cause libbzip2 to be
 * used with default settings (blockSize100k = 9, verbosity = 0,
 * workFactor = 0, small = 0).  this function will return tsf_true
 * if libbzip2 is supported, and tsf_false otherwise.  if tsf_false
 * is returned, the error will be set to TSF_E_NOT_SUPPORTED.  to
 * make life easier in some cases, you can pass NULL to this function
 * and the function will still perform the check if libbzip2 is
 * supported. */
tsf_bool_t      tsf_zip_wtr_attr_get_bzip2_default(tsf_zip_wtr_attr_t *attr);

/* get the attributes that correspond to the given mode. */
tsf_bool_t      tsf_zip_wtr_attr_get_for_mode(tsf_zip_wtr_attr_t *attr,
					      tsf_zip_mode_t mode);

/* get the default decompression settings.  this enables only those
 * complression schemes that are actually supported. */
void            tsf_zip_rdr_attr_get_default(tsf_zip_rdr_attr_t *attr);

/* get the settings with all compression disabled.  this sets the
 * allow bits to false, but all of the other fields are set to the
 * default settings - so this is the perfect starting place if you want
 * to enable only your approved compression schemes (just call this
 * and then set the bits of the schemes that make you happy). */
void            tsf_zip_rdr_attr_get_nozip(tsf_zip_rdr_attr_t *attr);

/* is all compression disabled? */
tsf_bool_t      tsf_zip_rdr_attr_is_nozip(tsf_zip_rdr_attr_t *attr);

/* convenience macro that checks if zlib support is available */
#define         tsf_zlib_supported() \
                    (tsf_zip_wtr_attr_get_zlib_default(NULL))

/* convenience macro that checks if libbzip2 support is available */
#define         tsf_libbzip2_supported() \
                    (tsf_zip_wtr_attr_get_bzip2_default(NULL))

/* create a zipping reader.  it also does buffering.  accepts a partial
 * reader as a delegate to actually read the data.  as such, this thing
 * can adapt to many different underlying devices.  returns an error
 * if you try to use a mode or attribute set that is not supported.
 * In particular, using TSF_ZIP_NONE is not supported!
 *
 * note that this function does not check if the selected mode is allowed
 * according to the attributes.  it is the caller's responsibility to
 * ensure that the mode is allowed. */
tsf_zip_rdr_t*  tsf_zip_reader_create(tsf_partial_reader_t reader,
                                      void *reader_arg,
                                      tsf_zip_mode_t mode,
                                      const tsf_zip_rdr_attr_t *attr);

/* destroy the reader. */
void            tsf_zip_reader_destroy(tsf_zip_rdr_t *reader);

/* read some data. */
tsf_bool_t      tsf_zip_reader_read(void *reader,
                                    void *data,
                                    uint32_t len);

/* create a zipping writer.  it also does buffering.  accepts a writer
 * as a delegate to actually write the data.  as such, this thing can
 * adapt to many different underlying devices.  note that it makes little
 * or no sense to pass in a buffering writer, since this writer already
 * does buffering (in fact, it has to do buffering to interact with the
 * underlying compression engine).  returns an error if you try to use
 * a mode or attribute set that is not supported.  in particular,
 * TSF_ZIP_NONE is not supported! */
tsf_zip_wtr_t*  tsf_zip_writer_create(tsf_writer_t writer,
                                      void *writer_arg,
                                      const tsf_zip_wtr_attr_t *attr);

/* destroy the writer. */
void            tsf_zip_writer_destroy(tsf_zip_wtr_t *writer);

/* write some data. */
tsf_bool_t      tsf_zip_writer_write(void *writer,
                                     void *data,
                                     uint32_t len);

/* flush data.  calling this will almost certainly hurt compression, so
 * call it only if you really have to.  if the full_flush parameter is
 * true, the flush will not only be complete enough to allow the reading
 * end to get all of the data, but will also result in the compressor
 * resetting its state (so to speak) in order to make all future data
 * be independant of the already written data.  passing tsf_true for
 * full_flush will help error recovery but will hurt your compression
 * ratios even more.
 *
 * note that when using libbzip2, full_flush is ignored.  in general,
 * any flushing done with libbzip2 will hurt a lot more than any flushing
 * done with zlib, regardless of the value of full_flush.  so keep that
 * in mind.  in many cases, the extremely bad performance that results
 * from flushing in libbzip2 will be sufficient reason to use zlib
 * instead. */
tsf_bool_t      tsf_zip_writer_flush(tsf_zip_wtr_t *writer,
                                     tsf_bool_t full_flush);

/* finish the stream.  this allows the compressor to output its end-of-
 * stream marker.  this also does an implicit flush, so calling flush
 * just before calling finish will not help (it shouldn't hurt, either).
 * not that if you do a flush but do not do a finish, the receiver
 * should be able to get all of the data but will probably report an
 * error at the end of the file, complaining that the end of stream was
 * never really hit.
 *
 * after finish returns, you may not use this writer anymore.
 * attempting to call writer, flush, or finish on a writer that has
 * already had finish called on it will lead to unpredictable behavior.
 * after calling finish, you should just call destroy. */
tsf_bool_t      tsf_zip_writer_finish(tsf_zip_wtr_t *writer);

/* create an adaptive zipping reader.  this reader will be able to switch
 * between different zipping strategies (including no zipping) as the input
 * data indicates.  however, it only works under two non-trivial
 * conditions:
 * 1) the input is formatted according to the serial protocol with
 *    stream file headers indicating compression settings, and
 * 2) you call tsf_adaptive_reader_hint_switch() at the right times. */
tsf_adpt_rdr_t* tsf_adaptive_reader_create(tsf_partial_reader_t reader,
					   void *reader_arg,
					   uint32_t nozip_buf_size,
					   const tsf_zip_rdr_attr_t *attr);

/* close the adaptive reader.  cleans up everything.  never fails. */
void            tsf_adaptive_reader_destroy(tsf_adpt_rdr_t *reader);

/* inform the reader that there may be a format change in the next byte.
 * if we are currently reading with compression turned off, this hint tells
 * the reader to inspect the next four bytes to see if it contains a code
 * corresponding to the start of a ZLIB or BZIP2 sequence.  if we are using
 * a ZLIB or BZIP2 reader, this specifies that an end-of-stream from the
 * underlying reader is correct (otherwise it is flagged as erroneous). */
void            tsf_adaptive_reader_hint_switch(tsf_adpt_rdr_t *reader);

/* the actual read method. */
tsf_bool_t      tsf_adaptive_reader_read(void *reader,
					 void *data,
					 uint32_t len);

/* determine the zip mode of the given file by openning it and reading the
 * first four bytes.  if the mode returned is TSF_ZIP_UNKNOWN then there
 * was an error.  if the error was TSF_E_NOT_SUPPORTED then the file is
 * not in a valid TSF stream file format. */
tsf_zip_mode_t  tsf_stream_file_get_mode(const char *filename);

/* same as above except that it reads the first four bytes from the given
 * file descriptor. */
tsf_zip_mode_t  tsf_stream_file_get_mode_from_fd(int fd);

/* open a file for reading.  if the zip reader attributes are NULL then they
 * are set to the defaults.  if the limits are NULL then no limits are used. */
tsf_stream_file_input_t*
                tsf_stream_file_input_open(const char *filename,
                                           tsf_limits_t *limits,
                                           tsf_zip_rdr_attr_t *attr);

/* make a stream file input hijack a file descriptor.  if the zip reader
 * attributes are NULL then they are set to the defaults.
 *
 * if keep is true, the file descriptor will be closed when
 * tsf_stream_file_input_close() is called.  otherwise, the file
 * descriptor will remain open. */
tsf_stream_file_input_t*
                tsf_stream_file_input_fd_open(int fd,
                                              tsf_limits_t *limits,
                                              tsf_zip_rdr_attr_t *attr,
                                              tsf_bool_t keep);

/* close a file that was openned for reading. */
void            tsf_stream_file_input_close(tsf_stream_file_input_t *file);

/* returns true if the file is compressed.  compressed files have some
 * limitations.  for example, mark & seek is not supported for compressed
 * files. */
#define         tsf_stream_file_input_is_compressed(file) \
                    ((file)->use_zip)

/* register a callback that will get called every time the input manager
 * encounters a new type. */
#define         tsf_stream_file_input_register_type_cback(in,cback,arg) \
                    (tsf_serial_in_man_register_type_cback((in)->in_man, \
                                                           (cback),(arg)))

/* unregister all type callbacks that match the given function pointer
 * and void* argument. */
#define         tsf_stream_file_input_unregister_type_cback(in,cback,arg) \
                    (tsf_serial_in_man_unregister_type_cback((in)->in_man, \
                                                             (cback),(arg)))

/* read a buffer from a file that was openned for reading.  this operation
 * is thread-safe. */
tsf_buffer_t*   tsf_stream_file_input_read(tsf_stream_file_input_t *file);

/* get the mark of the thing you are about to read.  the mark is something
 * that you get to allocate.  but you must treat it as an opaque!
 *
 * note that mark & seek is not supported for files that are compressed.
 * it is also not supported if the underlying file descriptor does not
 * support lseek().  if the file is compressed, you will get a
 * TSF_E_NOT_SUPPORTED error.  it is perfectly safe (and fast) to determine
 * if a file is compressed by calling this function.  you can of course
 * also use tsf_stream_file_input_is_compressed().  if the file does not
 * support lseek(), you will get some sort of system error, which may
 * differ depending on how well your system complies to POSIX. :-) */
tsf_bool_t      tsf_stream_file_input_get_mark(tsf_stream_file_input_t *file,
                                               tsf_stream_file_mark_t *mark);

/* seek to previously acquired mark. */
tsf_bool_t      tsf_stream_file_input_seek(tsf_stream_file_input_t *file,
                                           const tsf_stream_file_mark_t *mark);

/* open a file for writing.  if the zip attributes are NULL, the default
 * ones will be used (which means no compression). */
tsf_stream_file_output_t*
                tsf_stream_file_output_open(const char *filename,
                                            tsf_bool_t append,
                                            tsf_zip_wtr_attr_t *attr);

/* create a file atomically.  neither this call, nor subsequent writes to
 * the file will make it appear at the requested filename.  however, flushing
 * or closing the file will make it appear.  this is great if you need to
 * replace a file, but you cannot allow those reading the file to see a
 * part of it without seeing the whole thing. */
tsf_stream_file_output_t*
                tsf_stream_file_output_create_atomic(const char *filename,
						     tsf_zip_wtr_attr_t *attr);

/* make a stream file output hijack a file descriptor.  if keep is true, the
 * file descriptor will be closed when tsf_stream_file_output_close() is
 * called.  otherwise, it will remain open. */
tsf_stream_file_output_t*
                tsf_stream_file_output_fd_open(int fd,
                                               tsf_zip_wtr_attr_t *attr,
                                               tsf_bool_t keep);

/* flush and close a stream file output.  returns tsf_true if the stream
 * was in a good state after the flush, tsf_false if it wasn't.  the stream
 * will be in a bad state if an error had ever occurred while writing or
 * flushing the stream.  either way, the file gets closed. */
tsf_bool_t      tsf_stream_file_output_close(tsf_stream_file_output_t *file);

/* close the stream file output without flushing. */
void            tsf_stream_file_output_destroy(tsf_stream_file_output_t *file);

/* flush the stream file output.  if this fails, the stream becomes corrupt
 * at the point of failure and should not be used. */
tsf_bool_t      tsf_stream_file_output_flush(tsf_stream_file_output_t *file);

/* write something into the file.  if this fails, the stream becomes corrupt
 * at the point of failure and should not be used. */
tsf_bool_t      tsf_stream_file_output_write(tsf_stream_file_output_t *file,
                                             tsf_buffer_t *buffer);

/* returns tsf_true if BDB dictionary files are supported, tsf_false
 * otherwise. */
tsf_bool_t      tsf_bdb_file_supported(void);

/* open a BDB dictionary file.  it will allow you to store key-value
 * pairs in a searchable way.  if BDB support was not compiled in, you
 * will get the TSF_E_NOT_SUPPORTED error.
 *
 * NOTE: BDB dictionary file support is EXPERIMENTAL!  Don't use it to
 * store anything important. */
tsf_bdb_file_t* tsf_bdb_file_open(const char *filename,
                                  tsf_bool_t update,
                                  tsf_bool_t truncate);

/* close a BDB dictionary file. */
void            tsf_bdb_file_close(tsf_bdb_file_t *file);

/* sync a BDB dictionary file.  all changes are flushed to disk.  if this
 * fails, you can safely assume that the file is corrupt. */
tsf_bool_t      tsf_bdb_file_sync(tsf_bdb_file_t *file);

/* put an entry into a BDB dictionary file. */
tsf_bool_t      tsf_bdb_file_put(tsf_bdb_file_t *file,
                                 tsf_buffer_t *key,
                                 tsf_buffer_t *val);

/* retrieve an entry from a BDB dictionary file.  if the key does not
 * exist in the dictionary, you will get the TSF_E_ELE_NOT_FOUND error. */
tsf_buffer_t*   tsf_bdb_file_get(tsf_bdb_file_t *file,
                                 tsf_buffer_t *key);

/* remove an entry from a BDB dictionary file.  if the key does not
 * exist in the dictionary, you will get the TSF_E_ELE_NOT_FOUND error.
 * after removal, returns the old value. */
tsf_buffer_t*   tsf_bdb_file_remove(tsf_bdb_file_t *file,
                                    tsf_buffer_t *key);

/* determine if the file has the given type. */
#define         tsf_bdb_file_has_type(file,type) \
                    (tsf_ra_type_man_has_type((file)->types,type))

/* determine if the file has all of the types used by the given buffer. */
#define         tsf_bdb_file_has_buf_types(file,buf) \
                    (tsf_ra_type_man_has_buf_types((file)->types,(buf)))

/* calculates [sic] the size of the type code free list.  by 'calculates'
 * I mean that this is not an O(1) operation. */
#define         tsf_bdb_file_get_type_code_free_list_size(file)	  \
                    (tsf_ra_type_man_get_type_code_free_list_size( \
                         (file)->types))

/* FIXME: add tsf_define helpers for dealing with primitive generators,
   parsers, and copiers, as well as caching array-of-primitive types.

   ... this may not be needed ... */

#ifdef __cplusplus
};
#endif

#endif

