/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_bdb_file.c,v 1.1.1.1 2007/05/02 23:31:06 pizlo Exp $
 *
 * tsf_bdb_file.c -- support for Berkeley DB dictionary files.
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "tsf_internal.h"

#ifdef HAVE_BDB

#include "tsf_bdb_protocol.h"
#include "tsf_format.h"

#include <db.h>
#include <string.h>

#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#define Cdb(x) ((DB*)((x)->db))
#define Cdbc(x) ((DBC*)((x)->dbc))

static tsf_bool_t set_dbt_from_buffer(DBT *thang,
                                      tsf_buffer_t *buf,
                                      tsf_bdb_file_t *file) {
    uint8_t *cur;
    uint32_t type_code;

    memset(thang,0,sizeof(DBT));
    
    thang->size=(tsf_type_is_dynamic(buf->type)?
                 4+4*tsf_type_in_map_get_num_types(buf->types):0)+
                buf->size+4;
    
    thang->data=malloc(thang->size);
    if (thang->data==NULL) {
        tsf_set_errno("Could not malloc DBT buffer");
        return tsf_false;
    }
    
    cur=(uint8_t*)thang->data;
    
    type_code=tsf_ra_type_man_get_type_code(file->types,
					    buf->type);
    copy_htonl_incdst(cur,(uint8_t*)&type_code);
    
    if (tsf_type_is_dynamic(buf->type)) {
        uint32_t num,i;
        num=tsf_type_in_map_get_num_types(buf->types);
        copy_htonl_incdst(cur,(uint8_t*)&num);
        for (i=0;i<num;++i) {
            type_code=tsf_ra_type_man_get_type_code
		(file->types,
		 tsf_type_in_map_get_type(buf->types,i));
            copy_htonl_incdst(cur,(uint8_t*)&type_code);
        }
    }
    
    memcpy(cur,buf->data,buf->size);
                       
    return tsf_true;
}

static tsf_buffer_t *get_buffer_from_dbt(DBT *thang,
                                         tsf_bdb_file_t *file) {
    uint8_t *cur=(uint8_t*)thang->data,
            *end=cur+thang->size;
    
    uint32_t type_code;
    
    tsf_type_t *type;
    tsf_type_in_map_t *types;
    uint8_t *data;
    
    copy_ntohl_incsrc_bc((uint8_t*)&type_code,
                         cur,end,phase1_bounds_error);
    
    type=tsf_ra_type_man_find(file->types,type_code);
    if (type==NULL) {
        tsf_set_error(TSF_E_PARSE_ERROR,
                      "Type code " fui32 " not recognized",
                      type_code);
        return NULL;
    }
    
    if (tsf_type_is_dynamic(type)) {
        uint32_t num_types,
                 index;
        
        types=tsf_type_in_map_create();
        if (types==NULL) {
            return NULL;
        }
        
        copy_ntohl_incsrc_bc((uint8_t*)&num_types,
                             cur,end,phase2_bounds_error);
        
        tsf_type_in_map_prepare(types,num_types);
        
        while (num_types-->0) {
            tsf_type_t *cur_type;
            
            copy_ntohl_incsrc_bc((uint8_t*)&index,
                                 cur,end,phase2_bounds_error);
            
            cur_type=tsf_ra_type_man_find(file->types,index);
            if (cur_type==NULL) {
                tsf_set_error(TSF_E_PARSE_ERROR,
                              "Type code " fui32 " not recognized",
                              index);
                tsf_type_in_map_destroy(types);
                return NULL;
            }
            
            if (!tsf_type_in_map_append(types,cur_type)) {
                tsf_type_in_map_destroy(types);
                return NULL;
            }
        }
    } else {
        types=tsf_type_in_map_get_empty_singleton();
    }
    
    data=malloc(end-cur);
    if (data==NULL) {
        tsf_set_errno("Could not malloc data for buffer");
        tsf_type_in_map_destroy(types);
        return NULL;
    }
    
    memcpy(data,cur,end-cur);
    
    return tsf_buffer_create(type,
                             types,
                             data,
                             end-cur,
                             tsf_true);

phase2_bounds_error:
    tsf_type_in_map_destroy(types);
phase1_bounds_error:
    tsf_set_error(TSF_E_PARSE_ERROR,
                  "Bounds error in get_buffer_from_dbt()");
    return NULL;
}

static tsf_bool_t write_types(tsf_bdb_file_t *file) {
    DBT key,val;
    uint32_t size;
    uint8_t *cur;
    int res;
    
    memset(&key,0,sizeof(key));
    memset(&val,0,sizeof(val));
    
    key.data=TSF_BP_TYPES;
    key.size=strlen(TSF_BP_TYPES);
    
    /* must first calculate the size of the value thang */
    size=0;
    if (!tsf_ra_type_man_write(file->types,
			       tsf_size_calc_writer,
			       &size)) {
	return tsf_false;
    }

    val.size=size;
    val.data=cur=malloc(size);
    
    if (val.data==NULL) {
        tsf_set_errno("Could not malloc data for DBT");
        return tsf_false;
    }
    
    if (!tsf_ra_type_man_write(file->types,
			       tsf_buffer_only_writer,
			       &cur)) {
	free(val.data);
	return tsf_false;
    }

    res=Cdb(file)->put(Cdb(file),
                       NULL,
                       &key,
                       &val,
                       0);
    
    free(val.data);
    
    if (res!=0) {
        tsf_set_bdb_error(res,"Could not DB->put() the TSF_SP_TYPES record");
        return tsf_false;
    }
    
    return tsf_true;
}

static tsf_bool_t read_types(tsf_bdb_file_t *file) {
    DBT key,val;
    int res;

    memset(&key,0,sizeof(key));
    memset(&val,0,sizeof(val));
    
    key.data=TSF_BP_TYPES;
    key.size=strlen(TSF_BP_TYPES);
    
    res=Cdb(file)->get(Cdb(file),
                       NULL,
                       &key,
                       &val,
                       0);
    
    if (res==0) {
        tsf_light_buffer_t s;
        
        s.cur=val.data;
        s.end=s.cur+val.size;
	
	file->types=tsf_ra_type_man_read(tsf_buffer_only_reader,
					 &s);
	if (file->types==NULL) {
	    return tsf_false;
	}
    } else if (res==DB_NOTFOUND) {
	file->types=tsf_ra_type_man_create();
	if (file->types==NULL) {
	    return tsf_false;
	}

        if (file->update) {
	    if (!write_types(file)) {
		tsf_ra_type_man_destroy(file->types);
		return tsf_false;
	    }
        }
    } else {
        tsf_set_bdb_error(res,"Could not DB->get() the TSF_BP_TYPES record");
        return tsf_false;
    }
    
    return tsf_true;
}

tsf_bdb_file_t *tsf_bdb_file_open(const char *filename,
                                  tsf_bool_t update,
                                  tsf_bool_t truncate) {
    int res;
    tsf_bdb_file_t *ret;
    int minor,major,patch;
    
    /* check bdb version */
    db_version(&major,&minor,&patch);
    if (major != BDB_VERSION_MAJOR ||
        minor != BDB_VERSION_MINOR) {
        tsf_set_error(TSF_E_INTERNAL,
                      "Expected BDB library version %d.%d.%d but got "
                      "%d.%d.%d, which differs by more than just the "
                      "patch number.  If you install a new version of "
                      "BDB, please recompile the TSF library to avoid "
                      "seeing this error.",
                      BDB_VERSION_MAJOR,
                      BDB_VERSION_MINOR,
                      BDB_VERSION_PATCH,
                      major,
                      minor,
                      patch);
        return NULL;
    }
    
    if (truncate && !update) {
        tsf_set_error(TSF_E_INVALID_ARG,
                      "Cannot specify the truncate without the update "
                      "flag");
        return NULL;
    }
    
    ret=malloc(sizeof(tsf_bdb_file_t));
    if (ret==NULL) {
        tsf_set_errno("Could not malloc tsf_bdb_file_t");
        return NULL;
    }
    
    ret->update=update;
    
    if ((res=db_create((DB**)&ret->db,NULL,0))!=0) {
        tsf_set_bdb_error(res,"Could not db_create()");
        free(ret);
        return NULL;
    }
    
    /* may want to do crazy loop like in tsf_stream_file_output_open()
     * so that we know if we created the file or not. */
    
    if ((res=Cdb(ret)->open(Cdb(ret),
#if (BDB_VERSION_MAJOR == 4 && BDB_VERSION_MINOR > 0) ||\
    BDB_VERSION_MAJOR > 4
                            NULL,
#endif
                            filename,
                            NULL,
                            DB_BTREE,
                            (update?DB_CREATE:0)|
                            (update?0:DB_RDONLY)|
                            (truncate?DB_TRUNCATE:0),
                            0644))!=0) {
        tsf_set_bdb_error(res,"Could not DB->open() for %s",
                              filename);
        Cdb(ret)->close(Cdb(ret),0);
        free(ret);
        return NULL;
    }
    
    if (!read_types(ret)) {
        Cdb(ret)->close(Cdb(ret),0);
        free(ret);
        return NULL;
    }
    
    return ret;
}

void tsf_bdb_file_close(tsf_bdb_file_t *file) {
    if (!file->update) {
        write_types(file);
    }
    Cdb(file)->close(Cdb(file),0);
    tsf_ra_type_man_destroy(file->types);
    free(file);
}

tsf_bool_t tsf_bdb_file_sync(tsf_bdb_file_t *file) {
    int res;
    
    if (!file->update) {
        return tsf_true;
    }
    
    if (!write_types(file)) {
        return tsf_false;
    }
    
    if ((res=Cdb(file)->sync(Cdb(file),0))!=0) {
        tsf_set_bdb_error(res,"Could not DB->sync()");
        return tsf_false;
    }
    
    return tsf_true;
}

tsf_bool_t tsf_bdb_file_put(tsf_bdb_file_t *file,
                            tsf_buffer_t *key,
                            tsf_buffer_t *val) {
    /* things to do here:
     * 1) must get references for types in key and val
     * 2) must produce DBT's for both key and val
     *    2.1) if we fail, must remove references for types in key and val
     * 3) must call DB->put() with NOOVERWRITE
     *    3.1) if we fail, must remove references for types in key and val
     *         and must free up DBT's for key and val
     *    3.2) if we get KEYEXIST, then we must:
     *         3.2.1) remove the references for the key
     *         3.2.2) remove the references for the old value
     *         3.2.3) try again without NOOVERWRITE
     * 4) must free up DBT's for key and val
     */
    
    tsf_buffer_t *oldval;
    DBT keyt,valt,oldvalt;
    int res;
    
    /* 1 */
    
    if (!tsf_ra_type_man_get_buf_references(file->types,key)) {
        return tsf_false;
    }
    
    if (!tsf_ra_type_man_get_buf_references(file->types,val)) {
        tsf_ra_type_man_remove_buf_references(file->types,key);
        return tsf_false;
    }
    
    /* 2 */
    
    if (!set_dbt_from_buffer(&keyt,
                             key,
                             file)) {
        /* 2.1 */
        tsf_ra_type_man_remove_buf_references(file->types,key);
        tsf_ra_type_man_remove_buf_references(file->types,val);
        return tsf_false;
    }
    
    if (!set_dbt_from_buffer(&valt,
                             val,
                             file)) {
        /* 2.1 */
        tsf_ra_type_man_remove_buf_references(file->types,key);
        tsf_ra_type_man_remove_buf_references(file->types,val);
        free(keyt.data);
        return tsf_false;
    }
    
    /* 3 */
    
    if ((res=Cdb(file)->put(Cdb(file),
                            NULL,
                            &keyt,
                            &valt,
                            DB_NOOVERWRITE))!=0) {
        if (res!=DB_KEYEXIST) {
            /* 3.1 */
            tsf_set_bdb_error(res,
                "Could not DB->put() for a TSF buffer record");
            tsf_ra_type_man_remove_buf_references(file->types,key);
            tsf_ra_type_man_remove_buf_references(file->types,val);
            free(keyt.data);
            free(valt.data);
            return tsf_false;
        }
        
        /* 3.2.1 */
        tsf_ra_type_man_remove_buf_references(file->types,key);
        
        /* 3.2.2 */
        memset(&oldvalt,0,sizeof(oldvalt));
        
        if ((res=Cdb(file)->get(Cdb(file),
                                NULL,
                                &keyt,
                                &oldvalt,
                                0))!=0) {
            tsf_set_bdb_error(res,
                "Could not DB->get() an old (about to be overwritten) "
                "TSF buffer record");
            tsf_ra_type_man_remove_buf_references(file->types,val);
            free(keyt.data);
            free(valt.data);
            return tsf_false;
        }
        
        oldval=get_buffer_from_dbt(&oldvalt,file);
        if (oldval==NULL) {
            tsf_set_bdb_error(res,
                "Could not DB->get() an old (about to be overwritten) "
                "TSF buffer record");
            tsf_ra_type_man_remove_buf_references(file->types,val);
            free(keyt.data);
            free(valt.data);
            return tsf_false;
        }
        
        tsf_ra_type_man_remove_buf_references(file->types,oldval);
        tsf_buffer_destroy(oldval);
        
        /* 3.2.3 */
        if ((res=Cdb(file)->put(Cdb(file),
                                NULL,
                                &keyt,
                                &valt,
                                0))!=0) {
            tsf_set_bdb_error(res,
                "Could not DB->put() for a TSF buffer record (2nd attempt)");
            tsf_ra_type_man_remove_buf_references(file->types,val);
            free(keyt.data);
            free(valt.data);
            return tsf_false;
        }
    }
    
    /* 4 */

    free(keyt.data);
    free(valt.data);
    
    return tsf_true;
}

static tsf_bool_t prepare_key(tsf_bdb_file_t *file,
                              DBT *thang,
                              tsf_buffer_t *key) {
    if (!tsf_bdb_file_has_buf_types(file,key)) {
        tsf_set_error(TSF_E_ELE_NOT_FOUND,
                      "Item known not to be in the file because the key "
                      "contained types not in the type code mapping");
        return tsf_false;
    }
    
    if (!set_dbt_from_buffer(thang,
                             key,
                             file)) {
        return tsf_false;
    }
    
    return tsf_true;
}

tsf_buffer_t *tsf_bdb_file_get(tsf_bdb_file_t *file,
                               tsf_buffer_t *key) {
    DBT keyt,valt;
    int res;
    
    if (!prepare_key(file,&keyt,key)) {
        return NULL;
    }
    
    memset(&valt,0,sizeof(valt));
    
    res=Cdb(file)->get(Cdb(file),
                       NULL,
                       &keyt,
                       &valt,
                       0);
    
    free(keyt.data);
    
    if (res!=0) {
        if (res==DB_NOTFOUND) {
            tsf_set_error(TSF_E_ELE_NOT_FOUND,
                          "A search using DB->get() returned DB_NOTFOUND");
        } else {
            tsf_set_bdb_error(res,
                "Could not DB->get() for a TSF buffer record");
        }
        return NULL;
    }
    
    return get_buffer_from_dbt(&valt,file);
}

tsf_buffer_t *tsf_bdb_file_remove(tsf_bdb_file_t *file,
                                  tsf_buffer_t *key) {
    DBT keyt,valt;
    tsf_buffer_t *val;
    int res;
    
    if (!prepare_key(file,&keyt,key)) {
        return NULL;
    }
    
    memset(&valt,0,sizeof(valt));
    
    if ((res=Cdb(file)->get(Cdb(file),
                            NULL,
                            &keyt,
                            &valt,
                            0))!=0) {
        if (res==DB_NOTFOUND) {
            tsf_set_error(TSF_E_ELE_NOT_FOUND,
                          "A search using DB->get() returned DB_NOTFOUND");
        } else {
            tsf_set_bdb_error(res,
                "Could not DB->get() for a TSF buffer record");
        }
        free(keyt.data);
        return NULL;
    }
    
    val=get_buffer_from_dbt(&valt,file);
    if (val==NULL) {
        free(keyt.data);
        return NULL;
    }
    
    if ((res=Cdb(file)->del(Cdb(file),
                            NULL,
                            &keyt,
                            0))!=0) {
        if (res==DB_NOTFOUND) {
            tsf_set_error(TSF_E_INTERNAL,
                          "An attempt to delete a record using DB->del() "
                          "returned DB_NOTFOUND even though a search using "
                          "DB->get() performed immediately prior had "
                          "succeeded.  This implies a synchronization error.");
        } else {
            tsf_set_bdb_error(res,
                "Could not DB->del() for a TSF buffer record");
        }
        free(keyt.data);
        return NULL;
    }
    
    tsf_ra_type_man_remove_buf_references(file->types,key);
    tsf_ra_type_man_remove_buf_references(file->types,val);
    free(keyt.data);
    
    return val;
}

tsf_bool_t tsf_bdb_file_supported() {
    return tsf_true;
}

#else

tsf_bdb_file_t *tsf_bdb_file_open(const char *filename,
                                  tsf_bool_t update,
                                  tsf_bool_t truncate) {
    tsf_set_error(TSF_E_NOT_SUPPORTED,
                  "Berkeley DB support was not configured into this "
                  "TSF library");
    return NULL;
}

void tsf_bdb_file_close(tsf_bdb_file_t *file) {
    tsf_abort("Attempt to call tsf_bdb_file_close() when "
              "tsf_bdb_file_open() could not have possibly succeeded.");
}

tsf_bool_t tsf_bdb_file_sync(tsf_bdb_file_t *file) {
    tsf_abort("Attempt to call tsf_bdb_file_sync() when "
              "tsf_bdb_file_open() could not have possibly succeeded.");
}

tsf_bool_t tsf_bdb_file_put(tsf_bdb_file_t *file,
                            tsf_buffer_t *key,
                            tsf_buffer_t *val) {
    tsf_abort("Attempt to call tsf_bdb_file_put() when "
              "tsf_bdb_file_open() could not have possibly succeeded.");
}

tsf_buffer_t *tsf_bdb_file_get(tsf_bdb_file_t *file,
                               tsf_buffer_t *key) {
    tsf_abort("Attempt to call tsf_bdb_file_get() when "
              "tsf_bdb_file_open() could not have possibly succeeded.");
}

tsf_buffer_t *tsf_bdb_file_remove(tsf_bdb_file_t *file,
                                  tsf_buffer_t *key) {
    tsf_abort("Attempt to call tsf_bdb_file_remove() when "
              "tsf_bdb_file_open() could not have possibly succeeded.");
}

tsf_bool_t tsf_bdb_file_supported() {
    return tsf_false;
}

#endif


