/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_buffer.c,v 1.2 2007/07/08 08:15:01 pizlo Exp $
 *
 * tsf_buffer.c -- buffer code
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "tsf_internal.h"
#include "tsf_format.h"

#ifdef HAVE_PTHREAD
#include <pthread.h>
#endif

#ifdef DEBUG_BUFFER_ALLOC
static int64_t buffer_balance = 0;

int64_t tsf_buffer_get_balance() {
    return buffer_balance;
}

#define increment_buffer_balance() do {\
    buffer_balance++;\
} while(0)

#define decrement_buffer_balance() do {\
    buffer_balance--;\
} while(0)

#else

#define increment_buffer_balance() do {\
} while(0)

#define decrement_buffer_balance() do {\
} while(0)

#endif

static tsf_buffer_t empty_buffer;

static void init_empty_buffer(void) {
    empty_buffer.region=NULL;
    empty_buffer.type=tsf_type_create(TSF_TK_VOID);
    empty_buffer.types=tsf_type_in_map_get_empty_singleton();
    empty_buffer.data=NULL;
    empty_buffer.size=0;
    empty_buffer.keep=tsf_false;
}

#ifdef HAVE_PTHREAD
static pthread_once_t once_control=PTHREAD_ONCE_INIT;
#else
static tsf_bool_t once_control=tsf_false;
#endif

static void check_init_empty_buffer(void) {
#ifdef HAVE_PTHREAD
    /* FIXME: pthread_once can be expensive on some platforms.  And
       here we really need it to be fast. */
    pthread_once(&once_control,init_empty_buffer);
#else
    if (!once_control) {
	init_empty_buffer();
	once_control=tsf_true;
    }
#endif
}

tsf_buffer_t *tsf_buffer_get_empty_singleton(void) {
    check_init_empty_buffer();
    return &empty_buffer;
}

tsf_buffer_t *tsf_buffer_create(tsf_type_t *type,
                                tsf_type_in_map_t *types,
                                void *data,
                                uint32_t size,
                                tsf_bool_t keep) {
    tsf_buffer_t *ret=malloc(sizeof(tsf_buffer_t));
    if (ret==NULL) {
        tsf_set_errno("Could not allocate tsf_buffer_t struct.");
        if (keep) {
            free(data);
        }
        tsf_type_in_map_destroy(types);
        return NULL;
    }
    
    ret->region=NULL;
    ret->type=tsf_type_dup(type);
    ret->types=types;
    ret->data=data;
    ret->size=size;
    ret->keep=keep;
    
    ret->ref_count=1;

#ifdef HAVE_PTHREAD
    pthread_mutex_init(&ret->ref_count_lock,NULL);
#endif
    
    increment_buffer_balance();
    
    return ret;
}

tsf_buffer_t *tsf_buffer_create_in(tsf_type_t *type,
                                   tsf_type_in_map_t *types,
                                   void *data,
                                   uint32_t size,
                                   void *region) {
    tsf_buffer_t *ret=tsf_region_alloc(region,sizeof(tsf_buffer_t));
    if (ret==NULL) {
        tsf_set_errno("Could not tsf_region_alloc tsf_buffer_t struct.");
        return NULL;
    }
    
    ret->region=region;
    
    if (!tsf_region_add_cback(region,
                              (tsf_region_cback_t)tsf_type_destroy,
                              type)) {
        tsf_set_errno("Could not tsf_region_add_cback for "
                      "tsf_type_destroy");
        return NULL;
    }
    
    ret->type=tsf_type_dup(type);
    ret->types=types;
    ret->data=data;
    ret->size=size;
    ret->keep=tsf_false;
    
    return ret;
}

tsf_buffer_t *tsf_buffer_dup(tsf_buffer_t *buf) {
    if (buf==&empty_buffer) {
	return buf;
    }
    
    if (buf->region!=NULL || !buf->keep) {
        return tsf_buffer_clone(buf);
    }
    
#ifdef HAVE_PTHREAD
    pthread_mutex_lock(&buf->ref_count_lock);
#endif
    buf->ref_count++;
#ifdef HAVE_PTHREAD
    pthread_mutex_unlock(&buf->ref_count_lock);
#endif
    
    increment_buffer_balance();
    
    return buf;
}

tsf_buffer_t *tsf_buffer_clone(tsf_buffer_t *buf) {
    tsf_buffer_t *ret=malloc(sizeof(tsf_buffer_t));
    if (ret==NULL) {
        tsf_set_errno("Could not malloc tsf_buffer_t");
        return NULL;
    }
    
    ret->region=NULL;
    ret->size=buf->size;
    ret->data=malloc(buf->size);
    if (ret->data==NULL) {
        free(ret);
        tsf_set_errno("Could not malloc buffer");
        return NULL;
    }
    
    memcpy(ret->data,buf->data,ret->size);
    
    ret->type=tsf_type_dup(buf->type);
    ret->types=tsf_type_in_map_clone(buf->types);
    if (ret->types==NULL) {
        tsf_type_destroy(ret->type);
        free(ret->data);
        free(ret);
        return NULL;
    }
    
    ret->keep=tsf_true;
    
    ret->ref_count=1;

#ifdef HAVE_PTHREAD
    pthread_mutex_init(&ret->ref_count_lock,NULL);
#endif
    
    increment_buffer_balance();
    
    return ret;
}

tsf_buffer_t *tsf_buffer_clone_in(tsf_buffer_t *buf,
                                  void *region) {
    tsf_buffer_t *ret=tsf_region_alloc(region,sizeof(tsf_buffer_t));
    if (ret==NULL) {
        tsf_set_errno("Could not tsf_region_alloc tsf_buffer_t");
        return NULL;
    }
    
    ret->size=buf->size;
    ret->data=tsf_region_alloc(region,buf->size);
    if (ret->data==NULL) {
        tsf_set_errno("Could not tsf_region_alloc buffer");
        return NULL;
    }
    
    memcpy(ret->data,buf->data,ret->size);

    if (!tsf_region_add_cback(region,
                              (tsf_region_cback_t)tsf_type_destroy,
                              buf->type)) {
        tsf_set_errno("Could not tsf_region_add_cback for "
                      "tsf_type_destroy");
        return NULL;
    }
    
    ret->type=tsf_type_dup(buf->type);
    ret->types=tsf_type_in_map_clone_in(buf->types,region);
    if (ret->types==NULL) {
        return NULL;
    }
    
    ret->keep=tsf_false;
    
    return ret;
}

void tsf_buffer_destroy(tsf_buffer_t *buf) {
    if (buf==&empty_buffer) {
	return;
    }
    
    tsf_assert(buf->region==NULL);
    
    decrement_buffer_balance();
    
#ifdef HAVE_PTHREAD
    pthread_mutex_lock(&buf->ref_count_lock);
#endif

    if (--buf->ref_count>0) {
#ifdef HAVE_PTHREAD
        pthread_mutex_unlock(&buf->ref_count_lock);
#endif
        return;
    }
#ifdef HAVE_PTHREAD
    pthread_mutex_unlock(&buf->ref_count_lock);
#endif
    
    if (buf->keep) {
        free(buf->data);
    }
    
#ifdef HAVE_PTHREAD
    pthread_mutex_destroy(&buf->ref_count_lock);
#endif

    tsf_type_destroy(buf->type);
    tsf_type_in_map_destroy(buf->types);
    free(buf);
}

tsf_buffer_t *tsf_buffer_read(tsf_type_in_map_t *enclosing_types,
                              tsf_reader_t reader,
                              void *arg,
                              tsf_limits_t *limits) {
    uint32_t type_code;
    tsf_buffer_t *ret=malloc(sizeof(tsf_buffer_t));
    if (ret==NULL) {
        tsf_set_errno("Could not malloc tsf_buffer_t");
        return NULL;
    }
    
    ret->keep=tsf_true;
    ret->region=NULL;
    
    if (!reader(arg,&type_code,sizeof(type_code))) {
        free(ret);
        return NULL;
    }
    
    type_code=ntohl(type_code);
    
    if (type_code>=tsf_type_in_map_get_num_types(enclosing_types)) {
        tsf_set_error(TSF_E_PARSE_ERROR,
                      "Type code is outside of bounds of type map: "
                      fui32 " is the type code, while the type map "
                      "has a size of " fui32,
                      type_code,
                      tsf_type_in_map_get_num_types(enclosing_types));
        free(ret);
        return NULL;
    }
    
    ret->type=tsf_type_dup(
        tsf_type_in_map_get_type(enclosing_types,type_code));
    
    if (tsf_type_is_dynamic(ret->type)) {
        /* must read type translation table. */
        uint32_t num_types;
        uint32_t index;
        
        ret->types=tsf_type_in_map_create();
        if (ret->types==NULL) {
            tsf_type_destroy(ret->type);
            free(ret);
            return NULL;
        }
        
        if (!reader(arg,&num_types,sizeof(num_types))) {
            tsf_type_in_map_destroy(ret->types);
            tsf_type_destroy(ret->type);
            free(ret);
            return NULL;
        }
        
        num_types=ntohl(num_types);
        
        tsf_type_in_map_prepare(ret->types,num_types);
        
        while (num_types-->0) {
            if (!reader(arg,&index,sizeof(index))) {
                tsf_type_in_map_destroy(ret->types);
                tsf_type_destroy(ret->type);
                free(ret);
                return NULL;
            }
            
            index=ntohl(index);
            
            if (index>=tsf_type_in_map_get_num_types(enclosing_types)) {
                tsf_set_error(TSF_E_PARSE_ERROR,
                              "Index " fui32 " into enclosing type table "
                              "is outside of bounds; the enclosing table "
                              "has a size of " fui32,
                              index,
                              tsf_type_in_map_get_num_types(enclosing_types));
                tsf_type_in_map_destroy(ret->types);
                tsf_type_destroy(ret->type);
                free(ret);
                return NULL;
            }
            
            if (!tsf_type_in_map_append(
                     ret->types,
                     tsf_type_in_map_get_type(enclosing_types,index))) {
                tsf_type_in_map_destroy(ret->types);
                tsf_type_destroy(ret->type);
                free(ret);
                return NULL;
            }
        }
    } else {
        ret->types=tsf_type_in_map_get_empty_singleton();
    }

    if (!reader(arg,&ret->size,sizeof(ret->size))) {
        tsf_type_in_map_destroy(ret->types);
        tsf_type_destroy(ret->type);
        free(ret);
        return NULL;
    }
    
    ret->size=ntohl(ret->size);
    
    if (ret->size>tsf_limits_max_buf_size(limits)) {
        tsf_set_error(TSF_E_PARSE_ERROR,
                      "Buffer size exceeds limit");
        tsf_type_in_map_destroy(ret->types);
        tsf_type_destroy(ret->type);
        free(ret);
        return NULL;
    }
    
    ret->data=malloc(ret->size);
    if (ret->data==NULL) {
        tsf_set_errno("Could not allocate buffer data.");
        tsf_type_in_map_destroy(ret->types);
        tsf_type_destroy(ret->type);
        free(ret);
        return NULL;
    }
    
    if (!reader(arg,ret->data,ret->size)) {
        free(ret->data);
        tsf_type_in_map_destroy(ret->types);
        tsf_type_destroy(ret->type);
        free(ret);
        return NULL;
    }
    
    ret->ref_count=1;
    
#ifdef HAVE_PTHREAD
    pthread_mutex_init(&ret->ref_count_lock,NULL);
#endif
    
    increment_buffer_balance();
    
    return ret;
}

tsf_buffer_t *tsf_buffer_read_from_buf(tsf_type_in_map_t *enclosing_types,
                                       uint8_t **cur,
                                       uint8_t *end,
                                       void *region) {
    uint32_t type_code;
    tsf_type_t *type;
    tsf_buffer_t *ret=tsf_cond_alloc(region,sizeof(tsf_buffer_t));
    if (ret==NULL) {
        tsf_set_errno("Could not tsf_cond_alloc tsf_buffer_t");
        return NULL;
    }
    
    ret->keep=(region==NULL?tsf_true:tsf_false);
    ret->region=region;
    
    copy_ntohl_incsrc_bc((uint8_t*)&type_code,*cur,end,phase1_bounds_error);
    
    if (type_code>=tsf_type_in_map_get_num_types(enclosing_types)) {
        tsf_set_error(TSF_E_PARSE_ERROR,
                      "Type code is outside of bounds of type map: "
                      fui32 " is the type code, while the type map "
                      "has a size of " fui32,
                      type_code,
                      tsf_type_in_map_get_num_types(enclosing_types));
        if (region==NULL) {
            free(ret);
        }
        return NULL;
    }
    
    type=tsf_type_in_map_get_type(enclosing_types,type_code);
    
    if (region!=NULL) {
        if (!tsf_region_add_cback(region,
                                  (tsf_region_cback_t)tsf_type_destroy,
                                  type)) {
            tsf_set_errno("Could not tsf_region_add_cback for "
                          "tsf_type_destroy");
            return NULL;
        }
    }
    
    ret->type=tsf_type_dup(type);
    
    if (tsf_type_is_dynamic(ret->type)) {
        /* must read type translation table. */
        uint32_t num_types;
        uint32_t index;
        
        if (region==NULL) {
            ret->types=tsf_type_in_map_create();
            if (ret->types==NULL) {
                tsf_type_destroy(ret->type);
                free(ret);
                return NULL;
            }
        } else {
            ret->types=tsf_type_in_map_create_in(region);
            if (ret->types==NULL) {
                return NULL;
            }
        }
        
        copy_ntohl_incsrc_bc((uint8_t*)&num_types,
                             *cur,end,phase2_bounds_error);
        
        tsf_type_in_map_prepare(ret->types,num_types);
        
        while (num_types-->0) {
            copy_ntohl_incsrc_bc((uint8_t*)&index,
                                 *cur,end,phase2_bounds_error);
            
            if (index>=tsf_type_in_map_get_num_types(enclosing_types)) {
                tsf_set_error(TSF_E_PARSE_ERROR,
                              "Index " fui32 " into enclosing type table "
                              "is outside of bounds; the enclosing table "
                              "has a size of " fui32,
                              index,
                              tsf_type_in_map_get_num_types(enclosing_types));
                if (region==NULL) {
                    tsf_type_in_map_destroy(ret->types);
                    tsf_type_destroy(ret->type);
                    free(ret);
                }
                return NULL;
            }
            
            if (!tsf_type_in_map_append(
                     ret->types,
                     tsf_type_in_map_get_type(enclosing_types,index))) {
                if (region==NULL) {
                    tsf_type_in_map_destroy(ret->types);
                    tsf_type_destroy(ret->type);
                    free(ret);
                }
                return NULL;
            }
        }
    } else {
        ret->types=tsf_type_in_map_get_empty_singleton();
    }
    
    copy_ntohl_incsrc_bc((uint8_t*)&ret->size,
                         *cur,end,phase2_bounds_error);
    
    if (ret->size + *cur > end) {
        goto phase2_bounds_error;
    }
    
    ret->data=tsf_cond_alloc(region,ret->size);
    if (ret->data==NULL) {
        if (region==NULL) {
            tsf_type_in_map_destroy(ret->types);
            tsf_type_destroy(ret->type);
            free(ret);
        }
        tsf_set_errno("Could not tsf_cond_alloc buffer");
        return NULL;
    }
    
    memcpy(ret->data,*cur,ret->size);
    (*cur)+=ret->size;
    
    if (region==NULL) {
        ret->ref_count=1;
    
#ifdef HAVE_PTHREAD
        pthread_mutex_init(&ret->ref_count_lock,NULL);
#endif

        increment_buffer_balance();
    }
    
    return ret;

phase2_bounds_error:
    if (region==NULL) {
        tsf_type_destroy(ret->type);
        tsf_type_in_map_destroy(ret->types);
    }
phase1_bounds_error:
    if (region==NULL) {
        free(ret);
    }
    tsf_set_error(TSF_E_PARSE_ERROR,
                  "Bounds error in tsf_buffer_read_from_buf()");
    return NULL;
}

tsf_bool_t tsf_buffer_skip_in_buf(tsf_type_in_map_t *enclosing_types,
                                  uint8_t **cur,
                                  uint8_t *end) {
    uint32_t type_code,size;
    
    copy_ntohl_incsrc_bc((uint8_t*)&type_code,*cur,end,bounds_error);
    
    if (type_code>=tsf_type_in_map_get_num_types(enclosing_types)) {
        tsf_set_error(TSF_E_PARSE_ERROR,
                      "Type code is outside of bounds of type map: "
                      fui32 " is the type code, while the type map "
                      "has a size of " fui32,
                      type_code,
                      tsf_type_in_map_get_num_types(enclosing_types));
        return tsf_false;
    }
    
    if (tsf_type_is_dynamic(
            tsf_type_in_map_get_type(enclosing_types,type_code))) {
        /* must skip type translation table. */
        uint32_t num_types;
        
        copy_ntohl_incsrc_bc((uint8_t*)&num_types,*cur,end,bounds_error);
        
        (*cur)+=num_types*4;
    }

    copy_ntohl_incsrc_bc((uint8_t*)&size,*cur,end,bounds_error);
    
    if (size + *cur > end) {
        goto bounds_error;
    }
    
    (*cur)+=size;
    
    return tsf_true;
    
bounds_error:
    tsf_set_error(TSF_E_PARSE_ERROR,
                  "Bounds error in tsf_buffer_skip_in_buf()");
    return tsf_false;
}

tsf_bool_t tsf_buffer_write(tsf_buffer_t *buf,
                            tsf_type_out_map_t *type_map,
                            tsf_writer_t writer,
                            void *arg) {
    uint32_t size,num,i;
    uint32_t type_code;
    
    if (!tsf_type_out_map_get_type_code(type_map,
                                        buf->type,
                                        &type_code,
                                        NULL)) {
        return tsf_false;
    }
    
    type_code=htonl(type_code);
    
    if (!writer(arg,&type_code,sizeof(type_code))) {
        return tsf_false;
    }
    
    if (tsf_type_is_dynamic(buf->type)) {
        num=tsf_type_in_map_get_num_types(buf->types);
        size=htonl(num);
        if (!writer(arg,&size,sizeof(size))) {
            return tsf_false;
        }
        for (i=0;i<num;++i) {
            if (!tsf_type_out_map_get_type_code(
                     type_map,
                     tsf_type_in_map_get_type(buf->types,i),
                     &type_code,
                     NULL)) {
                return tsf_false;
            }
            type_code=htonl(type_code);
            if (!writer(arg,&type_code,sizeof(type_code))) {
                return tsf_false;
            }
        }
    }
    
    size=htonl(buf->size);
    if (!writer(arg,&size,sizeof(size))) {
        return tsf_false;
    }
    
    if (!writer(arg,buf->data,buf->size)) {
        return tsf_false;
    }
    
    return tsf_true;
}

tsf_bool_t tsf_buffer_write_to_buf(tsf_buffer_t *buf,
                                   tsf_type_out_map_t *type_map,
                                   uint8_t **target) {
    uint32_t type_code;
    
    if (!tsf_type_out_map_get_type_code(type_map,
                                        buf->type,
                                        &type_code,
                                        NULL)) {
        return tsf_false;
    }
    
    copy_htonl_incdst(*target,(uint8_t*)&type_code);

    if (tsf_type_is_dynamic(buf->type)) {
        uint32_t num,i;
        num=tsf_type_in_map_get_num_types(buf->types);
        copy_htonl_incdst(*target,(uint8_t*)&num);
        for (i=0;i<num;++i) {
            if (!tsf_type_out_map_get_type_code(
                     type_map,
                     tsf_type_in_map_get_type(buf->types,i),
                     &type_code,
                     NULL)) {
                return tsf_false;
            }
            copy_htonl_incdst(*target,(uint8_t*)&type_code);
        }
    }
    
    copy_htonl_incdst(*target,(uint8_t*)&buf->size);
    
    memcpy(*target,buf->data,buf->size);
    (*target)+=buf->size;
    
    return tsf_true;
}


