/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_dump_types.c,v 1.2 2007/05/10 05:13:04 pizlo Exp $
 *
 * tsf_dump_types.c -- textual dump of the types of any tsf stream
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "tsf.h"
#include "tsf_indent.h"
#include "tsf_format.h"
#include "tsf_util.h"
#include "tsf_comment_dump.h"

#include <stdio.h>
#include <stdlib.h>

static void usage() {
    fprintf(stderr,"Usage: tsf_dump_types < <filename>\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"tsf_dump_types produces a textual and (hopefully) human-readable dump\n");
    fprintf(stderr,"of all types in almost any TSF stream file.  No actual data is printed.\n");
    fprintf(stderr,"However, all type information, including all structure and choice\n");
    fprintf(stderr,"elements as well as all comments are printed.  tsf_dump_types can read\n");
    fprintf(stderr,"TSF stream files in the following formats:\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"-> Raw TSF files, which use no compression.  In this case only a\n");
    fprintf(stderr,"   reset types command is required at the beginning of the stream.\n");
    fprintf(stderr,"   If it is not present, you can easily inject it.  It is simply\n");
    fprintf(stderr,"   \\0 followed by the three characters 'TSF'.\n");
#ifdef TSF_HAVE_ZLIB
    fprintf(stderr,"\n");
    fprintf(stderr,"-> TSF files compressed with ZLib.  In this case only a 'TSFz' magic\n");
    fprintf(stderr,"   header is required; if it is not present, you can easily inject it.\n");
#endif
#ifdef TSF_HAVE_BZIP2
    fprintf(stderr,"\n");
    fprintf(stderr,"-> TSF files compressed with libbzip2.  In this case only a 'TSFb'\n");
    fprintf(stderr,"   magic header is required; if it is not present, you can easily\n");
    fprintf(stderr,"   inject it.\n");
#endif
    fprintf(stderr,"\n");
    fprintf(stderr,"Note that any file or stream produced with a tsf_stream_file_output_t\n");
    fprintf(stderr,"will be readable by tsf_dump_types without any alterations.\n");
    fprintf(stderr,"\n");
    exit(1);
}

static void print_type(uint32_t indent,
                       tsf_type_t *type) {
    if (tsf_type_has_comment(type)) {
        ++indent;
        printf("\n");
        tsf_dump_comment(stdout,indent,tsf_type_get_comment(type));
        printf("%s",tsf_indent_str(indent));
    }
    printf("%s",tsf_type_kind_lc_name(tsf_type_get_kind_code(type)));
    switch (tsf_type_get_kind_code(type)) {
        case TSF_TK_ARRAY:
            printf(" of ");
            print_type(indent,tsf_array_type_get_element_type(type));
            break;
        case TSF_TK_STRUCT:
            if (tsf_struct_type_get_num_elements(type)==0) {
                printf(" {}");
                break;
            }
            {
                uint32_t i;
                printf(" {\n");
                ++indent;
                for (i=0;i<tsf_struct_type_get_num_elements(type);++i) {
                    tsf_named_type_t *n=tsf_struct_type_get_element(type,i);
                    if (tsf_named_type_has_comment(n)) {
                        tsf_dump_comment(stdout,
                                         indent,
                                         tsf_named_type_get_comment(n));
                    }
                    printf("%s%s%s: ",
                           tsf_indent_str(indent),
			   tsf_named_type_get_optional(n)?"optional ":"",
                           tsf_named_type_get_name(n));
                    print_type(indent,tsf_named_type_get_type(n));
                    printf(";\n");
                }
                --indent;
                printf("%s}",tsf_indent_str(indent));
            }
            break;
        case TSF_TK_CHOICE:
            if (tsf_choice_type_get_num_elements(type)==0) {
                printf(" {}");
                break;
            }
            {
                uint32_t i;
                printf(" {\n");
                ++indent;
                for (i=0;i<tsf_choice_type_get_num_elements(type);++i) {
                    tsf_named_type_t *n=tsf_choice_type_get_element(type,i);
                    if (tsf_named_type_has_comment(n)) {
                        tsf_dump_comment(stdout,
                                         indent,
                                         tsf_named_type_get_comment(n));
                    }
                    printf("%s%s: ",
                           tsf_indent_str(indent),
                           tsf_named_type_get_name(n));
                    print_type(indent,tsf_named_type_get_type(n));
                    printf(";\n");
                }
                --indent;
                printf("%s}",tsf_indent_str(indent));
            }
            break;
        default:
            break;
    }
    if (tsf_type_has_comment(type)) {
        --indent;
    }
}

static void type_callback(tsf_serial_in_man_t *in_man,
                          tsf_type_t *type,
                          void *arg) {
    static uint32_t count=1;
    if (count==1) {
        printf("/* what follows are type declarations generated\n");
        printf(" * by tsf_dump_types.  these declarations are\n");
        printf(" * suitable for use with tsf_define. */\n");
    }
    printf("\ndeclare type_" fui32 ": ",count++);
    print_type(0,type);
    printf(";\n");
}

int main(int c,char **v) {
    tsf_stream_file_input_t *in;
    tsf_buffer_t *buf;
    
    if (c>1) {
        usage();
    }
    
    CHECKTSF(in=tsf_stream_file_input_fd_open(0,NULL,NULL,tsf_false));
    CHECKTSF(tsf_stream_file_input_register_type_cback(in,
                                                       type_callback,
                                                       NULL));
    
    for (;;) {
        buf=tsf_stream_file_input_read(in);
        if (buf==NULL && tsf_get_error_code()==TSF_E_EOF) {
            break;
        }
        
        CHECKTSF(buf);
        
        tsf_buffer_destroy(buf);
    }
    
    tsf_stream_file_input_close(in);
    
    return 0;
}


