/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_internal.h,v 1.2 2007/06/19 06:52:21 pizlo Exp $
 *
 * tsf_internal.h -- internal stuff for tsf
 * by Filip Pizlo, 2003, 2004, 2005
 */

#ifndef FP_TSF_INTERNAL_H
#define FP_TSF_INTERNAL_H

#include "tsf_internal_config.h"
#include "tsf.h"

#include <stdlib.h>
#include <stdarg.h>

#include "tsf_inttypes.h"

/* min and max */
#define tsf_min(a,b) ((a)<(b)?(a):(b))
#define tsf_max(a,b) ((a)>(b)?(a):(b))

/* the full error setting command */
void tsf_set_error_fullv(tsf_error_t code,
                         int sys_errno,
                         int bdb_error,
                         int deadly_signal,
                         int zlib_error,
                         const char *zlib_msg,
                         const char *format,
                         va_list lst1,
                         va_list lst2);

void tsf_set_error_full(tsf_error_t code,
                        int sys_errno,
                        int bdb_error,
                        int deadly_signal,
			int zlib_error,
			const char *zlib_msg,
                        const char *format,
                        ...);

/* this is an strerror thingy for zlib.  it will abort if called when
 * no zlib support is compiled. */
const char *tsf_zlib_strerror(int code);

/* this is an strerror thingy for libbzip2.  it will abort if called
 * when no libbzip2 support is compiled. */
const char *tsf_libbzip2_strerror(int code);

/* this is just plain dumb.  thanks a lot, GNU ppl. */
#ifdef HAVE_STPCPY
char *stpcpy(char *dest,const char *src);
#endif

/* tsf_sort just calls qsort if available.  otherwise it is some shitty
 * sorting algo that I implemented. */
void tsf_sort(void *base,size_t nmemb,size_t size,
              int (*compar)(const void *a,const void *b));

/* sorts an array of uint32_t in ascending order. */
void tsf_ui32_sort(uint32_t *base,size_t nmemb);

/* string creation stuff.  yes, to use tsf_vasprintf(), you must
 * pass in two va_lists!  and yes, they both need to be initialized
 * at format.  and no, they are not both guaranteed to be used. 
 *
 * on error, both functions just set errno. */
char *tsf_vasprintf(const char *format,va_list lst1,va_list lst2);

char *tsf_asprintf(const char *format,...);

/* create a "part" file. */
char *tsf_create_part_file(const char *filename,int mode);

/* hashtable stuff */
tsf_st_table *tsf_st_init_typetable();

/* limits stuff */

/* max number of types in one stream */
#define tsf_limits_max_types(limits) \
            ((limits)==NULL?UINT32_MAX:(limits)->max_types)

/* maximum size in bytes of a serialized type */
#define tsf_limits_max_type_size(limits) \
            ((limits)==NULL?UINT32_MAX:(limits)->max_type_size)

/* maximum total size of all types read in a session */
#define tsf_limits_max_total_type_size(limits) \
            ((limits)==NULL?UINT32_MAX:(limits)->max_total_type_size)

/* maximum size in bytes of a buffer */
#define tsf_limits_max_buf_size(limits) \
            ((limits)==NULL?UINT32_MAX:(limits)->max_buf_size)

/* container depth limit */
#define tsf_limits_depth_limit(limits) \
            ((limits)==NULL?UINT32_MAX:(limits)->depth_limit)

/* comments */
#define tsf_limits_allow_comments(limits) \
            ((limits)==NULL?tsf_true:(limits)->allow_comments)

/* maximum array length */
#define tsf_limits_max_array_length(limits) \
            ((limits)==NULL?UINT32_MAX:(limits)->max_array_length)

/* maximum struct size, in number of elements */
#define tsf_limits_max_struct_size(limits) \
            ((limits)==NULL?UINT32_MAX:(limits)->max_struct_size)

/* maximum choice size, in number of elements */
#define tsf_limits_max_choice_size(limits) \
            ((limits)==NULL?UINT32_MAX:(limits)->max_choice_size)

/* maximum string size */
#define tsf_limits_max_string_size(limits) \
            ((limits)==NULL?UINT32_MAX:(limits)->max_string_size)

/* endianness stuff */

/* get ntohl, htonl, ntohs, htons */
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#else
#ifdef HAVE_LITTLE_ENDIAN
static inline uint32_t ntohl(uint32_t x) {
    union {
        uint32_t x;
        uint8_t c[4];
    } u;
    u.x=x;
    u.c[0]^=u.c[3]^=u.c[0]^=u.c[3];
    u.c[1]^=u.c[2]^=u.c[1]^=u.c[2];
    return u.x;
}
static inline uint16_t ntohs(uint16_t x) {
    union {
        uint16_t x;
        uint8_t c[2];
    } u;
    u.x=x;
    u.c[0]^=u.c[1]^=u.c[0]^=u.c[1];
    return u.x;
}
#define htonl(x) (ntohl(x))
#define htons(x) (ntohs(x))
#else
#define ntohl(x) (x)
#define htonl(x) (x)
#define ntohs(x) (x)
#define htons(x) (x)
#endif
#endif

/* define stupid trivial stuff */
#define htonc(x) (x)
#define ntohc(x) (x)

#ifdef HAVE_BIG_ENDIAN

/* #undef NEED_INT_CONVERSION */

/* define long long equivalent of the arpa/inet functions */
#define htonll(x) (x)

/* define misaligned copy equivalents of all of the arpa/inet functions */
static inline void copy_htonc(uint8_t *dst,
                              uint8_t *src) {
    *dst=*src;
}

static inline void copy_htons(uint8_t *dst,
                              uint8_t *src) {
#ifdef CAN_DO_MISALIGNED_LDST
    *((uint16_t*)dst)=*((uint16_t*)src);
#else
    dst[0]=src[0];
    dst[1]=src[1];
#endif
}

static inline void copy_htonl(uint8_t *dst,
                              uint8_t *src) {
#ifdef CAN_DO_MISALIGNED_LDST
    uint32_t tmp=*((uint32_t*)src);
    *((uint32_t*)dst)=tmp;
#else
    dst[0]=src[0];
    dst[1]=src[1];
    dst[2]=src[2];
    dst[3]=src[3];
#endif
}

static inline void copy_htonll(uint8_t *dst,
                               uint8_t *src) {
#ifdef CAN_DO_MISALIGNED_LDST
    *((uint64_t*)dst)=*((uint64_t*)src);
#else
    dst[0]=src[0];
    dst[1]=src[1];
    dst[2]=src[2];
    dst[3]=src[3];
    dst[4]=src[4];
    dst[5]=src[5];
    dst[6]=src[6];
    dst[7]=src[7];
#endif
}

#define copy_ntohc(dst,src) copy_htonc(dst,src)
#define copy_ntohs(dst,src) copy_htons(dst,src)
#define copy_ntohl(dst,src) copy_htonl(dst,src)
#define copy_ntohll(dst,src) copy_htonll(dst,src)

#else

#define NEED_INT_CONVERSION

/* define long long equivalent of the arpa/inet functions */
static inline uint64_t htonll(uint64_t x) {
    /* probably not very efficient, but oh well. */
    union {
        uint64_t x;
        uint8_t c[8];
    } u;
    u.x=x;
    u.c[0]^=u.c[7]^=u.c[0]^=u.c[7];
    u.c[1]^=u.c[6]^=u.c[1]^=u.c[6];
    u.c[2]^=u.c[5]^=u.c[2]^=u.c[5];
    u.c[3]^=u.c[4]^=u.c[3]^=u.c[4];
    return u.x;
}

/* define misaligned copy equivalents of all of the arpa/inet functsions */
static inline void copy_htonc(uint8_t *dst,
                              uint8_t *src) {
    *dst=*src;
}

static inline void copy_htons(uint8_t *dst,
                              uint8_t *src) {
    dst[0]=src[1];
    dst[1]=src[0];
}

static inline void copy_htonl(uint8_t *dst,
                              uint8_t *src) {
    dst[0]=src[3];
    dst[1]=src[2];
    dst[2]=src[1];
    dst[3]=src[0];
}

static inline void copy_htonll(uint8_t *dst,
                               uint8_t *src) {
    dst[0]=src[7];
    dst[1]=src[6];
    dst[2]=src[5];
    dst[3]=src[4];
    dst[4]=src[3];
    dst[5]=src[2];
    dst[6]=src[1];
    dst[7]=src[0];
}

#define copy_ntohc(dst,src) copy_htonc(dst,src)
#define copy_ntohs(dst,src) copy_htons(dst,src)
#define copy_ntohl(dst,src) copy_htonl(dst,src)
#define copy_ntohll(dst,src) copy_htonll(dst,src)

#endif

#define ntohll(x) (htonll(x))

/* floating point stuff */

#ifdef TSF_HAVE_IEEE_FLOAT

#ifdef NEED_INT_CONVERSION
#define NEED_FLOAT_CONVERSION
#endif

static inline float htonf(float x) {
    uint32_t ret=htonl(*((long*)&x));
    return *((float*)&ret);
}

static inline double htond(double x) {
    uint64_t ret=htonll(*((uint64_t*)&x));
    return *((double*)&ret);
}

#define ntohf(x) (htonf(x))
#define ntohd(x) (htond(x))

#if !defined(NEED_INT_CONVERSION) &&\
    defined(CAN_DO_MISALIGNED_LDST) &&\
    !defined(CAN_DO_FLOAT_MISALIGNED_LDST)
static inline void copy_htonf(uint8_t *dst,
                              uint8_t *src) {
    dst[0]=src[0];
    dst[1]=src[1];
    dst[2]=src[2];
    dst[3]=src[3];
}

static inline void copy_htond(uint8_t *dst,
                              uint8_t *src) {
    dst[0]=src[0];
    dst[1]=src[1];
    dst[2]=src[2];
    dst[3]=src[3];
    dst[4]=src[4];
    dst[5]=src[5];
    dst[6]=src[6];
    dst[7]=src[7];
}
#else
#define copy_htonf(dst,src) copy_htonl(dst,src)
#define copy_htond(dst,src) copy_htonll(dst,src)
#endif

#define copy_ntohf(dst,src) copy_htonf(dst,src)
#define copy_ntohd(dst,src) copy_htond(dst,src)

#else

#define NEED_FLOAT_CONVERSION

#error "Currently, we only support IEEE floats!"

#endif

#if defined(NEED_FLOAT_CONVERSION) || defined(NEED_INT_COVERSION)
#define NEED_CONVERSION
#endif

/* destination-incrementing misaligned copy macros.  these do
 * no bounds checking. */

#define copy_htonc_incdst(dst,src) do {     \
    copy_htonc(dst,src);                    \
    (dst)=(void*)(((uint8_t*)(dst))+1);     \
} while (0)

#define copy_ntohc_incdst(dst,src) do {     \
    copy_ntohc(dst,src);                    \
    (dst)=(void*)(((uint8_t*)(dst))+1);     \
} while (0)

#define copy_htons_incdst(dst,src) do {     \
    copy_htons(dst,src);                    \
    (dst)=(void*)(((uint16_t*)(dst))+1);    \
} while (0)

#define copy_ntohs_incdst(dst,src) do {     \
    copy_ntohs(dst,src);                    \
    (dst)=(void*)(((uint16_t*)(dst))+1);    \
} while (0)

#define copy_htonl_incdst(dst,src) do {     \
    copy_htonl(dst,src);                    \
    (dst)=(void*)(((uint32_t*)(dst))+1);    \
} while (0)

#define copy_ntohl_incdst(dst,src) do {     \
    copy_ntohl(dst,src);                    \
    (dst)=(void*)(((uint32_t*)(dst))+1);    \
} while (0)

#define copy_htonll_incdst(dst,src) do {    \
    copy_htonll(dst,src);                   \
    (dst)=(void*)(((uint64_t*)(dst))+1);    \
} while (0)

#define copy_ntohll_incdst(dst,src) do {    \
    copy_ntohll(dst,src);                   \
    (dst)=(void*)(((uint64_t*)(dst))+1);    \
} while (0)

#define copy_htonf_incdst(dst,src) do {     \
    copy_htonf(dst,src);                    \
    (dst)=(void*)(((float*)(dst))+1);       \
} while (0)

#define copy_ntohf_incdst(dst,src) do {     \
    copy_ntohf(dst,src);                    \
    (dst)=(void*)(((float*)(dst))+1);       \
} while (0)

#define copy_htond_incdst(dst,src) do {     \
    copy_htond(dst,src);                    \
    (dst)=(void*)(((double*)(dst))+1);      \
} while (0)

#define copy_ntohd_incdst(dst,src) do {     \
    copy_ntohd(dst,src);                    \
    (dst)=(void*)(((double*)(dst))+1);      \
} while (0)

/* source-incrementing misaligned copy macros.  these do
 * no bounds checking. */

#define copy_htonc_incsrc(dst,src) do {     \
    copy_htonc(dst,src);                    \
    (src)=(void*)(((uint8_t*)(src))+1);     \
} while (0)

#define copy_ntohc_incsrc(dst,src) do {     \
    copy_ntohc(dst,src);                    \
    (src)=(void*)(((uint8_t*)(src))+1);     \
} while (0)

#define copy_htons_incsrc(dst,src) do {     \
    copy_htons(dst,src);                    \
    (src)=(void*)(((uint16_t*)(src))+1);    \
} while (0)

#define copy_ntohs_incsrc(dst,src) do {     \
    copy_ntohs(dst,src);                    \
    (src)=(void*)(((uint16_t*)(src))+1);    \
} while (0)

#define copy_htonl_incsrc(dst,src) do {     \
    copy_htonl(dst,src);                    \
    (src)=(void*)(((uint32_t*)(src))+1);    \
} while (0)

#define copy_ntohl_incsrc(dst,src) do {     \
    copy_ntohl(dst,src);                    \
    (src)=(void*)(((uint32_t*)(src))+1);    \
} while (0)

#define copy_htonll_incsrc(dst,src) do {    \
    copy_htonll(dst,src);                   \
    (src)=(void*)(((uint64_t*)(src))+1);    \
} while (0)

#define copy_ntohll_incsrc(dst,src) do {    \
    copy_ntohll(dst,src);                   \
    (src)=(void*)(((uint64_t*)(src))+1);    \
} while (0)

#define copy_htonf_incsrc(dst,src) do {     \
    copy_htonf(dst,src);                    \
    (src)=(void*)(((float*)(src))+1);       \
} while (0)

#define copy_ntohf_incsrc(dst,src) do {     \
    copy_ntohf(dst,src);                    \
    (src)=(void*)(((float*)(src))+1);       \
} while (0)

#define copy_htond_incsrc(dst,src) do {     \
    copy_htond(dst,src);                    \
    (src)=(void*)(((double*)(src))+1);      \
} while (0)

#define copy_ntohd_incsrc(dst,src) do {     \
    copy_ntohd(dst,src);                    \
    (src)=(void*)(((double*)(src))+1);      \
} while (0)

/* source-incrementing misaligned copy macros.  these do
 * bounds checks.  when a bounds check fails, they jump to
 * the label you give them. */

#define copy_htonc_incsrc_bc(dst,src,end,label) do {    \
    if (((uint8_t*)(end))-((uint8_t*)(src))<1) {        \
        goto label;                                     \
    }                                                   \
    copy_htonc(dst,src);                                \
    (src)=(void*)(((uint8_t*)(src))+1);                 \
} while (0)

#define copy_ntohc_incsrc_bc(dst,src,end,label) do {    \
    if (((uint8_t*)(end))-((uint8_t*)(src))<1) {        \
        goto label;                                     \
    }                                                   \
    copy_ntohc(dst,src);                                \
    (src)=(void*)(((uint8_t*)(src))+1);                 \
} while (0)

#define copy_htons_incsrc_bc(dst,src,end,label) do {    \
    if (((uint8_t*)(end))-((uint8_t*)(src))<2) {        \
        goto label;                                     \
    }                                                   \
    copy_htons(dst,src);                                \
    (src)=(void*)(((uint16_t*)(src))+1);                \
} while (0)

#define copy_ntohs_incsrc_bc(dst,src,end,label) do {    \
    if (((uint8_t*)(end))-((uint8_t*)(src))<2) {        \
        goto label;                                     \
    }                                                   \
    copy_ntohs(dst,src);                                \
    (src)=(void*)(((uint16_t*)(src))+1);                \
} while (0)

#define copy_htonl_incsrc_bc(dst,src,end,label) do {    \
    if (((uint8_t*)(end))-((uint8_t*)(src))<4) {        \
        goto label;                                     \
    }                                                   \
    copy_htonl(dst,src);                                \
    (src)=(void*)(((uint32_t*)(src))+1);                \
} while (0)

#define copy_ntohl_incsrc_bc(dst,src,end,label) do {    \
    if (((uint8_t*)(end))-((uint8_t*)(src))<4) {        \
        goto label;                                     \
    }                                                   \
    copy_ntohl(dst,src);                                \
    (src)=(void*)(((uint32_t*)(src))+1);                \
} while (0)

#define copy_htonll_incsrc_bc(dst,src,end,label) do {   \
    if (((uint8_t*)(end))-((uint8_t*)(src))<8) {        \
        goto label;                                     \
    }                                                   \
    copy_htonll(dst,src);                               \
    (src)=(void*)(((uint64_t*)(src))+1);                \
} while (0)

#define copy_ntohll_incsrc_bc(dst,src,end,label) do {   \
    if (((uint8_t*)(end))-((uint8_t*)(src))<8) {        \
        goto label;                                     \
    }                                                   \
    copy_ntohll(dst,src);                               \
    (src)=(void*)(((uint64_t*)(src))+1);                \
} while (0)

#define copy_htonf_incsrc_bc(dst,src,end,label) do {    \
    if (((uint8_t*)(end))-((uint8_t*)(src))<4) {        \
        goto label;                                     \
    }                                                   \
    copy_htonf(dst,src);                                \
    (src)=(void*)(((float*)(src))+1);                   \
} while (0)

#define copy_ntohf_incsrc_bc(dst,src,end,label) do {    \
    if (((uint8_t*)(end))-((uint8_t*)(src))<4) {        \
        goto label;                                     \
    }                                                   \
    copy_ntohf(dst,src);                                \
    (src)=(void*)(((float*)(src))+1);                   \
} while (0)

#define copy_htond_incsrc_bc(dst,src,end,label) do {    \
    if (((uint8_t*)(end))-((uint8_t*)(src))<8) {        \
        goto label;                                     \
    }                                                   \
    copy_htond(dst,src);                                \
    (src)=(void*)(((double*)(src))+1);                  \
} while (0)

#define copy_ntohd_incsrc_bc(dst,src,end,label) do {    \
    if (((uint8_t*)(end))-((uint8_t*)(src))<8) {        \
        goto label;                                     \
    }                                                   \
    copy_ntohd(dst,src);                                \
    (src)=(void*)(((double*)(src))+1);                  \
} while (0)

/* read a string using a reader */
char *tsf_read_string(tsf_reader_t reader,
                      void *arg,
                      size_t max_size);

/* write a string using a writer */
tsf_bool_t tsf_write_string(tsf_writer_t writer,
                            void *arg,
                            const char *str);

/* size calculating writer */
tsf_bool_t tsf_size_calc_writer(void *arg,
                                void *buf,
                                uint32_t len);

/* size aware - delegates to another reader, counts the bytes read,
 * and fires a TSF_E_NOT_ALLOWED error when the limit reaches 0.  name
 * should be a short (preferrably one-word) description of what you're
 * reading; it'll be used for the error message. */
typedef struct {
    tsf_reader_t reader;
    void *arg;
    uint32_t limit;
    const char *name;
} tsf_size_aware_rdr_t;
tsf_bool_t tsf_size_aware_reader(void *arg,
                                 void *buf,
                                 uint32_t len);

/* light-weight representation of a buffer */
typedef struct {
    uint8_t *cur,
            *end;
} tsf_light_buffer_t;

/* buffer-only writer.  writes to a buffer of a pre-defined size.  you can
 * then use this buffer with the buffer-only reader.  the argument is a
 * uint8_t**.  this function will dereference it and use it as an iterator
 * as it writes. */
tsf_bool_t tsf_buffer_only_writer(void *arg,
                                  void *buf,
                                  uint32_t len);

/* buffer-only reader.  reads from a light-weight buffer. */
tsf_bool_t tsf_buffer_only_reader(void *arg,
                                  void *buf,
                                  uint32_t len);

typedef struct {
    tsf_reader_t reader;
    void *arg;
} tsf_comparing_writer_t;

/* reads from a reader and compares to what you write to it.  fails if any
   byte doesn't match. */
tsf_bool_t tsf_comparing_writer(void *arg,
				void *buf,
				uint32_t len);

/* computes a hash of the text you give it.  the arg must point to a
   32-bit integer. */
tsf_bool_t tsf_hashing_writer(void *arg,
			      void *buf,
			      uint32_t len);

/* internal recursive function for reading types */
tsf_type_t *tsf_type_read_rec(tsf_reader_t reader,
                              void *arg,
                              tsf_limits_t *limits,
                              uint32_t depth);

/* read a type table */
tsf_type_table_t* tsf_type_table_read(const char *name,
                                      tsf_reader_t reader,
                                      void *arg,
                                      tsf_limits_t *limits,
                                      uint32_t size_limit,
                                      uint32_t depth);

/* clones a type, returning a new one just like it.  note this is
 * actually deep copy. */
tsf_type_t* tsf_type_clone(tsf_type_t *type);

/* recompute the hash code */
void tsf_type_recompute_hash(tsf_type_t *type);

/* write a type according to the serial protocol */
tsf_bool_t tsf_serial_out_man_write_type(tsf_writer_t writer,
                                         void *arg,
                                         tsf_type_t *type);

/* verify that the data is valid.  basically just does NULL checks. */
tsf_bool_t tsf_reflect_verify(tsf_reflect_t *data);

/* generate a gpint prototype for a generator.  a generator takes
 * the following arguments on the stack: a pointer to a tsf_type_t,
 * and a pointer to the user structure.  the generator returns a
 * pointer to a tsf_buffer_t, or 0 if an error occurred. */
gpc_proto_t *tsf_gpc_code_gen_generate_generator(tsf_type_t *type);

/* generate a gpint prototype for a parser.  a parser takes the
 * following arguments on the stack: a pointer to a tsf_buffer_t.
 * it returns a pointer to a user structure that is also a region. */
gpc_proto_t *tsf_gpc_code_gen_generate_parser(tsf_type_t *dest_type,
                                              tsf_type_t *src_type);

/* generate a gpint prototype for a converter.  a converter takes
 * the following arguments on the stack: a pointer to the destination,
 * which may be NULL indicating that the destination should be
 * allocated, and a pointer to the source.  the region register
 * may be NULL only if the destination is NULL; if the region
 * is NULL it means that a new region should be allocated.  a
 * converter returns the destination. */
gpc_proto_t *tsf_gpc_code_gen_generate_converter(tsf_type_t *dest_type,
                                                 tsf_type_t *src_type);

/* abort because we screwed up big time */
#define tsf_abort(msg) do {						\
	fprintf(stderr,"tsf_abort(\"%s\"): called at %s:%d\n",		\
		msg,__FILE__,__LINE__);					\
	abort();							\
    } while (0)

/* function version of abort */
void tsf_f_abort(const char *msg);

/* make an assertion that, if it fails, causes us to abort */
#define tsf_assert(exp) do {						\
	if (!(exp)) {							\
	    fprintf(stderr,"tsf_assert(%s): failed at %s:%d\n",		\
		    #exp,__FILE__,__LINE__);				\
	    abort();							\
	}								\
    } while (0)

#endif


