/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_native.c,v 1.1.1.1 2007/05/02 23:31:06 pizlo Exp $
 *
 * tsf_native.c -- code generation and structure mapping support
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "tsf_internal.h"
#include "tsf_format.h"
#include <string.h>

tsf_bool_t tsf_native_type_has_struct_mapping(tsf_type_t *type) {
    /* printf("in has struct mapping: type = %p\n",type); */
    switch (type->kind_code) {
        case TSF_TK_ARRAY:
            /* an array has a struct mapping if its elements have
             * struct mappings */
            return tsf_native_type_has_struct_mapping(
                type->u.a.element_type);
        case TSF_TK_STRUCT:
            if (!type->u.s.has_size) {
                return tsf_false;
            }
            {
                uint32_t i;
                for (i=0;i<tsf_struct_type_get_num_elements(type);++i) {
                    if (!tsf_struct_type_get_element(type,i)->has_offset) {
                        return tsf_false;
                    }
                    if (!tsf_native_type_has_struct_mapping(
                             tsf_struct_type_get_element(type,i)->type)) {
                        return tsf_false;
                    }
                }
            }
            break;
        case TSF_TK_CHOICE:
            return type->u.h.has_mapping;
        default:
            /* primitives, bits, and voids have struct mappings */
            break;
    }
    
    return tsf_true;
}

size_t tsf_native_type_get_size(tsf_type_t *type) {
    /* printf("in get size: type = %p\n",type); */
    tsf_assert(tsf_native_type_has_struct_mapping(type));
    
    switch (type->kind_code) {
        case TSF_TK_ARRAY:
            if (type->u.a.element_type->kind_code==TSF_TK_BIT) {
                return sizeof(tsf_native_bitvector_t);
            }
            if (type->u.a.element_type->kind_code==TSF_TK_VOID) {
                return sizeof(tsf_native_void_array_t);
            }
            return sizeof(tsf_native_array_t);
        case TSF_TK_STRUCT:
            return type->u.s.size;
        case TSF_TK_BIT:
            return sizeof(tsf_bool_t);
        case TSF_TK_VOID:
            return 0;
        case TSF_TK_CHOICE:
            return type->u.h.size;
        case TSF_TK_STRING:
            return sizeof(char*);
        case TSF_TK_ANY:
            return sizeof(tsf_buffer_t*);
        default:
            return tsf_primitive_type_kind_size_of(type->kind_code);
    }
}

static tsf_bool_t tsf_native_type_can_blit_recurse(tsf_type_t *dest,
                                                   tsf_type_t *src,
                                                   size_t dest_size) {
    uint32_t i;
    size_t sub_dest_size;
    if (dest->kind_code!=src->kind_code) {
        return tsf_false;
    }
    if (dest_size==0) {
        dest_size=tsf_native_type_get_size(dest);
    }
    if (tsf_native_type_get_size(src) > dest_size) {
        return tsf_false;
    }
    switch (dest->kind_code) {
        case TSF_TK_ARRAY:
        case TSF_TK_STRING:
        case TSF_TK_ANY:
            return tsf_false;
        case TSF_TK_CHOICE:
            if (tsf_choice_type_get_num_elements(src) >
                tsf_choice_type_get_num_elements(dest)) {
                return tsf_false;
            }
            if ((tsf_choice_type_has_non_void(src) &&
                 src->u.h.data_offset != dest->u.h.data_offset) ||
                src->u.h.value_offset != dest->u.h.value_offset) {
                return tsf_false;
            }
            if (dest->u.h.value_offset > dest->u.h.data_offset) {
                sub_dest_size=dest->u.h.value_offset-dest->u.h.data_offset;
            } else {
                sub_dest_size=dest->u.h.size-dest->u.h.data_offset;
            }
            for (i=0;i<tsf_choice_type_get_num_elements(src);++i) {
                tsf_named_type_t *sn=tsf_choice_type_get_element(src,i);
                tsf_named_type_t *dn=tsf_choice_type_get_element(dest,i);
                if (strcmp(sn->name,dn->name)) {
                    return tsf_false;
                }
                if (!tsf_native_type_can_blit_recurse(dn->type,
                                                      sn->type,
                                                      sub_dest_size)) {
                    return tsf_false;
                }
		if (dn->type->kind_code!=TSF_TK_VOID &&
		    (!tsf_native_choice_type_is_in_place(dest) ||
		     !tsf_native_choice_type_is_in_place(src))) {
		    return tsf_false;
		}
            }
            break;
        case TSF_TK_STRUCT:
            for (i=0;i<tsf_struct_type_get_num_elements(dest);++i) {
                tsf_named_type_t *dn=tsf_struct_type_get_element(dest,i);
                tsf_named_type_t *sn;
                
                if (tsf_type_get_static_size(dn->type)==0) {
                    continue;
                }
                
                sn=tsf_struct_type_find_node(src,dn->name);
                if (sn==NULL) {
                    return tsf_false;
                }
                
                if (!tsf_native_type_can_blit_recurse(dn->type,
                                                      sn->type,
                                                      0)) {
                    return tsf_false;
                }
                
                if (dn->offset != sn->offset) {
                    return tsf_false;
                }
            }
            break;
        default:
            break;
    }
    
    return tsf_true;
}

tsf_bool_t tsf_native_type_can_blit(tsf_type_t *dest,
                                    tsf_type_t *src) {
    if (!tsf_native_type_has_struct_mapping(dest) ||
        !tsf_native_type_has_struct_mapping(src)) {
        return tsf_false;
    }
    
    return tsf_native_type_can_blit_recurse(dest,src,0);
}

tsf_bool_t tsf_native_struct_type_map(tsf_type_t **type,
                                      const char *name,
                                      size_t offset) {
    tsf_type_t *ret;
    tsf_named_type_t *n;
    /* uint32_t i; */
    
    if ((*type)->ref_count==1) {
        ret=*type;
    } else {
        ret=tsf_type_clone(*type);
        if (ret==NULL) {
            return tsf_false;
        }
    }
    
    n=tsf_struct_type_find_node(ret,name);
    
    if (n==NULL) {
        tsf_type_destroy(ret);
        return tsf_false;
    }

    /* the code that follows never really did catch all of the possible
       bugs.  it only made things slower.  so I got rid of it. */
    
#if 0
    if (tsf_native_type_has_struct_mapping(n->type)) {
        size_t my_size=tsf_native_type_get_size(n->type);
        
        for (i=0;i<tsf_struct_type_get_num_elements(ret);++i) {
            tsf_named_type_t *n2=tsf_struct_type_get_element(ret,i);
            size_t other_size;
            
            if (!tsf_native_type_has_struct_mapping(n2->type) ||
                !n2->has_offset) {
                continue;
            }
            
            other_size=tsf_native_type_get_size(n2->type);
            
            if (offset+my_size>n2->offset &&
                n2->offset+other_size>offset) {
                tsf_set_error(TSF_E_BAD_STRUCT_MAP,
                              "Element '%s' overlaps with element '%s' in "
                              "tsf_native_struct_type_map()",
                              name,n2->name);
                tsf_type_destroy(ret);
                return tsf_false;
            }
        }
        
        if (ret->u.s.has_size) {
            if (offset+my_size>ret->u.s.size) {
                tsf_set_error(TSF_E_BAD_STRUCT_MAP,
                              "Element '%s' goes beyond the end of "
                              "the structure in "
                              "tsf_native_struct_type_map()",
                              name);
                tsf_type_destroy(ret);
                return tsf_false;
            }
        }
    }
#endif
    
    n->has_offset=tsf_true;
    n->offset=offset;
    
    if (ret!=*type) {
        tsf_type_destroy(*type);
        *type=ret;
    }
    
    return tsf_true;
}

tsf_bool_t tsf_native_struct_type_set_callbacks(tsf_type_t **type,
						void (*cons)(void*),
						void (*dest)(void*)) {
    tsf_type_t *ret=tsf_type_clone(*type);
    if (ret==NULL) {
	return tsf_false;
    }

    ret->u.s.constructor=cons;
    ret->u.s.destructor=dest;

    tsf_type_destroy(*type);
    *type=ret;

    return tsf_true;
}

tsf_bool_t tsf_native_struct_type_set_size(tsf_type_t **type,
                                           size_t size) {
    uint32_t i;
    tsf_type_t *ret=tsf_type_clone(*type);
    if (ret==NULL) {
        return tsf_false;
    }
    
    for (i=0;i<tsf_struct_type_get_num_elements(ret);++i) {
        tsf_named_type_t *n=tsf_struct_type_get_element(ret,i);
        size_t other_size;
        
        if (!tsf_native_type_has_struct_mapping(n->type) ||
            !n->has_offset) {
            continue;
        }
        
        other_size=tsf_native_type_get_size(n->type);
        
        if (n->offset+other_size > size) {
            tsf_set_error(TSF_E_BAD_STRUCT_MAP,
                          "Element '%s' goes beyond the end of "
                          "the structure in "
                          "tsf_native_struct_type_set_size()",
                          n->name);
            tsf_type_destroy(ret);
            return tsf_false;
        }
    }

    ret->u.s.has_size=tsf_true;
    ret->u.s.size=size;
    
    tsf_type_destroy(*type);
    *type=ret;
    return tsf_true;
}

tsf_bool_t tsf_native_choice_type_map(tsf_type_t **type,
				      tsf_bool_t in_place,
                                      size_t value_offset,
                                      size_t data_offset,
                                      size_t size) {
    tsf_type_t *ret=tsf_type_clone(*type);
    if (ret==NULL) {
        return tsf_false;
    }
    
    ret->u.h.has_mapping=tsf_true;
    ret->u.h.in_place=in_place;
    ret->u.h.value_offset=value_offset;
    ret->u.h.data_offset=data_offset;
    ret->u.h.size=size;
    
    tsf_type_destroy(*type);
    *type=ret;
    return tsf_true;
}

tsf_bool_t tsf_native_type_produce_some_mapping(tsf_type_t **type,
						tsf_bool_t do_align) {
    tsf_type_t *ret=tsf_type_clone(*type);
    uint32_t i;
    size_t total_size;
    
    if (ret==NULL) {
        return tsf_false;
    }
    
    switch ((*type)->kind_code) {
        case TSF_TK_ARRAY:
            if (!tsf_native_type_produce_some_mapping(&ret->u.a.element_type,
						      do_align)) {
                tsf_type_destroy(ret);
                return tsf_false;
            }
            break;
        case TSF_TK_STRUCT:
            total_size=0;
            
            for (i=0;
                 i<tsf_type_table_get_num_elements(ret->u.s.tt);
                 ++i) {
                tsf_named_type_t *n=
                    tsf_type_table_find_by_index(ret->u.s.tt,i);
                
                if (!tsf_native_type_produce_some_mapping(&n->type,
							  do_align)) {
                    tsf_type_destroy(ret);
                    return tsf_false;
                }
                
		if (do_align) {
		    n->offset=TSF_ALIGN(total_size);
		} else {
		    n->offset=total_size;
		}
                n->has_offset=tsf_true;
                
                total_size=n->offset+tsf_native_type_get_size(n->type);
            }
            
            ret->u.s.size=total_size;
            ret->u.s.has_size=tsf_true;
            break;
        case TSF_TK_CHOICE:
	    ret->u.h.in_place=tsf_true;
	    
            total_size=0;
            
            for (i=0;
                 i<tsf_type_table_get_num_elements(ret->u.h.tt);
                 ++i) {
                tsf_named_type_t *n=
                    tsf_type_table_find_by_index(ret->u.h.tt,i);
                
                size_t size;
                
                if (!tsf_native_type_produce_some_mapping(&n->type,
							  do_align)) {
                    tsf_type_destroy(ret);
                    return tsf_false;
                }
                
                size=tsf_native_type_get_size(n->type);
                
                if (size>total_size) {
                    total_size=size;
                }
            }

            ret->u.h.value_offset=0;

	    if (do_align) {
		ret->u.h.data_offset=TSF_ALIGN(sizeof(tsf_choice_t));
	    } else {
		ret->u.h.data_offset=sizeof(tsf_choice_t);
	    }

            ret->u.h.size=ret->u.h.data_offset+total_size;
            ret->u.h.has_mapping=tsf_true;
            break;
        default:
            break;
    }
    
    tsf_type_destroy(*type);
    *type=ret;
    return tsf_true;
}



