/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_parser.c,v 1.1.1.1 2007/05/02 23:31:06 pizlo Exp $
 *
 * tsf_parser.c -- parser implementation
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "tsf_internal.h"
#include "tsf_region.h"
#include "gpc.h"
#include <unistd.h>
#include <stdio.h>
#include <string.h>

#ifdef HAVE_PTHREAD
#include <pthread.h>
#endif

static int table_destroy_pair(tsf_type_t *type,
                              gpc_program_t *value,
                              void *arg) {
    tsf_type_destroy(type);
    gpc_program_destroy(value);
    return TSF_ST_CONTINUE;
}

tsf_parser_t *tsf_parser_create(tsf_type_t *type,
                                tsf_engine_t conf,
				const char *cache_dir) {
    tsf_parser_t *result;
    
    if (!tsf_native_type_has_struct_mapping(type)) {
        tsf_set_error(TSF_E_NO_STRUCT_MAP,NULL);
        return NULL;
    }
    
    result=malloc(sizeof(tsf_parser_t));
    if (result==NULL) {
        tsf_set_errno("Could not malloc tsf_parser_t");
        return NULL;
    }
    
    result->conf=conf;
    if (result->conf==TSF_EN_NOT_SUPPORTED) {
        free(result);
        return NULL;
    }

    if (cache_dir!=TSF_CD_DEFAULT &&
	cache_dir!=TSF_CD_NO_CACHE) {
	result->cache_dir=strdup(cache_dir);
	if (result->cache_dir==NULL) {
	    tsf_set_errno("Could not strdup()");
	    free(result);
	    return NULL;
	}
    } else {
	result->cache_dir=(char*)cache_dir;
    }
    
    result->parsers=tsf_st_init_typetable();
    if (result->parsers==NULL) {
	if (cache_dir!=TSF_CD_DEFAULT &&
	    cache_dir!=TSF_CD_NO_CACHE) {
	    free(result->cache_dir);
	}
        free(result);
        tsf_set_errno("Could not tsf_st_init_typetable()");
        return NULL;
    }
    
    result->expected_type=tsf_type_dup(type);

#ifdef HAVE_PTHREAD
    pthread_mutex_init(&result->lock,NULL);
#endif
    
    return result;
}

void tsf_parser_destroy(tsf_parser_t *parser) {
#ifdef HAVE_PTHREAD
    pthread_mutex_destroy(&parser->lock);
#endif
    tsf_st_foreach(parser->parsers,
                   (tsf_st_func_t)table_destroy_pair,
                   NULL);
    tsf_st_free_table(parser->parsers);
    tsf_type_destroy(parser->expected_type);
    if (parser->cache_dir!=TSF_CD_DEFAULT &&
	parser->cache_dir!=TSF_CD_NO_CACHE) {
	free(parser->cache_dir);
    }
    free(parser);
}

void *tsf_parser_parse(tsf_parser_t *parser,
                       tsf_buffer_t *buffer) {
    return tsf_parser_parse_in(parser,buffer,NULL);
}

void *tsf_parser_parse_in(tsf_parser_t *parser,
                          tsf_buffer_t *buffer,
                          void *region) {
    gpc_program_t *program;
    gpc_cell_t args[1];

#ifdef HAVE_PTHREAD
    pthread_mutex_lock(&parser->lock);
#endif
    if (!tsf_st_lookup(parser->parsers,
                       (char*)buffer->type,
                       (void*)&program)) {
        /* not found, must create! */
        
        tsf_type_t *type;
        
        gpc_proto_t *proto=
            tsf_gpc_code_gen_generate_parser(parser->expected_type,
                                             buffer->type);
        
        if (proto==NULL) {
#ifdef HAVE_PTHREAD
	    pthread_mutex_unlock(&parser->lock);
#endif
            return NULL;
        }
        
        gpc_code_gen_debug("parser",proto);
        
        program=gpc_program_from_proto(proto,
				       parser->conf,
				       parser->cache_dir);
        
        gpc_proto_destroy(proto);
        
        if (program==NULL) {
#ifdef HAVE_PTHREAD
	    pthread_mutex_unlock(&parser->lock);
#endif
            return NULL;
        }
        
        type=tsf_type_dup(buffer->type);
        
        if (tsf_st_insert(parser->parsers,(char*)type,program)<0) {
            tsf_set_errno("Could not tsf_st_insert() in "
                          "tsf_parser_parse()");
            tsf_type_destroy(type);
            gpc_program_destroy(program);
#ifdef HAVE_PTHREAD
	    pthread_mutex_unlock(&parser->lock);
#endif
            return NULL;
        }
    }
#ifdef HAVE_PTHREAD
    pthread_mutex_unlock(&parser->lock);
#endif
    
    args[0]=(gpc_cell_t)buffer;
    
    return (void*)gpc_program_run(program,region,args);
}

