/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_serial_out_man.c,v 1.1.1.1 2007/05/02 23:31:06 pizlo Exp $
 *
 * tsf_serial_out_man.c -- output manager for the serial format
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "tsf_internal.h"
#include "tsf_serial_protocol.h"

#ifdef HAVE_PTHREAD
#include <pthread.h>
#endif

static tsf_bool_t write_command(tsf_writer_t writer,
                                void *arg,
                                uint8_t command) {
    return writer(arg,&command,sizeof(command));
}

static tsf_bool_t write_reset(tsf_writer_t writer,
                              void *arg) {
    return writer(arg,TSF_SP_MAGIC_CODE,TSF_SP_MAGIC_LEN);
}

tsf_serial_out_man_t *tsf_serial_out_man_create(tsf_writer_t writer,
                                                void *arg) {
    tsf_serial_out_man_t *ret=malloc(sizeof(tsf_serial_out_man_t));
    if (ret==NULL) {
        tsf_set_errno("Could not malloc tsf_serial_out_man_t structure.");
        return NULL;
    }
    
    ret->types=tsf_type_out_map_create();
    if (ret->types==NULL) {
        free(ret);
        return NULL;
    }
    
    ret->writer=writer;
    ret->arg=arg;
    
    if (!write_reset(writer,arg)) {
        tsf_type_out_map_destroy(ret->types);
        free(ret);
        return NULL;
    }
    
#ifdef HAVE_PTHREAD
    pthread_mutex_init(&ret->lock,NULL);
#endif
    
    return ret;
}

void tsf_serial_out_man_destroy(tsf_serial_out_man_t *out_man) {
    tsf_type_out_map_destroy(out_man->types);

#ifdef HAVE_PTHREAD
    pthread_mutex_destroy(&out_man->lock);
#endif
    
    free(out_man);
}

tsf_bool_t tsf_serial_out_man_reset_types(tsf_serial_out_man_t *out_man) {
    tsf_type_out_map_t *new_types;
    
#ifdef HAVE_PTHREAD
    pthread_mutex_lock(&out_man->lock);
#endif
    
    new_types=tsf_type_out_map_create();
    if (new_types==NULL) {
#ifdef HAVE_PTHREAD
        pthread_mutex_unlock(&out_man->lock);
#endif
    
        return tsf_false;
    }
    
    if (!write_reset(out_man->writer,
                     out_man->arg)) {
#ifdef HAVE_PTHREAD
        pthread_mutex_unlock(&out_man->lock);
#endif
    
        tsf_type_out_map_destroy(new_types);
        return tsf_false;
    }
    
    tsf_type_out_map_destroy(out_man->types);
    out_man->types=new_types;
    
#ifdef HAVE_PTHREAD
    pthread_mutex_unlock(&out_man->lock);
#endif
    
    return tsf_true;
}

tsf_bool_t tsf_serial_out_man_write_type(tsf_writer_t writer,
                                         void *arg,
                                         tsf_type_t *type) {
    if (!write_command(writer,arg,TSF_SP_C_NEW_TYPE)) {
        return tsf_false;
    }
    return tsf_type_write(type,writer,arg);
}

tsf_bool_t tsf_serial_out_man_write(tsf_serial_out_man_t *out_man,
                                    tsf_buffer_t *buf) {
    uint32_t i;
    tsf_bool_t created,result;
    
#ifdef HAVE_PTHREAD
    pthread_mutex_lock(&out_man->lock);
#endif
    
    if (!tsf_type_out_map_get_type_code(out_man->types,
                                        buf->type,
                                        NULL,
                                        &created)) {
#ifdef HAVE_PTHREAD
        pthread_mutex_unlock(&out_man->lock);
#endif
    
        return tsf_false;
    }
    
    if (created) {
        if (!tsf_serial_out_man_write_type(out_man->writer,
                                           out_man->arg,
                                           buf->type)) {
#ifdef HAVE_PTHREAD
            pthread_mutex_unlock(&out_man->lock);
#endif
    
            return tsf_false;
        }
    }
    
    for (i=0;i<tsf_type_in_map_get_num_types(buf->types);++i) {
        if (!tsf_type_out_map_get_type_code(
		out_man->types,
		tsf_type_in_map_get_type(buf->types,i),
		NULL,
		&created)) {
#ifdef HAVE_PTHREAD
            pthread_mutex_unlock(&out_man->lock);
#endif
    
            return tsf_false;
        }
        
        if (created) {
            if (!tsf_serial_out_man_write_type(
                     out_man->writer,
                     out_man->arg,
                     tsf_type_in_map_get_type(buf->types,i))) {
#ifdef HAVE_PTHREAD
                pthread_mutex_unlock(&out_man->lock);
#endif
    
                return tsf_false;
            }
        }
    }
    
    if (!write_command(out_man->writer,out_man->arg,TSF_SP_C_DATA)) {
#ifdef HAVE_PTHREAD
        pthread_mutex_unlock(&out_man->lock);
#endif
    
        return tsf_false;
    }
    
    result=tsf_buffer_write(buf,
                            out_man->types,
                            out_man->writer,
                            out_man->arg);

#ifdef HAVE_PTHREAD
    pthread_mutex_unlock(&out_man->lock);
#endif
    
    return result;
}


