/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_stream_file_output.c,v 1.2 2007/06/19 06:52:21 pizlo Exp $
 *
 * tsf_stream_file_output.c -- file output!!
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "tsf_internal.h"
#include "tsf_serial_protocol.h"

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

tsf_stream_file_output_t *tsf_stream_file_output_open(const char *filename,
                                                      tsf_bool_t append,
                                                    tsf_zip_wtr_attr_t *attr) {
    int fd;
    tsf_stream_file_output_t *result;
    int flags=O_WRONLY|(append?O_APPEND:O_TRUNC);
    tsf_bool_t created=tsf_false;
    tsf_zip_wtr_attr_t defaults;

    if (attr==NULL) {
        attr=&defaults;
        tsf_zip_wtr_attr_get_default(attr);
    }
    
    /* this is written in a totally crazy frikkin' way so that it cleans
     * up after itself correctly in case of error. */
    
    for (;;) {
        /* first try to open an existing file */
        fd=open(filename,flags);
        
        if (fd>=0) {
            break;
        }
        
        if (errno!=ENOENT) {
            tsf_set_errno("Could not open %s for writing",
                          filename);
            return NULL;
        }
        
        /* the file didn't exist, so try creating it. */
        fd=open(filename,O_CREAT|O_EXCL|flags,0666);
        
        if (fd>=0) {
            created=tsf_true;
            break;
        }
        
        if (errno!=EEXIST) {
            tsf_set_errno("Could not open %s for writing",
                          filename);
            return NULL;
        }
    }
    
    result=tsf_stream_file_output_fd_open(fd,attr,tsf_true);
    if (result==NULL) {
        close(fd);
        
        /* this is where the craziness above gets used.  the loop above
         * ensures that we always know if we were the ones who created
         * the file.  if we were the ones who created, we delete it
         * now.  but if we were not the ones who created, we leave it
         * alone. */
	/* FIXME: not sure if this is correct.  even if we didn't create
	   the file, we still end up writing a bunch of garbage into it.
	   so maybe the correct thing to do is to either always delete it,
	   never delete it, or somehow preserve the old contents of the file
	   to be able to restore said contents here. */
        if (created) {
            unlink(filename);
        }
        return NULL;
    }
    
    return result;
}

tsf_stream_file_output_t *
tsf_stream_file_output_create_atomic(const char *filename,
				     tsf_zip_wtr_attr_t *attr) {
    char *part_flnm;
    tsf_stream_file_output_t *out;
    
    part_flnm=tsf_create_part_file(filename,0666);
    if (part_flnm==NULL) {
	return NULL;
    }
    
    out=tsf_stream_file_output_open(part_flnm,tsf_false,attr);
    if (out==NULL) {
	unlink(part_flnm);
	free(part_flnm);
	return NULL;
    }
    
    out->rename_from=part_flnm;
    out->rename_to=strdup(filename);
    if (out->rename_to==NULL) {
	tsf_set_errno("Could not strdup");
	tsf_stream_file_output_destroy(out);
	return NULL;
    }
    
    return out;
}

tsf_stream_file_output_t *
tsf_stream_file_output_fd_open(int fd,
			       tsf_zip_wtr_attr_t *attr,
			       tsf_bool_t keep) {
    tsf_stream_file_output_t *ret;
    tsf_zip_wtr_attr_t defaults;
    tsf_writer_t writer;
    
    if (attr==NULL) {
        attr=&defaults;
        tsf_zip_wtr_attr_get_default(attr);
    }
    
    ret=malloc(sizeof(tsf_stream_file_output_t));
    if (ret==NULL) {
        tsf_set_errno("Could not malloc tsf_stream_file_output_t");
        return NULL;
    }
    
    ret->rename_to=NULL;
    ret->rename_from=NULL;
    ret->all_good=tsf_true;
    ret->fd=fd;
    ret->keep_fd=keep;
    
    switch (attr->mode) {
        case TSF_ZIP_NONE:
            ret->use_zip=tsf_false;
            ret->writer.buf=tsf_buf_writer_create(ret->fd,
                                                  4096,
                                                  tsf_false);
            writer=tsf_buf_writer_write;
            break;
        case TSF_ZIP_ZLIB:
        case TSF_ZIP_BZIP2:
            if (!tsf_fd_writer(&fd,
                               attr->mode==TSF_ZIP_ZLIB?
                               TSF_SP_ZLIB_MAGIC_CODE:
                               TSF_SP_BZIP2_MAGIC_CODE,
                               TSF_SP_MAGIC_LEN)) {
                free(ret);
                return NULL;
            }
    
            ret->use_zip=tsf_true;
            ret->writer.zip=tsf_zip_writer_create(tsf_fd_writer,
                                                  &(ret->fd),
                                                  attr);
            writer=tsf_zip_writer_write;
            break;
        default:
            tsf_set_error(TSF_E_INVALID_ARG,
                          "%d is not a valid zip mode",
                          attr->mode);
            free(ret);
            return NULL;
    }

    if (ret->writer.ptr==NULL) {
        free(ret);
        return NULL;
    }
    
    ret->out_man=tsf_serial_out_man_create(writer,
                                           ret->writer.ptr);
    if (ret->out_man==NULL) {
        if (ret->use_zip) {
            tsf_zip_writer_destroy(ret->writer.zip);
        } else {
            tsf_buf_writer_destroy(ret->writer.buf);
        }
        free(ret);
        return NULL;
    }
    
    return ret;
}

static void do_renaming(tsf_stream_file_output_t *file) {
    if (file->all_good &&
	file->rename_to!=NULL &&
	file->rename_from!=NULL) {
	if (rename(file->rename_from,file->rename_to)<0) {
	    tsf_set_errno("Could not rename %s to %s",
			  file->rename_from,
			  file->rename_to);
	    file->all_good=tsf_false;
	} else {
	    free(file->rename_to);
	    free(file->rename_from);
	    file->rename_to=NULL;
	    file->rename_from=NULL;
	}
    }
}

tsf_bool_t tsf_stream_file_output_close(tsf_stream_file_output_t *file) {
    if (file->all_good) {
	if (file->use_zip) {
	    file->all_good=tsf_zip_writer_finish(file->writer.zip);
	} else {
	    file->all_good=tsf_buf_writer_flush(file->writer.buf);
	}
    }
    do_renaming(file);
    tsf_bool_t result=file->all_good;
    tsf_stream_file_output_destroy(file);
    return result;
}

void tsf_stream_file_output_destroy(tsf_stream_file_output_t *file) {
    tsf_serial_out_man_destroy(file->out_man);
    if (file->use_zip) {
        tsf_zip_writer_destroy(file->writer.zip);
    } else {
        tsf_buf_writer_destroy(file->writer.buf);
    }
    if (file->keep_fd) {
        close(file->fd);
    }
    if (file->rename_from!=NULL) {
	unlink(file->rename_from);
	free(file->rename_from);
    }
    if (file->rename_to!=NULL) {
	free(file->rename_to);
    }
    free(file);
}

tsf_bool_t tsf_stream_file_output_flush(tsf_stream_file_output_t *file) {
    if (!file->all_good) {
	tsf_set_error(TSF_E_BAD_STATE,
		      "Attempt to flush a stream that has already "
		      "experienced an error");
	return tsf_false;
    }
    if (file->use_zip) {
        file->all_good=tsf_zip_writer_flush(file->writer.zip,tsf_false);
    } else {
        file->all_good=tsf_buf_writer_flush(file->writer.buf);
    }
    do_renaming(file);
    return file->all_good;
}

tsf_bool_t tsf_stream_file_output_write(tsf_stream_file_output_t *file,
                                        tsf_buffer_t *buffer) {
    if (!file->all_good) {
	tsf_set_error(TSF_E_BAD_STATE,
		      "Attempt to write to a stream that has already "
		      "experienced an error");
	return tsf_false;
    }
    return file->all_good=tsf_serial_out_man_write(file->out_man,buffer);
}



