/*
 * $Header: /Users/pizlo/Repository/tsf/tsf_test.c,v 1.1.1.1 2007/05/02 23:31:06 pizlo Exp $
 *
 * tsf_test.c -- management of test context
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "tsf_internal.h"
#include "tsf_test.h"
#include <stdarg.h>
#include <stdio.h>
#include <sys/wait.h>
#include <unistd.h>

#ifdef HAVE_REGEX
#include <regex.h>
#endif

#ifdef FAULT_INJECT
#include "fault_inject.h"
#endif

typedef struct test_s {
    tsf_test_status_t status;
    char *name;
    char *message;
    
    struct test_s *next;
} test_t;

static test_t *head=NULL,*tail=NULL;

static char **stack=NULL,**top=NULL;

static unsigned test_count=0;

static tsf_bool_t do_forking,
                  parent=tsf_true;

static int result_pipe[2];

#ifdef FAULT_INJECT
static int stack_trace_pipe[2];
#endif

static tsf_bool_t premature_exit;

#ifdef HAVE_REGEX
static regex_t include_pattern;
#endif

void tsf_test_break() {
    /* do nothing.  expect user to set breakpoint here. */
}

void tsf_test_push(const char *msg,...) {
    va_list lst1,lst2;
    char *ele;
    
    va_start(lst1,msg);
    va_start(lst2,msg);
    
    ele=tsf_vasprintf(msg,lst1,lst2);
    if (ele==NULL) {
        tsf_abort("tsf_vasprintf() failed in tsf_test_push()");
    }
    
    if (stack==NULL) {
        stack=malloc(sizeof(char*));
        if (stack==NULL) {
            tsf_abort("malloc failed in tsf_test_push()");
        }
        top=stack;
    } else {
        char **new_stack=realloc(stack,sizeof(char*)*(top-stack+1));
        if (new_stack==NULL) {
            tsf_abort("realloc failed in tsf_test_push()");
        }
        top+=new_stack-stack;
        stack=new_stack;
    }
    
    *top++=ele;
}

void tsf_test_pop(void) {
    free(*--top);
}

static char *tsf_test_current_recurse(char **cur) {
    char *above,*result;
    
    if (cur==top) {
        return strdup("");
    }
    if (cur==top-1) {
        return strdup(*cur);
    }
    
    above=tsf_test_current_recurse(cur+1);
    tsf_assert(above!=NULL);
    
    result=tsf_asprintf("%s/%s",*cur,above);
    tsf_assert(result!=NULL);
    
    free(above);
    
    return result;
}

char *tsf_test_current(void) {
    return tsf_test_current_recurse(stack);
}

tsf_bool_t tsf_should_run_test(void) {
#ifdef HAVE_REGEX
    tsf_bool_t result;
    char *test_name=tsf_test_current();
    result=(0==regexec(&include_pattern,
                       test_name,
                       0,NULL,0));
    free(test_name);
    return result;
#else
    return tsf_false;
#endif
}

#ifdef FAULT_INJECT
static void new_trace_cback(void) {
    fi_write_trace(stack_trace_pipe[1]);
}
#endif

void tsf_test_init(int c,char **v,tsf_bool_t _do_forking) {
    const char *tsf_test_pattern;
    const char *tsf_test_nofork;
#ifdef HAVE_REGEX
    int re_res;
#endif
    tsf_test_pattern=getenv("TSF_TEST_PATTERN");
#ifdef HAVE_REGEX
    if (tsf_test_pattern==NULL) {
        tsf_test_pattern=".*";
    } else {
        fprintf(stderr,
                "NOTE: restricting the suite to tests whose full names "
                "conform to the regular expression '%s'.\n",
                tsf_test_pattern);
    }
    if ((re_res=regcomp(&include_pattern,
                        tsf_test_pattern,
                        REG_EXTENDED|REG_NOSUB))) {
        char err_buf[256];
        regerror(re_res,&include_pattern,err_buf,256);
        fprintf(stderr,"Error in test pattern: %s\n",err_buf);
        exit(1);
    }
#else
    if (tsf_test_pattern!=NULL) {
        fprintf(stderr,
                "NOTE: ignoring test pattern due to lack of regex support.\n");
    }
#endif
    tsf_test_nofork=getenv("TSF_TEST_NOFORK");
    if (tsf_test_nofork==NULL || strcasecmp(tsf_test_nofork,"yes")) {
        do_forking=_do_forking;
    } else {
        fprintf(stderr,"NOTE: test forking disabled by environment.\n");
        do_forking=tsf_false;
    }
#ifdef FAULT_INJECT
    if (!do_forking) {
        fprintf(stderr,"NOTE: forking must be enabled for fault injection!\n");
    } else {
        if (getenv("DYLD_INSERT_LIBRARIES")==NULL ||
            getenv("DYLD_FORCE_FLAT_NAMESPACE")==NULL) {
            char **v_copy;
            tsf_assert(putenv("DYLD_INSERT_LIBRARIES=./libfi_overrides.dylib")==0);
            tsf_assert(putenv("DYLD_FORCE_FLAT_NAMESPACE=yes")==0);
            v_copy=malloc((c+1)*sizeof(char*));
            tsf_assert(v_copy!=NULL);
            memcpy(v_copy,v,c*sizeof(char*));
            v_copy[c]=NULL;
            execv(v[0],v_copy);
        }
        fi_init_overrides();
        fi_set_new_trace_cback(new_trace_cback);
    }
#endif
    setvbuf(stderr,NULL,_IONBF,0);
}

static void display_status_code(tsf_test_status_t status) {
    switch (status) {
        case tsf_test_success: break;
        case tsf_test_failure:
            fprintf(stderr,"F");
            break;
        case tsf_test_error_dropped:
            fprintf(stderr,"D");
            break;
        case tsf_test_system_error:
        case tsf_test_tsf_error:
            fprintf(stderr,"E");
            break;
        case tsf_test_exit:
            fprintf(stderr,"X");
            break;
        case tsf_test_crash:
            fprintf(stderr,"C");
            break;
        default:
            tsf_abort("Unknown status code");
    }
}

static void parse_status(void) {
    int res;
    test_t *test;
    int wait_status,pid;
    char *buf=NULL; /* setting to NULL to make GCC happy; in reality, every
                       use of buf happens after buf has been set to
                       something other than this NULL. */
    size_t buf_size;
    size_t buf_off;
    tsf_bool_t keep_buf=tsf_true;
    tsf_test_status_t status;
    
    premature_exit=tsf_true;    /* will set this to false once enough
                                   of the result is read from the
                                   pipe */
    
    /* try to read the status code.  if it is not there at all, this counts
     * as a premature exit error. */
    for (;;) {
        res=read(result_pipe[0],&status,sizeof(status));
        if (res<0 && errno==EINTR) {
            continue;
        }
        break;
    }
    
    tsf_assert(res>=0); /* can't think of any good reason why this would fail */
    
    if (res==0) {
        /* premature exit; could be a tsf_test_exit or tsf_test_crash */
        
        /* nothing else to do here - drop out and do the wait madness */
    } else {
        /* there's something in that there pipe; could be one of:
         * tsf_test_success
         * tsf_test_failure
         * tsf_test_tsf_error
         */
        
        /* the plan:
         * 1) read the rest of the status
         * 2) read the message
         * 3) stuff both into the linked list
         * 4) drop down to do the wait business
         *
         * through steps 1-3, beware of formatting errors.  if a formatting
         * error is encountered, label the status as tsf_test_bizarre.
	 * note: it doesn't seem like I am labelling anything as
	 * tsf_test_bizarre at the moment... */
        
        if (res<sizeof(status)) {
            int len;
            buf=((char*)&status)+res;
            len=sizeof(status)-res;
            while (len>0) {
                res=read(result_pipe[0],buf,len);
                tsf_assert(res>=0);
                if (res==0) {
                    goto do_wait;
                }
                buf+=res;
                len-=res;
            }
        }
        
        /* ok, we now have a status */
        
        /* now read the actual message */
        buf_size=80;
        tsf_assert((buf=malloc(buf_size))!=NULL);
        keep_buf=tsf_false;
        buf_off=0;
        
        for (;;) {
            while (buf_off<buf_size-1) {
                res=read(result_pipe[0],
                         buf+buf_off,
                         buf_size-buf_off-1);
                tsf_assert(res>=0);
                if (res==0) {
                    buf[buf_off]=0;
                    premature_exit=tsf_false;
                    goto do_wait;
                }
                buf_off+=res;
            }
            buf_size+=80;
            tsf_assert((buf=realloc(buf,buf_size))!=NULL);
        }
    }
    
do_wait:
    for (;;) {
        pid=wait(&wait_status);
        
        if (pid>=0 && WIFSTOPPED(wait_status)) {
            continue;
        }
        
        if (pid>=0 || errno!=EINTR) {
            break;
        }
    }
    
    if (WIFEXITED(wait_status)) {
        if (WEXITSTATUS(wait_status)!=0 || premature_exit) {
            tsf_test_complete(tsf_test_exit,
                              tsf_error_none,
                              "Child process unexpectadly exited with code "
                              "%d.\n",
                              WEXITSTATUS(wait_status));
        } else {
            tsf_assert((test=malloc(sizeof(test_t)))!=NULL);
            
            test->name=tsf_test_current();
            test->next=NULL;
            test->status=status;
            test->message=buf;
            
            if (head==NULL) {
                head=tail=test;
            } else {
                tail->next=test;
                tail=test;
            }
            
            keep_buf=tsf_true;
        }
    } else if (WIFSIGNALED(wait_status)) {
        tsf_test_complete(tsf_test_crash,
                          tsf_error_none,
                          "Child process crashed with signal %d.\n",
                          WTERMSIG(wait_status));
    } else {
        tsf_test_complete(tsf_test_crash,
                          tsf_error_none,
                          "Child process terminated with unknown status "
                          "code %d.\n",
                          wait_status);
    }
    
    if (!keep_buf) {
        free(buf);
    }
}

void tsf_test_complete(tsf_test_status_t status,
                       tsf_error_status_t error_status,
                       const char *format,
                       ...) {
    va_list lst;

    if (do_forking) {
#ifdef FAULT_INJECT
        fi_set_checking_enabled(tsf_false);
#endif
        
        if (parent) {
            va_list lst2;
            test_t *test;
            va_start(lst,format);
            va_start(lst2,format);
            
            display_status_code(status);
            
            tsf_assert((test=malloc(sizeof(test_t)))!=NULL);
            
            test->status=status;
            test->name=tsf_test_current();
            tsf_assert((test->message=tsf_vasprintf(format,lst,lst2))!=NULL);
            test->next=NULL;
            
            if (head==NULL) {
                head=tail=test;
            } else {
                tail->next=test;
                tail=test;
            }
        } else {
            FILE *fout;
        
#ifdef FAULT_INJECT 
            if (fi_had_enomem()) {
                char *buf;
                va_list lst2;
                
                fi_got_enomem();
                switch (error_status) {
                    case tsf_error_none:
                        va_start(lst,format);
                        va_start(lst2,format);
                        tsf_assert(buf=tsf_vasprintf(format,lst,lst2));
                        va_end(lst);
                        va_end(lst2);
                        tsf_test_complete(tsf_test_error_dropped,
                                          tsf_error_none,
                                          "Was expecting an out-of-memory "
                                          "error, but it was not propagated "
                                          "correctly.  Instead got: %s",
                                          buf);
                        free(buf);
                        return;
                    case tsf_error_system:
                        if (errno==ENOMEM) {
                            tsf_test_complete(tsf_test_success,tsf_error_none,"Success");
                            return;
                        }
                        break;
                    case tsf_error_tsf:
                        if (tsf_get_error_code()==TSF_E_ERRNO &&
                            tsf_get_system_errno()==ENOMEM) {
                            tsf_test_complete(tsf_test_success,tsf_error_none,"Success");
                            return;
                        }
                        break;
                    default:
                        break;
                }
            }
#endif
            
            display_status_code(status);

            fout=fdopen(result_pipe[1],"w");
            tsf_assert(fout!=NULL);
            fwrite(&status,sizeof(status),1,fout);
            va_start(lst,format);
            vfprintf(fout,format,lst);
            va_end(lst);
            fclose(fout);
            exit(0);
        }
    } else {
        if (status!=tsf_test_success) {
            fprintf(stderr," FAILURE!\n\nIn %s:\n\n",tsf_test_current());
            va_start(lst,format);
            vfprintf(stderr,format,lst);
            va_end(lst);
            fprintf(stderr,"\n");
            exit(1);
        }
        test_count++;
    }
}

tsf_bool_t tsf_do_test_pre_hook(void) {
    fprintf(stderr,".");
    if (do_forking) {
        int res;
        tsf_assert(pipe(result_pipe)==0);
#ifdef FAULT_INJECT
        tsf_assert(pipe(stack_trace_pipe)==0);
#endif
        tsf_assert((res=fork())>=0);
        if (res==0) {
            /* child */
            parent=tsf_false;
            close(result_pipe[0]);
#ifdef FAULT_INJECT
            close(stack_trace_pipe[0]);
            fi_set_checking_enabled(tsf_true);
#endif
            return tsf_true;
        } else {
            /* parent */
            close(result_pipe[1]);
#ifdef FAULT_INJECT
            close(stack_trace_pipe[1]);
#endif
            parse_status();
            close(result_pipe[0]);
            return tsf_false;
        }
    } else {
        return tsf_true;
    }
}

void tsf_do_test_first_hook(void) {
#ifdef FAULT_INJECT
    if (do_forking) {
        fprintf(stderr,":");
        fi_enable();
    }
#endif
}

tsf_bool_t tsf_do_test_loop_hook(void) {
    tsf_bool_t result=tsf_false;
#ifdef FAULT_INJECT
    if (do_forking) {
        if (fi_read_trace(stack_trace_pipe[0]) &&
            fi_have_trace()) {
            fi_add_trace();
            result=tsf_true;
        } else {
            fi_disable();
        }
        close(stack_trace_pipe[0]);
    }
#endif
    return result;
}

void tsf_test_report(void) {
    if (do_forking) {
        unsigned num_total=0,
                 num_succeeded=0,
                 num_failed=0,
                 num_dropped=0,
                 num_system_error=0,
                 num_tsf_error=0,
                 num_exited=0,
                 num_crashed=0,
                 i;
        test_t *cur;
        
        for (cur=head;cur!=NULL;cur=cur->next) {
            num_total++;
            switch (cur->status) {
                case tsf_test_success:
                    num_succeeded++;
                    break;
                case tsf_test_failure:
                    num_failed++;
                    break;
                case tsf_test_error_dropped:
                    num_dropped++;
                    break;
                case tsf_test_system_error:
                    num_system_error++;
                    break;
                case tsf_test_tsf_error:
                    num_tsf_error++;
                    break;
                case tsf_test_exit:
                    num_exited++;
                    break;
                case tsf_test_crash:
                    num_crashed++;
                    break;
                default:
                    tsf_abort("Unknown status code");
            }
        }
        
        if (num_succeeded==num_total) {
            fprintf(stderr," OK! (%u tests)\n",num_total);
        } else {
            fprintf(stderr," FAILED!\n\n");
            fprintf(stderr,"Tests succeeded: %u\n",num_succeeded);
            if (num_failed!=0) {
                fprintf(stderr,"Tests failed:    %u\n",num_failed);
            }
            if (num_dropped!=0) {
                fprintf(stderr,"Errors dropped:  %u\n",num_dropped);
            }
            if (num_system_error!=0) {
                fprintf(stderr,"System errors:   %u\n",num_system_error);
            }
            if (num_tsf_error!=0) {
                fprintf(stderr,"TSF errors:      %u\n",num_tsf_error);
            }
            if (num_exited!=0) {
                fprintf(stderr,"Premature exits: %u\n",num_exited);
            }
            if (num_crashed!=0) {
                fprintf(stderr,"Crashes:         %u\n",num_crashed);
            }
            fprintf(stderr,"Total tests:     %u\n\n",num_total);
            
            for (cur=head,i=1;cur!=NULL;cur=cur->next,++i) {
                if (cur->status==tsf_test_success) {
                    continue;
                }
                
                fprintf(stderr,"Test #%d: %s: ",
                               i,cur->name);
                
                switch (cur->status) {
                    case tsf_test_failure:
                        fprintf(stderr,"(F) FAILURE!\n");
                        break;
                    case tsf_test_error_dropped:
                        fprintf(stderr,"(D) ERROR DROPPED!\n");
                        break;
                    case tsf_test_system_error:
                        fprintf(stderr,"(E) SYSTEM ERROR!\n");
                        break;
                    case tsf_test_tsf_error:
                        fprintf(stderr,"(E) TSF ERROR!\n");
                        break;
                    case tsf_test_exit:
                        fprintf(stderr,"(X) PREMATURE EXIT!\n");
                        break;
                    case tsf_test_crash:
                        fprintf(stderr,"(C) CRASH!\n");
                        break;
                    default:
                        tsf_abort("Unknown status code");
                }
                
                fprintf(stderr,"%s\n",cur->message);
            }
        }
    } else {
        fprintf(stderr," OK! (%u tests)\n",test_count);
    }
}

void tsf_test_done(void) {
    if (do_forking) {
        test_t *cur;
        for (cur=head;cur!=NULL;cur=cur->next) {
            if (cur->status!=tsf_test_success) {
                exit(1);
            }
        }
    }
}


