/*
 * $Header: /Users/pizlo/Repository/tsf/write_cases.c,v 1.1.1.1 2007/05/02 23:31:06 pizlo Exp $
 *
 * write_cases.c -- implementation of common test cases
 * by Filip Pizlo, 2003, 2004, 2005
 */

#include "write_cases.h"
#include "tsf_test.h"
#include "tsf.h"

static tsf_bool_t comments;
static doit_func_t toplevel_action;

static uint32_t comment_count=0;

static const char *make_comment() {
    static char buf[64];
    sprintf(buf,"comment: " fui32,++comment_count);
    return buf;
}

static void do_comment(tsf_type_t **type) {
    uint32_t i;
    
    if (!comments) {
        return;
    }
    
    TAT(tsf_type_set_comment(type,make_comment()));
    
    if (tsf_type_get_kind_code(*type)==TSF_TK_STRUCT) {
        for (i=0;
             i<tsf_struct_type_get_num_elements(*type);
             ++i) {
            TAT(tsf_struct_type_set_comment(
                    type,
                    tsf_named_type_get_name(
                        tsf_struct_type_get_element(*type,i)),
                    make_comment()));
        }
    } else if (tsf_type_get_kind_code(*type)==TSF_TK_CHOICE) {
        for (i=0;
             i<tsf_choice_type_get_num_elements(*type);
             ++i) {
            TAT(tsf_choice_type_set_comment(
                    type,
                    tsf_named_type_get_name(
                        tsf_choice_type_get_element(*type,i)),
                    make_comment()));
        }
    }
}

static tsf_reflect_t *produce_clones_array(tsf_reflect_t *obj,
                                           uint32_t size) {
    tsf_type_t *array_type;
    tsf_reflect_t *r;
    
    TAT(obj);
    
    TAT(array_type=tsf_type_create_array(tsf_reflect_get_type(obj)));
    
    do_comment(&array_type);
    
    TAT(r=tsf_reflect_create(array_type));
    
    while (size-->0) {
        TAT(tsf_array_reflect_append(r,tsf_reflect_clone(obj)));
    }
    
    tsf_reflect_destroy(obj);
    tsf_type_destroy(array_type);
    
    return r;
}

static tsf_reflect_t *produce_struct(const char *first,
                                     ...) {
    va_list lst;
    tsf_type_t *struct_type;
    tsf_reflect_t *result;
    const char *name;
    tsf_reflect_t *value;
    
    TAT(struct_type=tsf_type_create(TSF_TK_STRUCT));
    
    if (first!=NULL) {
        
        va_start(lst,first);
        name=first;
        
        for (;;) {
            value=va_arg(lst,tsf_reflect_t*);
            
            TAT(tsf_struct_type_append(&struct_type,name,
                    tsf_type_dup(tsf_reflect_get_type(value))));
            
            name=va_arg(lst,const char*);
            if (name==NULL) {
                break;
            }
        }
        
        va_end(lst);
        
    }
    
    do_comment(&struct_type);
    
    TAT(result=tsf_reflect_create(struct_type));
    
    tsf_type_destroy(struct_type);
    
    if (first!=NULL) {
        
        va_start(lst,first);
        name=first;
        
        for (;;) {
            value=va_arg(lst,tsf_reflect_t*);
            
            TAT(tsf_struct_reflect_set_element(result,name,value));
            
            name=va_arg(lst,const char *);
            if (name==NULL) {
                break;
            }
        }
        
        va_end(lst);
    }
    
    return result;
}

static tsf_reflect_t *produce_choice(tsf_type_t *type,
                                     const char *value,
                                     tsf_reflect_t *data) {
    tsf_reflect_t *result;
    
    do_comment(&type);
    
    TAT(result=tsf_reflect_create(type));
    
    tsf_type_destroy(type);
    
    if (value==NULL) {
        TAT(tsf_choice_reflect_set_unknown(result));
    } else {
        TAT(tsf_choice_reflect_set(result,
                                   value,
                                   data));
    }
    
    return result;
}

static tsf_reflect_t *produce_any(tsf_reflect_t *sub_r) {
    tsf_buffer_t *buf;
    tsf_reflect_t *r;
    
    TAT(sub_r);
    
    TAT(buf=tsf_reflect_make_buffer(sub_r));
    
    TAT(r=tsf_reflect_create_any(buf));
    
    return r;
}

static void doit_with_any(tsf_reflect_t *r) {
    toplevel_action(produce_any(r));
}

static void doit_with_two_anies(tsf_reflect_t *r) {
    toplevel_action(produce_any(produce_any(r)));
}

static void do_array_tests(doit_func_t func,
                           uint32_t size) {
    DO_TEST(func(produce_clones_array(tsf_reflect_create_int8_t(1),size)));
    DO_TEST(func(produce_clones_array(tsf_reflect_create_uint8_t(2),size)));
    DO_TEST(func(produce_clones_array(tsf_reflect_create_int16_t(3003),size)));
    DO_TEST(func(produce_clones_array(tsf_reflect_create_uint16_t(4004),size)));
    DO_TEST(func(produce_clones_array(tsf_reflect_create_int32_t(500005),size)));
    DO_TEST(func(produce_clones_array(tsf_reflect_create_uint32_t(600006),size)));
    DO_TEST(func(produce_clones_array(tsf_reflect_create_int64_t(700007),size)));
    DO_TEST(func(produce_clones_array(tsf_reflect_create_uint64_t(800008),size)));
    DO_TEST(func(produce_clones_array(tsf_reflect_create_float(9.7),size)));
    DO_TEST(func(produce_clones_array(tsf_reflect_create_double(10.9),size)));

    DO_TEST(func(produce_clones_array(
        tsf_reflect_create_string("hello, world!"),size)));
}

static void do_tests_with_doit(doit_func_t func) {
    DO_TEST(func(tsf_reflect_create_int8_t(1)));
    DO_TEST(func(tsf_reflect_create_uint8_t(2)));
    DO_TEST(func(tsf_reflect_create_int16_t(3003)));
    DO_TEST(func(tsf_reflect_create_uint16_t(4004)));
    DO_TEST(func(tsf_reflect_create_int32_t(500005)));
    DO_TEST(func(tsf_reflect_create_uint32_t(600006)));
    DO_TEST(func(tsf_reflect_create_int64_t(700007)));
    DO_TEST(func(tsf_reflect_create_uint64_t(800008)));
    DO_TEST(func(tsf_reflect_create_float(9.7)));
    DO_TEST(func(tsf_reflect_create_double(10.9)));
    
    DO_TEST(func(tsf_reflect_create_string("hello, world!")));
    
    TEST_PUSH("size=0");
    do_array_tests(func,0);
    TEST_POP();
    TEST_PUSH("size=1");
    do_array_tests(func,1);
    TEST_POP();
    TEST_PUSH("size=2");
    do_array_tests(func,2);
    TEST_POP();
    TEST_PUSH("size=254");
    do_array_tests(func,254);
    TEST_POP();
    TEST_PUSH("size=255");
    do_array_tests(func,255);
    TEST_POP();
    TEST_PUSH("size=256");
    do_array_tests(func,256);
    TEST_POP();
    
    DO_TEST(func(produce_struct(NULL)));
    DO_TEST(func(produce_struct("a",
                                tsf_reflect_create_int8_t(1),
                                NULL)));
    DO_TEST(func(produce_struct("a",
                                tsf_reflect_create_int8_t(2),
                                "b",
                                tsf_reflect_create_float(5.0),
                                "c",
                                tsf_reflect_create_uint32_t(5000),
                                NULL)));
    DO_TEST(func(produce_struct("a",
                                tsf_reflect_create_int8_t(3),
                                "b",
                                produce_struct("c",
                                               tsf_reflect_create_uint32_t(5000),
                                               NULL),
                                NULL)));
    
    DO_TEST(func(produce_struct(
        "a",
        produce_any(tsf_reflect_create_int8_t(4)),
        "b",
        produce_any(tsf_reflect_create_float(5.0)),
        "c",
        produce_any(tsf_reflect_create_uint32_t(5000)),
        NULL)));
    
    DO_TEST(func(produce_choice(
        tsf_type_create_choice(NULL),
        NULL,
        NULL)));
    
    DO_TEST(func(produce_choice(
        tsf_type_create_choice("blah",
                               tsf_type_create(TSF_TK_VOID),
                               NULL),
        NULL,
        NULL)));
    
    DO_TEST(func(produce_choice(
        tsf_type_create_choice("blah",
                               tsf_type_create(TSF_TK_VOID),
                               NULL),
        "blah",
        tsf_reflect_create_void())));
    
    DO_TEST(func(produce_choice(
        tsf_type_create_choice("blah",
                               tsf_type_create(TSF_TK_VOID),
                               "bleh",
                               tsf_type_create(TSF_TK_VOID),
                               NULL),
        NULL,
        NULL)));
    
    DO_TEST(func(produce_choice(
        tsf_type_create_choice("blah",
                               tsf_type_create(TSF_TK_VOID),
                               "bleh",
                               tsf_type_create(TSF_TK_VOID),
                               NULL),
        "blah",
        tsf_reflect_create_void())));
    
    DO_TEST(func(produce_choice(
        tsf_type_create_choice("blah",
                               tsf_type_create(TSF_TK_VOID),
                               "bleh",
                               tsf_type_create(TSF_TK_VOID),
                               NULL),
        "bleh",
        tsf_reflect_create_void())));
    
    DO_TEST(func(produce_choice(
        tsf_type_create_choice("blah",
                               tsf_type_create(TSF_TK_FLOAT),
                               NULL),
        NULL,
        NULL)));
    
    DO_TEST(func(produce_choice(
        tsf_type_create_choice("blah",
                               tsf_type_create(TSF_TK_FLOAT),
                               NULL),
        "blah",
        tsf_reflect_create_float(5.0))));
    
    DO_TEST(func(produce_choice(
        tsf_type_create_choice("blah",
                               tsf_type_create(TSF_TK_FLOAT),
                               "bleh",
                               tsf_type_create(TSF_TK_DOUBLE),
                               NULL),
        NULL,
        NULL)));
    
    DO_TEST(func(produce_choice(
        tsf_type_create_choice("blah",
                               tsf_type_create(TSF_TK_FLOAT),
                               "bleh",
                               tsf_type_create(TSF_TK_DOUBLE),
                               NULL),
        "blah",
        tsf_reflect_create_float(5.0))));
    
    DO_TEST(func(produce_choice(
        tsf_type_create_choice("blah",
                               tsf_type_create(TSF_TK_FLOAT),
                               "bleh",
                               tsf_type_create(TSF_TK_DOUBLE),
                               NULL),
        "bleh",
        tsf_reflect_create_double(5.0))));
}

void do_tests(tsf_bool_t _comments,
              doit_func_t action) {
    comments=_comments;
    toplevel_action=action;
    
    TEST_PUSH("pure");
    do_tests_with_doit(toplevel_action);
    TEST_POP();
    
    TEST_PUSH("with_any");
    do_tests_with_doit(doit_with_any);
    TEST_POP();
    
    TEST_PUSH("with_two_level_any");
    do_tests_with_doit(doit_with_two_anies);
    TEST_POP();
}

